/*
 ===========================================================================
 =                                                                         =
 =  (C) Copyright 1995, Computer Technology Institute (CTI)                =
 =                                                                         =
 =  Permission to use, copy, modify and distribute this software and       =
 =  its documentation for non-commercial use and without fee is hereby     =
 =  granted, provided that the above copyright notice appear in all        =
 =  copies and that both that copyright notice and this permission         =
 =  notice and warranty disclaimer appear in supporting documentation,     =
 =  and that the name of Computer Technology Institute (CTI) or any of     =
 =  its entities not be used in advertising or publicity pertaining to     =
 =  distribution of the software and its documentation without specific    =
 =  prior permission.                                                      =
 =                                                                         =
 =  Computer Technology Institute (CTI) disclaims all warranties with      =
 =  regard to this software and makes no representations about the         =
 =  suitability of this software and its documentation for any purpose.    =
 =  It is provided "as is" without expressed or implied warranty.          =
 =                                                                         =
 ===========================================================================
 =                                                                         =
 =                        Project CTI-Print                                =
 =                                                                         =
 = File:                                                                   =
 =   test_printer.c                                                        =
 =                                                                         =
 = Synopsis:                                                               =
 =   This file contains the functions which check the parallel port for    =
 =   possible problems such as no paper, printer off line etc.             =
 =                                                                         =
 = Author:                                                                 =
 =   Panos Dimakopoulos, Systems Programmer,                               =
 =   Computer Technology Institute,                                        =
 =   Division of Computing Facilities,                                     =
 =   P.O. Box 1122,                                                        =
 =   261 10  Patras,                                                       =
 =   Greece                                                                =
 =   (e-mail: dimakop@cti.gr)                                              =
 =   Tel: +30 61 992061                                                    =
 =   Fax: +30 61 993973                                                    =
 =                                                                         =
 ===========================================================================
*/


/****************************************************************************
 * Modification History:
 * 	Revision 1.1	94/10/10	12:58:58
 *		 Creation.
 * 	Revision 1.2	95/01/30	19:04:04
 *		 Level added to log functions.
 * 	Revision 1.3	95/02/10	18:35:08
 *		 Full CTI address.
 * 	Revision 1.4	95/02/14	13:18:25
 *		 New Copyright notice.
 * 	Revision 1.5	95/08/24	16:51:16
 *		 Porting to Solaris.
 *
 */

#include <sys/types.h>
#include <fcntl.h>
#include <sys/ioccom.h>
#ifdef SVR4
#include <sys/bpp_io.h>
#else
#include <sbusdev/bpp_io.h>
#endif
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <hpiii.h>

static char idtestpr[] = "@(#) test_printer.c 95/08/24 - V 1.5" ;
static int wait_time = 30;	/* Time to wait between two retries (secs) */
static int retries = 0;		/* Number of retries if printer comes off-line
				 * 0 means for ever
				 */
static struct bpp_error_status b_status;

/* Both test_status() and read_status() return the value that is assigned
 * to the pin_status field if this is relevant to the malfunction. Otherwise
 * -1 is returned.
 *  0 is returned on success.
 */
int test_status( fd )
int fd;
{
	if ( ioctl(fd, BPPIOC_TESTIO, &b_status) >= 0 )
		return( 0 );
	else {
		if (errno != EIO) {
			perror("Failure due to non port reasons");
			return( -1 );
		}
		perror("Data transfer impossible via parallel port ");
		return( read_status(fd) ) ;
	}
	
}

int read_status( fd )
int fd;
{
	if ( ioctl(fd, BPPIOC_GETERR, &b_status) < 0 ) {
		perror("Cannot read parallel port status");
		return( -1 );
	} else
		return( b_status.pin_status );
}

check_printer( fd, fp )
int fd;
FILE *fp;
{
	int n;

	if ( (n=test_status(fd)) == 0 ) {
		fprintf(fp,"Printer should be OK (status %d)\n",n);
		return(FILTSUCC);
	}

	if ( n & BPP_ERR_ERR ) {
		if ( n & BPP_PE_ERR ) {
			fprintf(fp,"Printer out of paper\n");
			return(FILTRETRY);
		}
	} else {
		fprintf(fp,"Printer off line or other problem\n");
		return(FILTRETRY);
	}

	fprintf(fp,"Printer off line or other problem\n");
	return(FILTRETRY);
}

void wait_printer()
{
	int e, tries;

	tries = retries;
	while ( (e=check_printer(STDOUT,stderr)) != FILTSUCC )	{
		log(2,"One more retry, tries=%d", tries);
		check_ifexit(e);
		if (retries == 0) continue;
		else
			if (--tries == 0)	break;
	}
	if (e == FILTSUCC)	return;
	else			fexit(FILTABORT);
}

/* Exits if SLEEPTIME is set to 0 */
check_ifexit(i)
int i;
{
	if (!wait_time)	fexit(i);
	else {
		log(2,"Let's sleep for %d secs", wait_time);
		sleep(wait_time);
	}
}

get_retries()
{
	char *s;
	int r;

	s = getenv("RETRIES") ;
	if ( s ) {
		sscanf(s,"%d", &r);
		retries = r;
	}

	s = getenv("SLEEPTIME") ;
	if ( s ) {
		sscanf(s,"%d", &r);
		wait_time = r;
	}
}

fexit(i)
int i;
{
	log(1,"Exit value = %d", i);
	exit(i);
}
