/***************************************************************************
 * U. Minnesota LPD Software * Copyright 1987, 1988, Patrick Powell
 * version 3.3.0 Justin Mason July 1994
 * version 4.1.2 Stefano Ianigro June 1998
 ***************************************************************************
 * MODULE: servicereq.c
 * service requests to lpd
 ***************************************************************************
 * Description:
 * The LPD daemon waits for a request from an associated process using a socket.
 * The main lpd daemon forks a server process,  which connects to the
 * requesting process.  The two processes communicate over the socket.
 * A message packet containing a request is the first information transferred.
 * Some requests will need to have error messages and diagnostics generated
 * by LPD sent to the user;  the "Echo_on_stdout" flag causes errors to be
 * echoed there as well.
 * The following operations are supported:
 *	    \1Printer\n
 * 	    	check the queue for jobs and print any found.
 *          -- acknowlege request and close socket
 *          -- start up the Printer
 *	    \2Printer\n
 *	    	receive a job from another machine and queue it.
 *          -- a complex file transfer protocol is used to transfer files.
 *	    \3Printer [users ...] [jobs ...]\n
 *	    	return the current state of the queue (short form).
 *          -- errors and diagnostics returned as well
 *	    \4Printer [users ...] [jobs ...]\n
 *	    	return the current state of the queue (long form).
 *          -- errors and diagnostics returned as well
 *	    \5Printer Person [users ...] [jobs ...]\n
 *	    	remove jobs from the queue.
 *          -- errors and diagnostics returned as well
 *	    \6Printer Person operation 
 *	    	control operations like enable/disable queueing, etc. 
 *          -- errors and diagnostics returned as well
 *	    \7Printer [users ...] [jobs ...]\n
 *	    	return the current state of the queue (old short form).
 *          -- errors and diagnostics returned as well
 *	    \8Printer [users ...] [jobs ...]\n
 *	    	return the current state of the queue (long form).
 *          -- errors and diagnostics returned as well
 *	    \9Printer Person history person Printer
 *	    	show history of printer "Printer" for user "person"
 *          -- errors and diagnostics returned as well
 ***************************************************************************/

#include "lp.h"
#include "common/rmjob.h"
#include "common/displayq.h"
#include "library/errormsg.h"
#include "LPD/recvfiles.h"

/* command line buffer. This is significantly smaller than most buffers in
 * this code to protect against buffer overflow attacks, a la the Internet
 * Worm vs. finger, and the NCSA httpd (as discussed on bugtraq, Feb 95).
 */
static char command[256];

static char *cmd_names[] = {
    "ERROR",				/* 0 */
    "start printer",			/* 1 */
    "receive files",			/* 2 */
    "displayq (short)",			/* 3 */
    "displayq (long)",			/* 4 */
    "remove job",			/* 5 */
    "control operation",		/* 6 */
    "displayq (short, old format)",	/* 7 */
    "displayq (long, old format)",	/* 8 */
    "history",	    			/* 9 */
};

/***************************************************************************
 * check permissions to see if Host has access to the Printer
 * Return: 1 if permissions to use Printer, 0 otherwise
 ***************************************************************************/

static int
chkhost(void) {
    assert (FQDN && *FQDN);
    assert (Host && *Host);

    if (Get_pc_entry (Printer, Status_pc_vars, Status_pc_len) == 0) {
        log (XLOG_INFO, "recvfiles: non-existent printer: %s");
    }

    if (Get_pc_entry (Printer, Status_pc_vars, Status_pc_len) == 0) {
	log (XLOG_INFO, "recvfiles: non-existent printer: %s", Printer);
    }
    if ((hostcmp (FQDN, Host) == 0) ||		/* this Host */
	(Checkperm (FQDN, (char *) 0, Printer, (int *) 0, (int *) 0, 0) > 0))
    {
	return True;
    }
    return False;
}

/***************************************************************************
 *   1. Reads the first line from the socket
 *   2. Determine service request
 *   3. Call suitable dispatch function
 ***************************************************************************/
void 
servicereq (void)
{
    int n;			/* ACME Integer, Inc. */

    /*
     * use almost bombproof way to read command line from socket
     */
    n = bpread (1, command, sizeof (command));
    if (n < 2) {
	fatal (XLOG_INFO, "servicereq: bad command line");
    }
    Request = command[0];
    if (Request < 1 || Request > 9) {
	fatal (XLOG_INFO, "servicereq: bad request (%d) %s", 
	       Request, &command[1]);
    }
    DEBUG0 ("servicereq: %s requests %d (%s) %s", 
	    FQDN, Request, cmd_names[Request], &command[1]);

    splitline (&command[1]);
    if ((Request != REQ_CONTROL) && (Request != REQ_HISTORY)) {
	/*
	 * extract Printer Name from command
	 */
	Printer = Parms[0].str;
	/*
	 * check to see if you can do the job
	 */
	if (chkhost () == 0) {
	    printf ("permission denied for printer '%s'.", Printer);
	    close (0);
	    close (1);
	    fatal (XLOG_INFO, "Host %s cannot access %s", FQDN, Printer);
	}
    }
    switch (Request) {
	/*
	 * start Printer
	 */
    case REQ_START:
	/* signal the other end that you are doing the request */
	putchar (0);		/* fd 1 is the socket */
	(void) fflush (stdout);	/* send it */
	(void) close (1);	/* disconnect */
	if (dup2 (0, 1) < 0) {
	    logerr_die (XLOG_INFO, "dup2 failed");
	}
	Startprinter ();	/* start Printer */
	break;
	/*
	 * get remote jobs
	 */
    case REQ_RECV:		/* get files from remote site */
	recvfiles ();
	Startprinter ();	/* start Printer */
	break;
	/*
	 * display current queue
	 */
    case REQ_DSHORT:		/* display the queue (short form) */
    case REQ_DLONG:		/* display the queue (long form) */
	Queue_format = Request;
	Shift_parms (1);
	Displayq (True, False);
	break;
	/*
	 * remove job from queue
	 */
    case REQ_REMOVE:		/* remove a job from the queue */
	/* echo errors on stdout as well as stderr */
	Echo_on_stdout = 1;
	Person = Parms[1].str;
	Shift_parms (2);
	mail_to_user_for_remote_remove = 1;
	rmjob ();
	mail_to_user_for_remote_remove = 0;
	break;
	/*
	 * perform control function
	 */
    case REQ_CONTROL:
	Echo_on_stdout = 1;
	Person = Parms[0].str;
	Shift_parms (1);
	(void) Control_ops ();
	break;
	/*
	 * display current queue in old format 
	 */
    case REQ_DSHORT_OLD:
	Queue_format = Request;
	Shift_parms (1);
	Displayq (True, True);
	break;
	/*
	 * perform history function
	 */
    case REQ_HISTORY:
	Echo_on_stdout = 1;
	Person = Parms[0].str;
	Shift_parms (1);
	(void) Control_ops ();
	break;
    default:
	fatal (XLOG_INFO, "%s: wierd request -- ignoring; '\\%03o %s'",
		Request, &command[1]);
	break;
    }
}
