/***************************************************************************
 * U. Minnesota LPD Software * Copyright 1987, 1988, Patrick Powell
 * version 3.3.0 Justin Mason July 1994
 ***************************************************************************
 * MODULE: Startprinter.c
 * Printer queue job handler
 *
 * StartPrinter is responsible for handling the Printer queue.
 * 1.  Get the printcap entry.
 * 2.  Check the spool directory to see if printing is enabled
 * 3.  Check to see if there are any jobs
 * 4.  If a remote Printer, call remotePrinter()
 *     If a local Printer, call localPrinter()
 ***************************************************************************/

#include "lp.h"
#include "library/errormsg.h"
#include "LPD/remoteprinter.h"
#include "LPD/localprinter.h"
#include "LPD/setproctitle.h"
#include "LPD/lpd.h"
#include "common/setstatus.h"

extern int Single_printer, Inetd_mode;

static int job_removed;		/* set if the current job gets lprmed */
static plp_signal_t handle_remove ();	/* handle 'remove job' signals */

static int
check_limits (struct plp_queue *q) {
    long limit;
    /*
     * Check the size.  The mx field in the printcap entry specifies
     * the maximum size of job which can be queued.  Here we check on
     * the current limit on jobs to be _printed_.
     * 
     * If current_limit() returns -1 then there is no limit on the size of jobs to print.
     */
    limit = current_limit ();
    DEBUG2 ("check_limits: current value %ld for %s", (long) limit, q->q_name);
    if ((limit != -1L) && ((q->q_size / 1024) > limit)) {
	DEBUG2 ("check_limits: not doing %s, size = %ld, limit = %ld",
		 q->q_name, (long) (q->q_size / 1024), (long) limit);
	return False;
    }
    return -1;
}

int 
opt_match (char *opt, char *loaded) {
    /*
     * It is possible for a printer to be capable of accepting jobs for more than one
     * forms type at a time.  The most likely example would be a printer with headed
     * paper loaded in one bin and plain in the other.  We want to be able to send both
     * plain and headed jobs to this printer.  To do this we have
     * introduced the concept of multiply loaded forms.  If the loaded forms 
     * spec is of the form form1/form2 then jobs of type form1 and form2 
     * can be printed.
     * 
     * This now copes with more than two forms, eg a/b/c
     */
    char *lf, *form;

    lf = strdup (loaded);
    lf[39] = '\0';

    form = strtok (lf, "/");
    while (form != NULL) {
	if (strsame (opt, form)) {
	    free (lf);
	    return False;
	}
	form = strtok (NULL, "/");
    }
    free (lf);
    return True;
}

static int
check_forms (struct plp_queue *q) {
    extern char *current_options (void);
    char *options;		/* current printer options (= forms) */

    /*
     * Print the job if
     * 
     * 1. neither the job nor the printer specify special options or 
     * 2. they both do and they're the same
     */
    if (FL && *FL && (options = current_options ())) {
	if (!((q->q_options == NULL &&
	       (options == NULL || strsame (options, DEF_OPTIONS)
		|| strsame (options, "none"))) ||
	      (q->q_options != NULL &&
	       options != NULL && !opt_match (q->q_options, options))))
	{
	    DEBUG2 ("unspool: not doing %s, options = %s, current = %s",
		     q->q_name,
		     q->q_options ? q->q_options : "NULL",
		     options ? options : "NULL");
	    return False;
	}
    }
    return True;
}

/*
 * do_job_now() makes the decision about whether a job should be sent to the 
 * printer now on not.  Reasons for not sending a job include 
 * 1. the job is held 
 * 2. the job is larger than the current print limit 
 * 3. the job's forms do not match the forms currently set on the printer
 * 
 * If the printer is remote we just pass the job through.
 * This routine returns 0 when the job should not be printed; 1 when should be printed.
 */
static int
do_job_now (struct plp_queue *q) {
    DEBUG3 ("do_job_now: %s, RM = %s; value of q_held: %d",
	     q->q_name, (RM != NULL) ? RM : "NULL", q->q_held);

    /* remote printer - let the job through */
    if ((RM != NULL) && (*RM != '\0')) {
	/* But keep held jobs in the queue */
	if (q->q_held) 
	    {
		DEBUG3 ("do_job_now: %s RM = %s will not be printed; reason: held.",
			q->q_name, (RM != NULL) ? RM : "NULL");
		return False;
	    } 
	else 
	    {
		DEBUG3 ("do_job_now: %s RM = %s will be printed.",
			q->q_name, (RM != NULL) ? RM : "NULL");
		return True;
	    }
    }
    /* Held job - not to be printed now */
    if (q->q_held) {
	DEBUG3 ("do_job_now: %s will not be printed; reason: held.",
		q->q_name);
	return False;
    }
    
    /* check forms */
    if (!check_forms (q)) {
	DEBUG3 ("do_job_now: %s will not be printed; reason: forms.",
		q->q_name);
	return False;
    }
    
    /* check limts */
    if (LI && *LI && (!check_limits (q)))
	{
	    DEBUG3 ("do_job_now: %s will not be printed; reason: limits.",
		    q->q_name);
	    return False;
	}
    /*
     * If we've got this far then we haven't found a reason not to print the job.
     */
    return True;
}

/* check the lockfile to see if the queue should be reread.  */
int
reread_queue (int LO_fd, struct stat *statb) {
    if (fstat (LO_fd, statb) < 0) {
	logerr_die (XLOG_NOTICE, "cannot stat %s", LO);
    }
    return ((statb->st_mode & FORCE_REQUE) ||
	    (statb->st_mode & DISABLE_PRINT));
}

/***********************************************************
 * unspool (init, jobhandler, closer)
 *
 * do{
 *   check queue for new jobs
 *   done_a_job = 0
 *   for(n = 0;n < Jobcount; ++n)
 *      if queue needs to be resorted then
 *         set done_a_job and break from "for" loop;
 *      while not successful and retry count is less than a limit do
 *         init();
 *         success = jobhandler(job n);
 *     -- note: job can be busy (no action), done successfully,
 *              done unsuccessfully and should be abandoned,
 *              or done unsuccessfully and should be repeated
 *      if action was done then set done_a_job and remove job
 *          else call closer();
 * }while (done_a_job)
 * call closer()
 ***********************************************************/

static void
unspool (int (*init) (), int (*dojob) (), void (*closer) (),
	void (*err) ())
{
    int n;			/* job number */
    int retries;		/* number of retries */
    int success;		/* job success status */
    struct plp_queue *q;	/* job entry  */
    int done_a_job;		/* job done in this pass */
    FILE *cfp;			/* control file */
    int init_done;		/* initialization done */
    int not_nows;		/* number of inactive (eg. held) jobs */
    int no_more_jobs;		/* flag: no more jobs to process */

    no_more_jobs = not_nows = init_done = False;

    do {
	do {
startcheck:
	    /* check for more jobs: if there aren't any, then break from
	     * the loops and let the server process exit.
	     */
	    Jobcount = Getq ();
	    if (Jobcount <= not_nows) {
		no_more_jobs = True;
		break;
	    }

	    /*
	     * search the spool directory for more work.
	     */
	    DEBUG4 ("unspool: %d jobs to unqueue", Jobcount);
	    /*
	     * reset "update queue order" flag
	     */
	    if (fstat (fileno (Lfd), &LO_statb) < 0) {
		logerr_die (XLOG_NOTICE, "unspool: cannot stat %s", LO);
	    }
	    DEBUG5 ("unspool: '%s' perms 0%o", LO, LO_statb.st_mode);
	    if (LO_statb.st_mode & DISABLE_PRINT) {
		DEBUG0 ("unspool: unspooling disabled for printer %s, SS = %s", Printer, (SS) ? SS : "none");;
		/* @@@ Fix from Dave Atkin */
		return; /* Avoid infinite loop */
		/* end fix */
		/* break;  */
	    }
#ifdef HAVE_FCHMOD
	    if (fchmod (fileno (Lfd), (int) (LO_statb.st_mode & CLEAR_REQUE)) < 0)
#else
	    if (chmod (LO, (int) (LO_statb.st_mode & CLEAR_REQUE)) < 0)
#endif
	    {
		logerr_die (XLOG_NOTICE, "unspool: cannot chmod '%s'", LO);
	    }
	    done_a_job = False;
	    /*
	     * scan the queue
	     */
	    DEBUG4 ("unspool: running through jobs");

	    not_nows = False;
	    for (n = 0; n < Jobcount; ++n) {
	        q = &Queue[n];

		/*
		 * leave until other jobs have gone through, forms are fitted, etc.
		 */
		if (!do_job_now (q)) {
		    DEBUG5 ("unspool: job %s: can't do now", q->q_name);
		    not_nows++;
		    continue;
		}

		/*
		 * lock the control file
		 */
		if ((cfp = Lockcf (q->q_name)) == NULL) {
		    DEBUG4 ("unspool: job %s: already being processed", q->q_name);
		    not_nows++;
		    continue;
		}

		Setlockfile (LO, Lfd, getpid (), q->q_name);
		/*
		 * try printing this file until successful, number of attempts
		 * exceeds limit, or abort try sending indefinately until successful
		 */
		success = JFAIL;

		/*
		 * If RT is 0 then retry forever (this is needed for pad printers).
		 *
		 * Paul Haldane says: FYI - pad printers are
		 * printers connected to X.25 terminal lines - we
		 * used to drive almost all of our printers that way -
		 * now we have only one or two - worked by making
		 * call to the printer's X.25 address - as many
		 * machines could drive the same printer it is quite
		 * normal to have many retries.
		 */
		for (retries = 0; (success == JFAIL || success == JFULL)
				      && (retries <= RT || RT <= 0); ++retries)
		{
		    if (job_removed) {
			DEBUG2 ("unspool: job %s: removed", q->q_name);
			break;	/* out of retry loop */
		    }
			
		    if (success == JFULL) {
			/* remote fs is full -- sleep 10 mins and retry */
			(void) sleep (600);
			init_done = False;
		    }
		    /*
		     * STDIO file error reset, needed on some implementations
		     */
		    clearerr (cfp);
		    /* seek to the start of the file */
		    if (fseek (cfp, 0L, 0) < 0) {
			logerr_die (XLOG_INFO, "unspool: fseek failed (%s)",
				    q->q_name);
		    }

		    DEBUG0 ("unspool: job %s: starting, attempt %d...", 
			    q->q_name, retries);

		    /* initialize */
		    if (init_done == 0) {
			init_done = (*init) ();
			if (init_done != JSUCC) {
			    setstatus ("Can't initialize device: %s.", Last_errormsg);
			    return;
			}
		    }
		    if (!called_from_one_job) {
			setstatus ("Printing (started at %s, attempt %d).",
				   Time_str (), retries + 1);
		    }
		    if (Errors_file && *Errors_file)
			unlink (Errors_file);		/* empty it */

		    /* try to print the job */
		    DEBUG2 ("unspool: job %s: starting unqueue", q->q_name);

		    success = (*dojob) (cfp, q);

		    DEBUG2 ("unspool: job %s: retry %d (of %d): status %d",
			     q->q_name, retries + 1, RT, success);

		    if (success == JFAIL) {
			init_done = False;

		    } else if (success == JNOCONF) {
			/* we didn't get a final confirm; job has 
			 * probably printed, but the link is closed
			 * anyway, so we'll need to re-open it.
			 */
			init_done = False;
			success = JSUCC;
		    }
		    if (success != JSUCC)
			sleep (5);
		}

		if (job_removed) {
		    DEBUG2 ("unspool: job %s: out of retries: job removed", q->q_name);
		    job_removed = False;

		} else {
		    DEBUG2 ("unspool: job %s: out of retries: status %d",
			     q->q_name, success);

		    if ((success == JFAIL) || (success == JFULL) ||
				(success == JABORT))
		    {
			DEBUG2 ("unspool: job %s: failed to print", q->q_name);
			(*err) ();
			init_done = False;

		    } else {
			DEBUG2 ("unspool: job %s: printed successfully", q->q_name);
		    }

		    /* send mail if user asks for it or if job failed */
		    if (MAILNAME[0] || (success != JSUCC)) {
			sendmail (q, success);
		    }
		}
		done_a_job = True;
  
                /* Remove job only if it has been successful or aborted,
                 * otherwise it should be left for future attempts. */
                if (success != JFAIL) {
		    Remove_job (cfp, q);
		}

		(void) fclose (cfp);
		if (reread_queue (fileno (Lfd), &LO_statb)) {
		    DEBUG5 ("unspool: rechecking queue");
		    goto startcheck;
		}
	    }

	    Reapchild ();
	} while (done_a_job);		/* until we've run out of active jobs */
    } while (no_more_jobs == 0);

    /*
     * queue finished
     */
    DEBUG0 ("unspool: no more work to do");

    if ((SS && LO_statb.st_mode & DISABLE_PRINT) || (!called_from_one_job)) {
	; /* do not change the status!! */
    } else {
	setstatus ("Closing since %s.", Time_str ());
    }
    (*closer) ();
}

/***************************************************************************
 * Start_server()
 *
 * start multiple servers for a single queue. The SV printcap parameter
 * has a list of the servers. These servers use the spool queues
 * directory, and have a lock file in the server directory.
 *
 * This routine will extract the names of each server, fork a process
 * for it, and then will read the printcap information for the server
 * from the queue.
 *
 * The printer names are separated by commas.
 ***************************************************************************/

static void
Start_server (void) {
    char *ps;
    char *last;
    pid_t pid;
    char parent;
    char *Inherit_from;

/* Changed logic of this function so that process for each printer
 * is child of main daemon (which immediately exits here).  Previous
 * code lead to first n-1 processes being children of last one and if
 * this last one exits any of the others who are sleeping get a HUP
 * and exit.  I don't understand why this hasn't been a problem up
 * till now.
 *                   -- Paul Haldane (EUCS)
 */

    parent = False;
    last = NULL;
    ps = SV;
    /* make a copy so we can inherit from the original printcap entry...
     * it'll won't be changed until we exit, so just forget about freeing
     * it. */
    Inherit_from = strdup (Printer);

    while (ps) {
	Printer = ps;
	if ((last = (char *) strchr (ps, ','))) {
	    *last = '\0';
	    ps = last + 1;
	} else {
	    ps = NULL;
	    parent = True;
	}

	if ((pid = fork ()) < 0) {
	    logerr_die (XLOG_INFO, "Start_server: fork failed");

	} else if (pid == 0) {
	    ps = NULL;		/* so we fall out of loop */
	    parent = False;
	}
    }
    if (parent) {		/* we are the parent */
	exit (0);
    }

    /* make a copy to avoid getting scribbled on by Get_multi_pc_entry...
     * it'll won't be changed until we exit, so just forget about freeing
     * it. */
    Printer = strdup (Printer);

    if (Get_multi_pc_entry (Printer, Inherit_from, All_pc_vars,
	All_pc_len) == False)
    {
	fatal (XLOG_INFO, "trying to start non-existent server");
    }
    LO = Printer;
    DEBUG3 ("Start_server: started");
}

void
Startprinter (void) {
    pid_t pid;			        /* process id */
    char *ps;		                /* ACME Pointers */
    char st[64];                        /* ACME Characters, Inc. */
    FILE *afp;			        /* attach file pointer */
    static char afname[MAXPARMLEN];	/* attach file name */
    static char spoold[MAXPARMLEN];	/* actual spool directory */

    *spoold = '\0';

    /*
     * handle rmjob requests.
     */
    job_removed = False;
    (void) plp_signal (SIGINT, handle_remove);

reattach:  /* restart from here if printer is attached to another */
    /*
     * get the printcap entry
     */
    if (Get_pc_entry (Printer, All_pc_vars, All_pc_len) == False) {
	log (XLOG_INFO, "trying to start non-existent Printer");
	exit (0);
    }
    /*
     * set the Debug flag with values from the printcap
     */
    if (DB && Debug == 0) {
	Debug = DB;
    }
    if (SS && *SS) {
	/*
	 * we have a spool server that is to be started
	 */
	ps = PS;
	(void) strncpy (st, ST, sizeof (st) -1);
	if (Get_multi_pc_entry (SS, Printer, All_pc_vars,
		All_pc_len) == False)
	{
	    log (XLOG_INFO, "queue %s does not exist!", SS);
	    exit (0);
	}
	PS = ps;
	(void) strncpy (ST, st, strlen (st) - 1);
	LO = Printer;
	SV = NULL;
    }
    if (SD == NULL || *SD == '\0') {
	/*
	 * no spooling directory, not a printer
	 */
	DEBUG0 ("Startprinter: not a printer -- no 'sd'");
	exit (0);
    }
    if (strsame (SD, spoold)) {		/* we've been here before */
	fatal (XLOG_INFO, "Printer %s attached to itself", Printer);
    }

    if (RM && NW && !*spoold) {
	(void) Remote_start ();
	exit (0);
    }

    chdir_SD ();

    /*
     * check to see if attached to another printerq, alter printer if attached
     */
    if (*Attach_file && ((afp = fopen_daemon (Attach_file, "r")) != NULL)) {
	if (!*spoold)	
	    (void) strcpy (spoold, SD);		/* set actual spool directory */

	if (fscanf (afp, "%s", afname) == 1) {
	    if (strsame (afname, Printer)) {
		fatal (XLOG_INFO, "Printer %s attached to itself", Printer);
	    }
	    Printer = afname;
	}
	DEBUG2 ("printer reattached to %s", Printer);
	(void) fclose (afp);
	goto reattach;
    }
    if (*spoold) {		/* printer re-attached */
	SD = spoold;		/* set spool directory */
	if (!RP || !*RP)	/* set remote printer if necessary */
	    RP = Printer;
	if (!RM || !*RM)	/* set remote machine */
	    RM = FQDN;
	chdir_SD ();
    }
    /*
     * check to see if there is a daemon running
     */
    Lfd = Getlockfile (LO, &pid, (char *) 0, 0, &LO_statb);
    if (Lfd == NULL) {
	DEBUG2 ("Startprinter: server present, pid %u", pid);

	if (chmod (LO, (int) (LO_statb.st_mode & 0777) | FORCE_REQUE) < 0) {
	    logerr (XLOG_INFO, "cannot force requeue on lockfile %s", LO);
	}
	exit (0);
    }
    (void) fseek (Lfd, 0L, 0);
    /*
     * check for unspooling enabled
     */
    if (LO_statb.st_mode & DISABLE_PRINT) {
	DEBUG0 ("Startprinter: unspooling disabled for printer %s.", Printer);
	exit (0);
    }
    /*
     * handle multiple servers
     */
    if (SV && *SV) {
	Closelockfile (LO, Lfd);
	Lfd = 0;
	Start_server ();
	/*
	 * Check on the status of the queue lock file
	 */
	Lfd = Getlockfile (LO, &pid, (char *) 0, 0, &LO_statb);
	if (Lfd == NULL) {
	    DEBUG2 ("Startprinter: server present, pid %u", pid);
	    if (chmod_daemon (LO, (int) (LO_statb.st_mode & 0777) | FORCE_REQUE) < 0) {
		logerr (XLOG_INFO, "cannot force requeue on lockfile %s", LO);
	    }
	    exit (0);
	}
    }
    /*
     * set up log file
     */
    (void) fseek (Lfd, 0L, 0);

    if (!Single_printer)	/* queue debugging mode */
	Setuplog (LF, 0);

    if (fseek (Lfd, 0L, 0) < 0) {
	logerr_die (XLOG_INFO, "Startprinter: T3 fseek failed");
    }

    if (!Single_printer && !Inetd_mode) {
	/* OK, you are the server, go to it. set process group. */
	if (plp_setsid () == -1)
	    logerr_die (XLOG_INFO, "setsid failed");
    }

    Setlockfile (LO, Lfd, getpid (), "(checking queue)");
    if (LO_statb.st_mode & DISABLE_PRINT) {
	DEBUG0 ("Startprinter: unspooling disabled for printer %s.", Printer);
    } else {
	if (!called_from_one_job) {
	    setstatus ("Printing since %s.", Time_str_long ());
	}
	setproctitle ("%s: unqueueing jobs", Printer);
    }
    /*
     * search the spool directory for work and sort by queue order.
     */
    if (RM && !QH) {
	unspool (Remote_setup, Remote_dojob, Remote_finish, Remote_error);
    } else {
	unspool (Local_setup, Local_dojob, Local_finish, Local_error);
    }
    DEBUG2 ("Startprinter: work done");

    if (LO_statb.st_mode & DISABLE_PRINT) {
	; /* do not change the status!! */
    } else {
	DEBUG2 ("Startprinter: setting status of %s to ready.", Printer);
	if (!called_from_one_job) {
	    setstatus ("Ready since %s.", Time_str_long ());
	}
    }

    /* setproctitle unneeded -- we're about to exit */
    Closelockfile (LO, Lfd);
    exit (0);
}

int
Job_removed (void) {
    if (job_removed) {
        DEBUG0 ("current job is lprmed: faking successful xfer");
    }
    return (job_removed);
}

extern void Print_intr ();

/*
 * handle 'remove-job' signals. Used to signal an lpd to stop
 * processing the current job, and to skip it.
 */
static plp_signal_t
handle_remove (int sig) {
    Print_intr ();	/* in case we're printing */
    /* Fix from Dave Atkin */
    if (qh_pid) {
        DEBUG3 ("handle_remove: Sending SIGINT to queue handler pid %d", qh_pid);
        (void) kill (qh_pid, SIGINT);    /* Tell queue handler about it */
    }
    /* end fix */
    job_removed = True;
#ifdef SVR4
    plp_signal (sig, handle_remove);
#endif
}
