/***************************************************************************
 * U. Minnesota LPD Software * Copyright 1987, 1988, Patrick Powell
 * version 3.3.0 Justin Mason July 1994
 * version 4.1.2 Stefano Ianigro June 1988
 ***************************************************************************
 * MODULE: lpq.c
 ***************************************************************************/

#include "lp.h"
#include "common/displayq.h"
#include "library/errormsg.h"
#include "library/readconf.h"
#include "LPQ/termclear.h"
#include "patchlevel.h"

/***************************************************************************
 *   lpq - spool queue examination program
 *   lpq [-Pprinter ][-a][-h[a[ll]]][-O][-l][+[n]][-D[n]][ job # ... ][ user ... ]
 *   lpstat <options>
 * DESCRIPTION
 *   lpq examines the spooling area used by lpd(8) for printing
 *   files on the line printer, and reports the status of the
 *   specified jobs or all jobs associated with a user. lpq
 *   invoked without any arguments reports on the printer given
 *   by the default printer (see -P option).  For each job sub-
 *   mitted (i.e. invocation of lpr(1)) lpq reports the user's
 *   name, current rank in the queue, the names of files compris-
 *   ing the job, the job identifier (a number which may be sup-
 *   plied to lprm(1) for removing a specific job), and the total
 *   size in bytes.  Job ordering is dependent on the algorithm
 *   used to scan the spooling directory and is FIFO (First in
 *   First Out), in order of priority level.  File names compris-
 *   ing a job may be unavailable (when lpr(1) is used as a sink
 *   in a pipeline) in which case the file is indicated as
 *   ``(stdin)''.  The following options are available.
 *   -P printer     spool queue name
 *   -a             All printers listed in /etc/printcap(5) file are reported.
 *   -l             long display format
 *   -O             old display format
 *   [+[n]]         redisplay the spool queue
 *   -D[n]          debug level n
 *   -h [all][user] history of printed jobs - all or for one user (c perms required)
 *   [ job # ... ][ user ... ]
 * This program may also be called as lpstat; it has then the (nearly) 
 * same functionalty as the lpstat command used in UNIX SYS V Rel.4
 ***************************************************************************/

static int Delay = 0;			/* time to wait */
static int Do_all_pr = False;		/* do all printers */
static int Name_all_pr = False;		/* list all printers by name */
static char **Printer_name;
static int Printer_count = 0;
static int Max_printer_count = 0;
static int Old_long_format;
static int history = 0;
static int all_history = 0;
static int d_found = 0, l_found = 0;
static char history_user[10];

void display_history (void), history_all (void), show_history (void);
void check_lpd (void), show_default_destination (void);
void  show_lpstat_v_option (char *), show_lpstat_a_option (char *);
void show_lpstat_o_option (char *), show_lpstat_p_option (char *);
void show_lpstat_u_option (char *);
void display_lpstatus (void), display_lp_printer (void);
void display_lp_printer_user (char *);
int do_checkperms_lpstat (char *, char *, char *, int *);
static void no_printer (char *);
char *short_time_for_lpq (time_t);
extern void Print_queuehead_hist (void), Print_comment (char *);
extern int Checkactive (void);
static char *short_time_lpstat (time_t);
extern int printer_compare ();

static void 
Usemsg (void)
{
    if (sysv_lp) {
    (void) fprintf (stderr,
    "Usage: lpstat [options] [request-ids]\n\
          [-a printers]                   (show acceptance status)\n\
          [-c classes]                    (show available classes)\n\
          [-d]                            (show default destination)\n\
          [-D]                            (-p only: describe printers)\n\
          [-f forms]                      (show available forms)\n\
          [-l]                            (be verbose)\n\
          [-o printers,classes,req-ids]   (show status of requests)\n\
          [-p printers]                   (show available printers)\n\
          [-P]                            (show paper types)\n\
          [-R]                            (show rank in queue)\n\
          [-r]                            (show status of Spooler)\n\
          [-s]                            (show summary status)\n\
          [-t]                            (show status of everything)\n\
          [-u users]                      (show status of user requests)\n\
          [-v printers [-H]]              (show devices used by printers)\n\
                  (\"all\" allowed with -a,-c,-f,-o,-p,-u,-v options)\n");
    } else {
    (void) fprintf (stderr,
		    "Usage: lpq [-O] [-Pprinter] [-a[ll]] [+[n]] [-h [all|user]] [-Dn] (job# | user)*\n");
    }
}

plp_signal_t 
cleanup (int signal)
{
    Term_finish ();
    exit (1);
}

static void 
Check_already (int c, int *v)
{
    if (*v) {
	if (c != 'a') {
	Usemsg ();
	Diemsg ("-%c has already been specified", c);
	}
    }
    *v = 1;
}

/***************************************************************************
 * 1. set up the name routine
 * 2. extract parameters
 *  -Pprinter sets Printer to printer
 *  -Dn set Debug to n
 *  -a   sets Do_all_pr
 *  -n   sets Name_all_pr
 *  -l   set Queue_format = REQ_DLONG
 *  +nn
 ***************************************************************************/
static void 
Get_parms (int argc, char *argv[])
{
    int c, i;			/* ACME Integers, Inc. */
    char *arg;			/* argument from list */
    int opts_available = 0;
    Queue_format = REQ_DSHORT;
    Old_long_format = False;
    Do_all_pr = Name_all_pr = False;
    Max_printer_count = Printer_count = 0;
    Delay = 0;

    Opt_flag = "+";
    c = 0;

    /* called as lpstat */
    if (sysv_lp) {
	/* First we have to search on argv if there is a -D or -l option */
	for (i = 0; i < argc; i++) {
	    if (strcmp (argv[i], "-D") == 0) { /* found -D */
		d_found = 1;
	    }
	    if (strcmp (argv[i], "-l") == 0) { /* found -l */
		l_found = 1;
	    }
	}
	/* in case the user sets both of them... */
	if (d_found == 1 && l_found == 1) {
	    d_found = 0;
	    l_found = 1;
	}
	
	/* evaluate arguments */
	while ((c = Getopt_lpstat (argc, argv, "a:c?dDf?lPp:o:RrsStu:v:")) != EOF) {
	    switch (c) {
	    case 'a':
		opts_available = 1;
		show_lpstat_a_option (Optarg);
		break;
		/* shows the classes; not used in PLP */
	    case 'c':
		opts_available = 1;
		(void) printf ("character set\n");
		break;
		
               /* prints short information
                * handled above
                */
	    case 'D':
		break;

		/* get the default printer; this is
		 * a. if the environment variable PRINTER is set: 
		 *    this printer or
		 * b. the printer with the name "lp" or
		 * c. the first printer found in the printcap database
		 */
	    case 'd':
		opts_available = 1;
		show_default_destination ();
		break;

		/* No forms used in PLP */
	    case 'f':
		opts_available = 1;
		{
		    /* Forms mounted valid for local printers only */
		    if (!RP && !RM) {
			char *forms_mounted;
			/* Forms mounted */
			(void) fprintf (stdout, "Form mounted: ");
			forms_mounted = current_options ();
			if (forms_mounted) {
			    if (!strsame (forms_mounted, DEF_OPTIONS)) {
				(void) fprintf (stdout, "%s", forms_mounted);
			    }
			} else {
			    (void) fprintf (stdout, "none");
			}
			(void) fprintf (stdout, "\n");
		    }
		}
		break;

		/* be verbose; handled above */
	    case 'l':
		break;

		/* shows the output of queues/printers */
	    case 'o':
		opts_available = 1;
		show_lpstat_o_option (Optarg);
		break;

		/* prints the status of the printer(s) */
	    case 'p':
		opts_available = 1;
		show_lpstat_p_option (Optarg);
		break;

		/* displays the status if lpd is running */
	    case 'r':
		opts_available = 1;
		check_lpd ();
		break;

		/* shows the printer destinations */
	    case 's':
		opts_available = 1;
		check_lpd ();
		show_default_destination ();
		show_lpstat_v_option (Optarg);
		(void) printf ("character set\n");
		break;

		/* shows everything */
	    case 't':
		opts_available = 1;
		check_lpd ();
		show_default_destination ();
		show_lpstat_v_option (Optarg);
		show_lpstat_a_option (Optarg);
		show_lpstat_p_option (Optarg);
		(void) printf ("character set\n");
		show_lpstat_o_option (Optarg);
		break;

		/* Shows the job(s) of a specified user;
		 * default is all
		 */
	    case 'u':
		opts_available = 1;
		if (!Optarg) {
		    show_lpstat_o_option (Optarg);
		} else {
		    show_lpstat_u_option (Optarg);
		}
		break;

		/* From the man page of lpstat (Solaris 2.5.1):
		 * -v [list]
		 * Print the names of  printers  and  the  path
		 * names  of the devices associated with them or
		 * remote system  names  for  network  printers:
		 * list is a list of printer names
		 */
	    case 'v':
		opts_available = 1;
		show_lpstat_v_option (Optarg);
		break;

	    default:
		Usemsg ();
		exit (1);
	    }
	}

	/* if there are no options (execept -l or -D), then we 
	 * do a lpstat -u user 
	 */
	if (!opts_available) {
	    show_lpstat_u_option (Person);
	}
	/* No more to do here so we can exit now */
	exit (0);
    } else { 	/* called as lpq */
	while ((c = Getopt (argc, argv, "P:D:ah:lnOst+?")) != EOF) {
	    switch (c) {
	    case 'O':
		Old_long_format = True;
		Queue_format = REQ_DSHORT_OLD;
		break;
	    case 'P':
		if (!Optarg) {
		    break;
		}
		if (Printer_count == Max_printer_count) {
		    if (Max_printer_count == 0) {
			Max_printer_count = 3;
			Printer_name = (char **) malloc
			    (Max_printer_count * sizeof (char *));
		    } else {
			Max_printer_count += 3;
			Printer_name = (char **) realloc (Printer_name,
							  Max_printer_count * sizeof (char *));
		    }
		}
		Printer_name[Printer_count] = Optarg;
		++Printer_count;
		break;
	    case 'D':
		if (!Optarg)
		    break;
		Check_already (c, &Debug);
		Parse_debug (Optarg);
		break;
	    case 'a':
		Check_already (c, &Do_all_pr);
		break;
	    case 'h':
		if (!Use_history) {
		    (void) printf ("History feature not active.\n");
		    exit (0);
		} else {
		    history = 1;
		}
		if (Optarg) {
		    /* we have the option "a" or exactly "all" or...*/
		    if (((Optarg[0] == 'a') && (Optarg[1] == '\0')) || 
			((Optarg[0] == 'a') && (Optarg[1] == 'l') && 
			 (Optarg[2] == 'l')&& (Optarg[3] == '\0'))) {
			all_history = 1;
			history_user[0] = '\0';
		    } else {
			/* a username here; 
			 * we will copy only max 8 bytes in the array history_user!
			 */
			(void) strncpy (history_user, Optarg, 8);
			all_history = 0;
		    }
		}
		break;
	    case 'l':
		Queue_format = REQ_DLONG;
		break;
	    case 'n':
		Check_already (c, &Name_all_pr);
		break;
	    case 's':
		Queue_format = REQ_DSHORT;
		break;
	    case '+':
	    /**************************************
	       [+[n]] set Delay to n
	    **************************************/
	    Check_already (c, &Delay);
	    if (Optarg) {
		Delay = atoi (Optarg);
		if (Delay <= 0) Delay = 30;
	    } else {
		Delay = 30;
	    }
	    break;
	    default:
		Usemsg ();
		exit (1);
	    }
	}
    }

    allocParmsIfNeeded ();
    for (; Optind < argc; ++Optind) {
	arg = argv[Optind];
	Parms[Parmcount].str = arg;
	if (isdigit (*arg)) {
	    Parms[Parmcount].num = atoi (arg);
	} else {
	    Parms[Parmcount].num = -1;
	}
	++Parmcount;
	growParmsIfNeeded ();
    }
}

/***************************************************************************
 * 1. get the printer list
 * 2. for each printer in the list call Displayq
 ***************************************************************************/

static void 
Display_all (void) {
    char **list, **printer_list = NULL;
    int number_of_printers = 0, i = 0;

    for (list = All_printers (); *list; ++list) {
	number_of_printers++;
    }
    /* getting memeory for the 2-dim array printer_list */
    printer_list = (char **) malloc ((size_t) number_of_printers * sizeof (char *));
    if (printer_list == NULL) {
	logerr_die (XLOG_INFO, "lpq: Malloc for **printer_list failed");
	}
    for (list = All_printers (); *list; ++list, i++) {
	printer_list[i] = (char *) malloc ((size_t) 22 * sizeof (char));
	if (printer_list[i] == NULL) {
	    logerr_die (XLOG_INFO, "lpq: Malloc for *printer_list failed");
	    }
	    (void) strncpy (printer_list[i], *list, 21);
	}
    /* Now sort the array of printers */
    qsort ((char **) printer_list, (size_t) number_of_printers,
	   (size_t) sizeof (printer_list), printer_compare);
    for (i = 0; i < number_of_printers; i++) {
	Printer = printer_list[i];
	Displayq (True, Old_long_format);
    }
}

/*
 * Name_all: added by jmason for shell-script writers; name all the printers.
 */

static void 
Name_all (void)
{
    char **list, **printer_list;
    int number_of_printers = 0, i = 0;

    for (list = All_printers (); *list; ++list) {
	number_of_printers++;
    }
    /* getting memeory for the 2-dim array printer_list */
    malloc_or_die (printer_list, number_of_printers);
    for (list = All_printers (); *list; ++list, i++) {
	malloc_or_die (printer_list[i], sizeof( char) * 21 + 1);
	printer_list[i] = *list;
    }
    /* Now sort the array of printers */
    qsort ((char **) printer_list, (size_t) number_of_printers,
	   (size_t) sizeof (printer_list), printer_compare);
    for (i = 0; i < number_of_printers; i++) {
	printf ("%s\n", printer_list[i]);
    }
}

int 
main (int argc, char *argv[], char *envp[])
{
    struct passwd *pw_ent;	/* user entry in /etc/passwd */
    int i;
    char *pgm_name;

    /* #ifdef NON_SETUID
    for (i = 0; i < argc; i++) {
	fprintf (stderr, "argv[%d]: %s\n", i, argv[i]);
    } 
    #endif */

    /* check the program name */
    if ((pgm_name = strrchr (argv[0], '/'))
	|| (pgm_name = strrchr (argv[0], '\\'))
	|| (pgm_name = strrchr (argv[0], ':')))
	pgm_name++;
    else
	pgm_name = argv[0];

    /* if we are called as "lpstat", set the flag to 1 (True) */
    if (strcmp ("lpstat", pgm_name) == 0) {
	sysv_lp = 1;
    } else {
	sysv_lp = 0;
    }

    set_user_uid (0);
    root_to_user ();

    /*
     * Set fd 0, 1, 2 to /dev/null if not open
     */
    Std_environ (argc, argv, envp);
    Setup_waitpid ();
    Readconf ();
    /*
     * get the user name
     */
    if ((pw_ent = getpwuid (UserUID)) == 0) {
	logerr_die (XLOG_INFO, "getpwuid failed on uid %d", UserUID);
    }
    (void) strcpy (LOGNAME, pw_ent->pw_name);
    Person = LOGNAME;

    Get_parms (argc, argv);

    DEBUG3 ("called as %s", pgm_name);

    /* pick up the default printer name */
    if (Printer_count == 0 && !Do_all_pr && !Name_all_pr) {
	Get_Printer (0);
    }

    /* set up the from information */
    FQDN = Host;
    From = ShortHost;

    (void) plp_signal (SIGPIPE, (plp_sigfunc_t)SIG_IGN);
    (void) plp_signal (SIGHUP, cleanup);
    (void) plp_signal (SIGINT, cleanup);
    (void) plp_signal (SIGQUIT, cleanup);
    (void) plp_signal (SIGTERM, cleanup);

    do {
	/*
	 * if we are to clear the terminal, do so
	 */
	if (Delay) {
	    Term_clear ();
	}
	/*
	 * if we have all printers to display, do so
	 */
	if (history) {
	    display_history ();
	    continue;
	}
	if (Do_all_pr) {
	    Display_all ();
	} else if (Name_all_pr) {
	    Name_all ();
	} else if (Printer_count == 0) {
	    Displayq (True, Old_long_format);
	} else {
	    for (i = 0; i < Printer_count; ++i) {
		Printer = Printer_name[i];
		Displayq (True, Old_long_format);
	    }
	}
	if (Delay) {
	    sleep ((unsigned) Delay);
	}
    } while (Delay);
    return False;
}


/* here are all the functions needed for displaying the history 
 * they are "stolen" from above (thanks Justin!)
 */

void
display_history (void)
{
    int i;

    if (Do_all_pr) {
	history_all ();
    } else if (Name_all_pr) {
	Name_all ();
    } else if (Printer_count == 0) {
	show_history ();
    } else {
	for (i = 0; i < Printer_count; ++i) {
		Printer = Printer_name[i];
		show_history ();
	}
    }
    return;
}

void 
history_all (void)
{
    char **list;

    for (list = All_printers (); *list; ++list) {
	Printer = *list;
	show_history ();
    }
}

void
show_history (void)
{
    /* most of this code comes from displayq.c */

    FILE *afp;                  /* attach file pointer */
    char afname[MAXPARMLEN];    /* attach file name */
    short att_mark;             /* Flag if a printer is attached 
				 * to another printer 
				 */
    char history_file[255];
    FILE *history_fp;
    time_t secs;
    char user[32];
    char machine[32];
    char buf[128];
    int seq_number;
    char priority[2];
    char jobname[22];
    int result;
    int perms = 'C';
    int permission = 0;
    int head_not_printed = 1;

    extern void Remote_history (char *);

    att_mark = False;


reattach:                       /* restart from here if printer is attached to 
				   another */
    /*
     * get the printcap entry
     */
    if (Get_pc_entry (Printer, Status_pc_vars, Status_pc_len) == 0) {
	(void) printf ("Printer '%s' does not exist!\n", Printer);
	(void) fflush (stdout);
	if (ThisIsLPD) {		/* we're the daemon -- log this error */
	    log (XLOG_WARNING,
		 "displayq: printer '%s' does not have a printcap entry!", Printer);
	}
	return;
    }

    /* No history for the save queue! */
    if (SA) 
	return;

    if (SD == 0 || *SD == 0) {
        log (XLOG_WARNING, "Printer '%s' does not have a spool directory!",
             Printer);
        return;
    }
    /* check permissions */
    permission = Checkperm (FQDN, Person, First_name, &perms, (int *) 0, 0);

    DEBUG5 ("show_history: doing %s on host %s", Printer, Host);
    /*
     * If you have an NFS spool dir, use the remote lpd.
     */
    if (RM && *RM && NW) {
        Remote_history (history_user);
        return;
    }
    /* chdir to spool directory */
    chdir_SD ();

    /*
     * check to see if attached to another printerq, alter printer if attached
     */
    if (Attach_file && *Attach_file && ((afp = fopen_daemon (Attach_file, "r")))) {
        if (fscanf (afp, "%s", afname) == 1) {
            if (strsame (afname, Printer)) {
                fatal (XLOG_INFO, "Printer '%s' attached to itself", Printer);
            }
            (void) strcpy (unattached_printer, Printer);
            Printer = strdup (afname);
            att_mark = True;
            fclose (afp);
	    
        } else {
            fatal (XLOG_INFO, "attach file for printer '%s' corrupted -- remove by hand!",
		   Printer);
        }
	
        DEBUG2 ("show_history: printer %s reattached to %s",
                unattached_printer, Printer);
        fclose (afp);
        goto reattach;
    }

    /* construct the history file name */
    (void) sprintf (history_file, "%s/.history", SD);
    user_to_daemon ();
    if ((history_fp = fopen (history_file, "r")) == NULL) {
        (void) fprintf (stdout, 
			"No history available for the queue/printer '%s'.\n",
			Printer);
        (void) fflush (stdout);
        daemon_to_user ();
    } else {
        while (fscanf (history_fp, "%ld", &secs) != EOF) {
            (void) fscanf (history_fp, "%s", user);
            (void) fscanf (history_fp, "%s", machine);
            (void) fscanf (history_fp, "%d", &seq_number);
            (void) fscanf (history_fp, "%s", priority);
	    (void) fgets (buf, 128, history_fp);
	    /* strip the "/n" */
	    buf[strlen (buf)-1] = '\0';
	    result = (int) atoi (&buf[strlen (buf)-2]);
	    buf[strlen (buf)-2] = '\0';
	    /* all data are read in; now evaluate it */
            DEBUG4 ("C_history: checking history for %s and jobs of %s.", 
                    user, Person);

	    /* Checking permissions; a user with control permissions is able to
	     * see the whole history file 
	     */
	    if (permission) {
		if (all_history) {
		    ;
		} else {
		    if (strlen (history_user) > 0) {
			if (strcmp (user, history_user) == 0) {
			    ;
			} else {
			    continue;
			}
		    } else {
			if (strcmp (user, Person) == 0) {
			    ;
			} else {
			    continue;
			}
		    }
		}
	    } else {
		if (strcmp (user, Person) == 0) {
		    ;
		} else {
		    continue;
		}
	    }
	    if (head_not_printed) {
		Print_queuehead_hist ();
		head_not_printed = 0;
	    }
	    (void) strncpy (jobname, buf, 21);
	    jobname[21] = '\0';
	    user[9] = '\0';
            (void) fprintf (stdout,"%-15s %-9s %-16.*s %3d %s %-22s ",
                            Time_str_hist (secs),
                            user, 
                            16, machine, 
                            seq_number, 
                            priority, 
                            jobname); 
	    switch (result) {
	    case -1:
		(void) fprintf (stdout, " SENT\n");
		break;
	    case 0:
		(void) fprintf (stdout, "ERROR\n");
		break;
	    case 1:
		(void) fprintf (stdout, "  OK\n");
		break;
	    case 2:
	    case 3:
		(void) fprintf (stdout, "ERROR\n");
		break;
	    case 4:
		(void) fprintf (stdout, " SENT\n");
		break;
	    case 5:
		(void) fprintf (stdout, "DELETED\n");
		break;
	    case 6:
		(void) fprintf (stdout, " MOVED\n");
		break;
	    default:
		(void) fprintf (stdout, "UNKOWN\n");
		break;
	    }
            (void) fflush (stdout);
        }
        fclose (history_fp);
	daemon_to_user ();
    }
    if (RM && *RM) {
	DEBUG4 ("show_history: getting Remote_history():");
	if (all_history) {
	    (void) strcpy (history_user, "all");
	}
	Remote_history (history_user);
	if (all_history) {
	    history_user[0] ='\0';
	}
    }
    all_history = 0;
    return;
}

#define TWENTY_FOUR_HOURS       (24 * 60 * 60)

static void
no_printer (char *pr) {
    (void) fprintf (stdout, "Printer '%s' does not exist!\n", pr);
    (void) fflush (stdout);
    if (ThisIsLPD) {		/* we're the daemon -- log this error */
	log (XLOG_WARNING,
		"displayq: printer '%s' does not have a printcap entry!", pr);
    }
    return;
}

void
check_lpd (void) {
    pid_t f;

    (void) Checklockfile (Masterlock, &f, (char *) 0, 0, &LO_statb);
    (void) fprintf (stdout, "scheduler is%srunning\n", f ? " " : " not ");
    return;
}

void
show_default_destination (void) {
    Get_Printer (0);
    (void) fprintf (stdout, "system default destination: %s\n", 
		    Printer);
    (void) fflush (stdout);
    return;
}

void
show_lpstat_a_option (char *Optarg) {
    char **list;

    if ( (!Optarg) || (strcmp (Optarg, "all") == 0) ) {
	for (list = All_printers (); *list; ++list) {
	    Printer = *list;
	    Get_pc_entry (Printer, Status_pc_vars, Status_pc_len);
	    chdir_SD ();
	    (void) fprintf (stdout, "%s", Printer);
	    (void) Checkactive ();
	    if (LO_statb.st_mode & (DISABLE_QUEUE)) {
		(void) fprintf (stdout, " not" );
	    }
	    (void) fprintf (stdout, " accepting requests ");
	    (void) fflush (stdout);
	    printstatus (0);
	}
    } else {
	char buf[64];
	int length_opt, length = 0, length_buf = 0;
			
	length_opt = strlen (Optarg);
	if (length_opt > 63) {
	    errno = E2BIG;
	    logerr_die (XLOG_INFO, "");
	    exit (1);
	}
	while (length <= length_opt) {
	    Optarg += length_buf;
	    (void) sscanf (Optarg, "%s", buf);
	    if (Get_pc_entry (buf, Status_pc_vars, Status_pc_len) == 0) {
		no_printer (buf);
	    } else {
		Printer = buf;
		chdir_SD ();
		(void) Checkactive ();
		(void) fprintf (stdout, "%s", Printer);
		if (LO_statb.st_mode & (DISABLE_QUEUE)) {
		    (void) fprintf (stdout, " not" );
		}
		(void) fprintf (stdout, " accepting requests ");
		printstatus (0);
	    }
	    length += strlen (buf) + 1;
	    length_buf = strlen (buf) + 1;
	}
	(void) fflush (stdout);
    }
    return;
}

void
show_lpstat_o_option (char *Optarg) {
    char **list;

    if ( (!Optarg) || (strcmp (Optarg, "all") == 0) ) {
	for (list = All_printers (); *list; ++list) {
	    Printer = *list;
	    display_lp_printer ();
	}
    } else {
	char buf[64];
	int length_opt, length = 0, length_buf = 0;
	
	length_opt = strlen (Optarg);
	if (length_opt > 63) {
	    errno = E2BIG;
	    logerr_die (XLOG_INFO, "");
	    exit (1);
	}
	while (length <= length_opt) {
	    Optarg += length_buf;
	    (void) sscanf (Optarg, "%s", buf);
	    Printer = buf;
	    display_lp_printer ();
	    length += strlen (buf) + 1;
	    length_buf = strlen (buf) + 1;
	}
    }
    return;
}

void
show_lpstat_p_option (char *Optarg) {
    char **list;

    if ( (!Optarg) || (strcmp (Optarg, "all") == 0) ) {
	for (list = All_printers (); *list; ++list) {
	    Printer = *list;
	    display_lpstatus ();
	}
    } else {
	char buf[64];
	int length_opt, length = 0, length_buf = 0;
	
	length_opt = strlen (Optarg);
	if (length_opt > 63) {
	    errno = E2BIG;
	    logerr_die (XLOG_INFO, "");
	    exit (1);
	}
	while (length <= length_opt) {
	    Optarg += length_buf;
	    (void) sscanf (Optarg, "%s", buf);
	    Printer = buf;
	    display_lpstatus ();
	    length += strlen (buf) + 1;
	    length_buf = strlen (buf) + 1;
	}
    }
    return;
}

void
show_lpstat_u_option (char *Optarg) {
    char **list;
    char user[9];
    char *user_buf;
    char buf[64];
    int length_opt, length = 0, length_buf = 0;

    if ( (!Optarg) || (strcmp (Optarg, "all") == 0) ) {
	for (list = All_printers (); *list; ++list) {
	    Printer = *list;
	    display_lp_printer ();
	}
    } else {
	length_opt = strlen (Optarg);
	if (length_opt > 63) {
	    errno = E2BIG;
	    logerr_die (XLOG_INFO, "");
	    exit (1);
	}
	malloc_or_die (user_buf, strlen (Optarg)* sizeof (char) + 1);
	(void) sprintf (user_buf, Optarg);
    
	length = 0;
	while (length <= length_opt) {
	    user_buf += length_buf;
	    (void) sscanf (user_buf, "%s", buf);
	    (void) strncpy (user, buf, 8);
	    for (list = All_printers (); *list; ++list) {
		Printer = *list;
		if (!user_buf) {
		    display_lp_printer ();
		} else {
		    display_lp_printer_user (user);
		}
	    }
	    length_opt = strlen (user_buf);
	    length += strlen (buf) + 1;
	    length_buf = strlen (buf) + 1;
	}
    }
    return;
}

void
show_lpstat_v_option (char *Optarg) {
    char **list;

    if ( (!Optarg) || (strcmp (Optarg, "all") == 0) ) {
	for (list = All_printers (); *list; ++list) {
	    Printer = *list;
	    if (Get_pc_entry (Printer, Status_pc_vars, 
			      Status_pc_len) == 0) {
		no_printer (Printer);
	    }
	    if (RP && RM) { /* remote: show only the rm entry */
		(void) fprintf (stdout, "system for %s: %s\n", 
			 Printer, RM);
	    } else {
		if (SV) { /* special case: we have servers */
		    (void) fprintf (stdout, "device for %s: %s\n", 
			     Printer, SV);
		}
		else { /* shows the associated device */
		    (void) fprintf (stdout, "device for %s: %s\n", 
			     Printer, LP ? LP : "/dev/null");
		}
	    }
	}
    } else {
	/* we have to separate the arguments :-( */
	char buf[64];
	int length_opt, length = 0, length_buf = 0;
	
	length_opt = strlen (Optarg);
	if (length_opt > 63) {
	    errno = E2BIG;
	    logerr_die (XLOG_INFO, "");
	    exit (1);
	}
	while (length <= length_opt) {
	    Optarg += length_buf;
	    (void) sscanf (Optarg, "%s", buf);
	    if (Get_pc_entry (buf, Status_pc_vars, 
			      Status_pc_len) == 0) {
		no_printer (buf);
	    } else {
		if (RP && RM) {
		    (void) fprintf (stdout, "system for %s: %s\n", 
				    buf, RM);
		} else {
		    (void) fprintf (stdout, "device for %s: %s\n", 
				    buf, LP ? LP: "/dev/null");
		}
	    }
	    length += strlen (buf) + 1;
	    length_buf = strlen (buf) + 1;
	}
    }
    (void) fflush (stdout);
    return;
}

void
display_lpstatus (void) {
    int i, server = 0;
    char buf[BUFSIZ];
    struct plp_queue *q_lp;
    int perms = 'R';		  /* file perms */

    if (Get_pc_entry (Printer, Status_pc_vars, Status_pc_len) == 0) {
	no_printer (Printer);
    } else {
	chdir_SD ();
	(void) fprintf (stdout, "printer %s ", Printer);
	Jobcount = Getq ();
	server = Checkactive ();
	if (Jobcount == 0) {
	    (void) fprintf (stdout, "is idle. ");
	} else if (Jobcount > 0 && server == 0) {
	    q_lp = &Queue[0];
	    (void) fprintf (stdout, "faulted printing %s-%03d. ", Printer, q_lp->q_num);
	} else {
	    for (i = 0; i < Jobcount; i++) {
		q_lp = &Queue[i];
		if (q_lp->q_daemon) {
		    (void) fprintf (stdout, "now printing %s-%03d. ", 
				    Printer, q_lp->q_num);
		    break;
		}
	    } 
	}
	if (LO_statb.st_mode & (DISABLE_PRINT)) {
	    (void) fprintf (stdout, "disabled " );
	} else {
	    (void) fprintf (stdout, "enabled ");
	}
	printstatus (1);
	
	(void) fflush (stdout);
    }
    /* -D option set in command line */
    if (d_found) {
	(void) fprintf (stdout, "        Description: ");
	if (CM && *CM) {
	    Print_comment (CM);
	} 
	(void) fprintf (stdout, "\n");
    } else /* -l option set in command line */
	if (l_found) {
	    /* Forms mounted valid for local printers only */
	    if (!RP && !RM) {
		char *forms_mounted;
		/* Forms mounted */
		(void) fprintf (stdout, "        Form mounted: ");
		buf[0] ='\0';
		forms_mounted = current_options ();
		if (forms_mounted) {
		    if (!strsame (forms_mounted, DEF_OPTIONS)) {
			(void) fprintf (stdout, "%s", forms_mounted);
		    }
		} else {
		    (void) fprintf (stdout, "none");
		}
		(void) fprintf (stdout, "\n");
	    }
	    
	    /* content types; "any" will be printed (it's only an emulation... 8:) */
	    (void) fprintf (stdout, "        Content types: any\n");
	    
	    /* Printer types: we'll read the pt entry from the printcap */
	    (void) fprintf (stdout, "        Printer types: ");
	    if (PT && *PT) {
		(void) fprintf (stdout, "%-s\n", PT);
	    } else {
		(void) fprintf (stdout, "unknown\n");
	    }

	    /* description of the printer */
	    (void) fprintf (stdout, "        Description: ");
	    if (DS && *DS) {
		(void) fprintf (stdout, "%-s\n", DS);
	    } else {
		(void) fprintf (stdout, "\n");
	    }

	    /* connection, interface, after fault; 
	     * this is only for local printers 
	     */
	    if (LP && *LP) {
		(void) fprintf (stdout, "        Connection: direct\n");
		(void) fprintf (stdout, "        Interface: %-s\n", LP);
		(void) fprintf (stdout, "        After fault: ");
		if (RT > 0) {
		    (void) fprintf (stdout, "after %d retries stop\n", RT);
		} else {
		    (void) fprintf (stdout, "stop\n");
		}	    
	    }

	    /* users allowed; we'll get the entry from printer_perms */
	    (void) fprintf (stdout, "        Users allowed:\n");
	    if (do_checkperms_lpstat (FROMHOST, LOGNAME, First_name, &perms) <= 0) {
		(void) fprintf (stdout, "(all)\n");
	    }

	    /* Forms allowed */
	    (void) fprintf (stdout, "        Forms allowed:\n");
	    if (FL && *FL) {
		(void) fprintf (stdout, "        %22s\n", FL);
	    } else {
		(void) fprintf (stdout, "                (none)\n");
	    }
    
	}
    
    (void) fflush (stdout);
    return;
}

void
display_lp_printer (void) {
    int server = 0;
    int i;
    char buf[128], buf_tmp[20];
    struct plp_queue *q_lpp;

    if (Get_pc_entry (Printer, Status_pc_vars, Status_pc_len) == 0) {
	no_printer (Printer);
	return;
    } else {
	chdir_SD ();
	Jobcount = Getq ();
	server = Checkactive ();
	for (i = 0; i < Jobcount; i++) {
	    q_lpp = &Queue[i];
	    (void) sprintf (buf_tmp, "%s-%03d", Printer, q_lpp->q_num);
	    (void) fprintf (stdout, "%-s %*.*s %*s %9ld %s", 
			    buf_tmp,
			    30 + (int) strlen (q_lpp->q_user) - (int) strlen (buf_tmp), 
			    (int) strlen (q_lpp->q_user),
			    q_lpp->q_user, 
			    (int) 10 - (int) strlen (q_lpp->q_user), " ",
			    q_lpp->q_size, 
			    short_time_lpstat (q_lpp->q_time));
	    if (q_lpp->q_daemon) {
		if (RP && RM) {
		    (void) sprintf (buf, "%s@%s", RP, RM);
		    (void) fprintf (stdout, " on %s\n", buf);
		} else {
		    (void) fprintf (stdout, " on %s\n", Printer);
		}
	    } else {
		(void) fprintf (stdout, "\n");
	    }
	}
    }
    (void) fflush (stdout);
    return;
}

void
display_lp_printer_user (char *user) {
    int server = 0;
    int i;
    char buf[128], buf_tmp[20];
    struct plp_queue *q_lpu;
    if (Get_pc_entry (Printer, Status_pc_vars, Status_pc_len) == 0) {
	no_printer (Printer);
	return;
    } else {
	chdir_SD ();
	Jobcount = Getq ();
	server = Checkactive ();
	for (i = 0; i < Jobcount; i++) {
	    q_lpu = &Queue[i];
	    if (strcmp (user, q_lpu->q_user) == 0) {
		(void) sprintf (buf_tmp, "%s-%03d", Printer, q_lpu->q_num);
		(void) fprintf (stdout, "%-s %*.*s %*s %9ld %s", 
				buf_tmp,
				30 + strlen (q_lpu->q_user) - strlen (buf_tmp),
				(int) strlen (q_lpu->q_user),
				q_lpu->q_user, 
				(int) 10 - strlen (q_lpu->q_user), " ",
				q_lpu->q_size, 
				short_time_lpstat (q_lpu->q_time));
		if (q_lpu->q_daemon) {
		    if (RP && RM) {
			(void) sprintf (buf, "%s@%s", RP, RM);
			(void) fprintf (stdout, " on %s\n", buf);
		    } else {
			(void) fprintf (stdout, " on %s\n", Printer);
		    }
		} else {
		    (void) fprintf (stdout, "\n");
		}
	    }
	}
    }
    (void) fflush (stdout);
    return;
}

/* This part comes form displayq.c
 * short_time includes a mod from Ed Santiago to cut down the date length
 * even more so lpq output fits into 80 columns again.
 */
static char *
short_time_lpstat (time_t tvec) {
    static char s[40];

    /*
     * 012345678901234567890123456789
     * Mon May 30 20:53:27 CDT 1988
     */
    if (tvec == (time_t)0) 
	return ("-");

    (void) strcpy (s, ctime (&tvec));
    s[16] = '\0';
    return (&s[4]);
}
/* short_time includes a mod from Ed Santiago to cut down the date length
 * even more so lpq output fits into 80 columns again. It comes from displayq.c
 */

char *
short_time_for_lpq (time_t tvec) {
    static char s[40];

    /*
     * 012345678901234567890123456789
     * Mon May 30 20:53:27 CDT 1988
     */
    if (tvec == (time_t)0) return ("-");
    
    (void) strcpy (s, ctime (&tvec));
    s[16] = '\0';
    if (time((time_t *) NULL) - tvec < TWENTY_FOUR_HOURS) {
        return (&s[11]);
    } else {
        s[10] = '\0';
        return (&s[4]);
    }
}

int
do_checkperms_lpstat (char *host, char *user, char *queue, int *perms) {
    int result;
    struct plp_queue *q_for_msg;

    result = Checkperm (host, user, queue, perms, (int *) NULL, 1);

    q_for_msg = Queue;
    if (result == -2) {		/* -2 means "out of pages" */
	return result;
    } else if (result <= 0) {
	return result;
    } else {
	return 1;
    }
}

int
printer_compare (char **l, char **r)
{
    return (strcmp (*l, *r));
}
