/***************************************************************************
 * U. Minnesota LPD Software * Copyright 1987, 1988, Patrick Powell
 * version 3.3.0 Justin Mason July 1994
 * version 4.1.2 Stefano Ianigro June 1998
 *
 * Changelog:
 * 16.08.1994: Added functions: downq, changepriority (Stefano Ianigro)
 *             Changed the ouptut format; it distinguishes between queues
 *             and printers
 *
 * 04.09.1994: Changed the output format so that the contents of the queue
 *             and the printer status will be displayed 
 *
 * 24.01.1995: now uses False, True; also, changed names of static
 *             flag vars j, m to more descriptive names. -- jm.
 *
 * 22.08.1995: changed the stop/disable function so that a message can be 
 *             issued; this message must begin with a "#" and must came after the 
 *             last printer name. If there is no message, setstatus will be
 *             called with a standard message. ---st.
 *
 * 27.11.1995: Introduced the new function "copy"; it will be used for copying
 *             jobs from one spool directory to another one. ---st.
 *
 * 04.10.1996: flag for the use with the Tk-Lpc program introduced. ---st.
 *
 * 05.06.1997: implemented the function "print_one_job". ---st.
 *
 * 22.06.1997: implemented the history functions. ---st.
 * 
 * 24.11.1997: To make it useable with CDE on Linux machines, the status 
 *             command will display a short status of a printer. The 
 *             long form has now the command lstatus. ---st.

 ***************************************************************************
 * MODULE: Control_ops.c
 */

#include "lp.h"
#include "library/errormsg.h"
#include "library/utils.h"
#include "common/control_ops.h"
#include "common/displayq.h"
#include "common/setstatus.h"
#include "common/requeue.h"
#include "common/hold.h"
#include "common/move.h"
#include "common/copy.h"
#ifdef HAVE_UTIME_H
#include <utime.h>
#endif
#ifdef HAVE_SYS_TERMIOS_H
#include <termios.h>
#include <sys/ioctl.h>
#endif

/***************************************************************************
 *  1. get the printcap entries
 *  2. check for permissions
 *  3. determine the function to be carried out
 *  4. carry out the function
 ***************************************************************************/
/*
 * dispatch is a data structure that contains the names and flags of control commands.
 */
struct dispatch {
    char *name;         /* command */
    int distinct;       /* how many of the characters are neccessary to distinct */
    int (*routine) ();  /* name of the routine */
    int flags;          /* attributes of the routine */
    char *summary;      /* help text */
};

static int
#ifdef C_BAD
C_bad (),			/* not used normally */
#endif
    C_forms (), C_help (), C_abort (), C_clean (), C_copy (), C_disable (),
    C_enable (), C_exit (), C_kill (), C_restart (), C_lpq (), C_lprm (), C_lstatus (),
    C_start (), C_status (), C_stop (), C_topq (), C_remote (), C_lpd (),
    C_down (), C_up (), C_attach (), C_unattach (), C_preprint (), C_print_one (),
    C_hold (), C_release (), C_move (), C_downq (), C_chprio (), C_requeue (),
    C_history (), C_listforms ();

void default_message_stop (), default_message_down (), default_message_disable ();
extern void move (void), copy (void), print_one (int), release (void);
extern int printer_comp (), check_if_printer_is_attached (void);
extern void Print_queuehead_hist (void);
extern int Match_entry (struct plp_queue *); 

#define NEED_OPT		1	/* we check for other options */
#define ALL_ALLOWED		2	/* check for all */
#define NEED_PRIV		4	/* privileged operation */
#define ALL_DEF			8	/* if no parameter, all is default */
#define NO_PR_LIST		0x10	/* printer list */
#define REMOTE_OP		0x20	/* remote allowed */
#define SERVER_OP		0x40	/* server allowed */
#define ON_REMOTE		0x80	/* not if you have RM and NW */
#define IS_OPT(x) ((x)&NEED_OPT)
#define IS_ALL(x) ((x)&ALL_ALLOWED)
#define IS_PRIV(x) ((x)&NEED_PRIV)
#define IS_ALL_DEF(x) ((x)&ALL_DEF)
#define IS_NO_PR_LIST(x) ((x)&NO_PR_LIST)
#define IS_REMOTE(x) ((x)&REMOTE_OP)
#define IS_SERVER(x) ((x)&SERVER_OP)
#define IS_ON_REMOTE(x) ((x)&ON_REMOTE)

static struct dispatch dispatch[] = 
{
    {"", 0, 0, 0, "not a command"},
    {"?", 1, C_help, 0, "? is same as 'help'"},
    {"abort", 2, C_abort,
     NEED_OPT | NEED_PRIV | ALL_ALLOWED | REMOTE_OP | SERVER_OP | ON_REMOTE,
     "abort ( all | Printer ):                     kill off server and disable                                                     printing"},
    {"attach", 2, C_attach, NEED_OPT | NO_PR_LIST | REMOTE_OP | ON_REMOTE | NEED_PRIV,
     "attach Printer destPrinter:                  attach queue to another printer"},
    {"chprio", 2, C_chprio, NEED_PRIV | NEED_OPT | REMOTE_OP | ON_REMOTE | NO_PR_LIST,
     "chprio Printer (user|host|job|*) priority:   change priority of a job"},
    {"clean", 2, C_clean, NEED_OPT | NEED_PRIV | ALL_ALLOWED | REMOTE_OP | ON_REMOTE,
     "clean ( all | Printer ):                     remove all crud from queue \n                                             (dangerous!)"},
    {"copy", 2, C_copy, NEED_OPT | NO_PR_LIST | ALL_ALLOWED | REMOTE_OP | ON_REMOTE,
     "copy Printer (user|host|job|*) destPrinter:  copy specified job(s) to another \n                                             printer"}, 
    {"cp", 2, C_copy, NEED_OPT | NO_PR_LIST | ALL_ALLOWED | REMOTE_OP | ON_REMOTE,
     "cp Printer (user|host|job|*) destPrinter:    same as copy"}, 
    {"disable", 2, C_disable, NEED_OPT | NEED_PRIV | ALL_ALLOWED | REMOTE_OP | ON_REMOTE,
     "disable ( all | Printer ) [#message]:        disable queueing"},
    {"down", 2, C_down, NEED_OPT | NEED_PRIV | ALL_ALLOWED | REMOTE_OP | SERVER_OP | ON_REMOTE,
     "down ( all | Printer ) [#message]:           disable queueing and printing"},
    {"downq", 1, C_downq, NEED_OPT | NO_PR_LIST | REMOTE_OP | ON_REMOTE | NEED_PRIV,
     "downq Printer (user|host|job|*):             move job to the end of queue"},
    {"enable", 2, C_enable, NEED_OPT | NEED_PRIV | ALL_ALLOWED | ON_REMOTE | REMOTE_OP,
     "enable ( all | Printer ):                    enable queuing"},
    {"exit", 2, C_exit, 0, "exit:                                        terminate lpc"},
    {"forms", 1, C_forms, NEED_OPT | NO_PR_LIST | REMOTE_OP | ON_REMOTE | NEED_PRIV,
     "forms Printer forms:                         change loaded forms"},
    {"help", 2, C_help, 0, 
     "help [all] [command]:                        print command summary"},
    {"history", 2, C_history, NEED_OPT | ALL_ALLOWED | REMOTE_OP | ON_REMOTE,
     "history [user|all] ( all | Printer ):        print history of a queue/all queues"},      
    {"hold", 2, C_hold, NEED_OPT | NO_PR_LIST | ALL_ALLOWED | REMOTE_OP | ON_REMOTE | NEED_PRIV,
     "hold Printer (user|host|job|*):              hold specified job(s)"},
    {"kill", 1, C_kill,
     NEED_OPT | NEED_PRIV | ALL_ALLOWED | REMOTE_OP | SERVER_OP | ON_REMOTE,
     "kill ( all | Printer ):                      kill off server and then \n                                             restart printing"},
    {"listforms", 2, C_listforms, ALL_DEF | ALL_ALLOWED | REMOTE_OP | ON_REMOTE | NEED_OPT,
     "listforms ( Printer | all )                  list loaded forms"},
    {"lpd", 3, C_lpd, 0,
     "lpd:                                         check out lpd process"},
    {"lpq", 3, C_lpq, 0,
     "lpq ( parms ):                               call lpq"},
    {"lprm", 3, C_lprm, 0,
     "lprm ( parms ):                              call lprm"},
    {"lstatus", 2, C_lstatus, ALL_DEF | ALL_ALLOWED | NEED_OPT,
     "lstatus [ terse | all ] [ Printer ]:         print status (long format)"},
    {"move", 2, C_move, NEED_OPT | NO_PR_LIST | ALL_ALLOWED | REMOTE_OP | ON_REMOTE,
     "move Printer (user|host|job|*) destPrinter:  move specified job to another \n                                             printer"},
    {"mv", 2, C_move, NEED_OPT | NO_PR_LIST | ALL_ALLOWED | REMOTE_OP | ON_REMOTE,
     "mv Printer (user|host|job|*):                same as move"},
    {"one", 1, C_print_one, NEED_OPT | NO_PR_LIST | NEED_PRIV | REMOTE_OP | ON_REMOTE, 
     "one Printer jobnumber:                       prints only this job"},
    {"preprint", 2, C_preprint, NEED_PRIV | NEED_OPT | NO_PR_LIST,
     "preprint Printer jobnumber                   pre-print selected pages of a job"},
    {"quit", 1, C_exit, 0, 
     "quit (same as exit):                         terminate lpc"},
    {"remote", 3, C_remote, 0,
     "remote command:                              do the action on remote printer"},
    {"release", 3, C_release, NEED_OPT | NO_PR_LIST | REMOTE_OP | ON_REMOTE | NEED_PRIV,
     "release Printer (user|host|job|*):           release specified job(s)"},
    {"requeue", 3, C_requeue, NEED_OPT | NO_PR_LIST | ON_REMOTE | NEED_PRIV,
     "requeue Printer jobnumber:                   requeue specified job to the same \n                                             printer"},
    {"rq", 2, C_requeue, NEED_OPT | NO_PR_LIST | ON_REMOTE | NEED_PRIV,
     "rq Printer jobnumber:                        same as requeue"},
    {"restart", 3, C_restart, NEED_OPT | ALL_ALLOWED | REMOTE_OP | SERVER_OP,
     "restart ( all | Printer ):                   start a server"},
    {"start", 4, C_start, NEED_OPT | NEED_PRIV | ALL_ALLOWED | REMOTE_OP | SERVER_OP | ON_REMOTE,
     "start ( all | Printer ):                     enable printing and start server"},
    {"status", 4, C_status, ALL_DEF | ALL_ALLOWED | NEED_OPT,
     "status [ all ] [ Printer ]:                  print status (short format)"},
    {"stop", 3, C_stop, NEED_OPT | NEED_PRIV | ALL_ALLOWED | REMOTE_OP | SERVER_OP | ON_REMOTE,
     "stop ( all | Printer ) [#message]:           disable further printing"},
    {"topq", 1, C_topq, NEED_OPT | NO_PR_LIST | REMOTE_OP | ON_REMOTE | NEED_PRIV,
     "topq Printer (user|host|job|*):              move job(s) to the top of queue"},
    {"unattach", 2, C_unattach, NEED_OPT | NO_PR_LIST | REMOTE_OP | ON_REMOTE | NEED_PRIV,
     "unattach Printer:                            revert queue original printer"},
    {"up", 2, C_up, NEED_OPT | NEED_PRIV | ALL_ALLOWED | REMOTE_OP | SERVER_OP | ON_REMOTE,
     "up ( all | Printer ):                        enable queuing and printing"},
};

static int ncmds = sizeof (dispatch) / sizeof (struct dispatch);

static void Show_ops (void), Print_sum (int), touch (char *);

static int find_command (char *), service (struct dispatch *),
    remote_cando (int), preprint (void),
    demote (char *, char *), changepriority (char *, char *),
    promote (char *, char *);

static int op_init;			  /* flag for operation initialization */
static int q_on_startstop = True;	  /* displayq on start,stop,disable,enable */
static int q_on_abortpromote = True;      /* displayq on abort,promote,topq */
static int number_of_printers = 0;        /* number of printers in the printcap file */
static int Is_all;			  /* doing for all Printers */
static int Is_terse;                      /* terse flag */
static int called_from_unattach = 0;      /* flag for routine unattach */
static int all_history = 0;               /* flag if all the history is wanted*/
char command_name[16];
char history_user[10];
static char p_tmp[22];

static char **lpc_printer_list = NULL;
static unsigned int line_counter;

/***************************************************************************
 * This routine is called with the parameters in the Parms[] array.
 * It will check to see if the command is valid,  and if so, will
 * do any other actions.
 ***************************************************************************/
int
Control_ops (void) {
    int i;                      /* SI and indices Ltd. */
    int command;		/* command we are going to do */
    char **list;		/* list of Printers */
    struct winsize w;
    static unsigned short rows;
    char dummys[2];
    int printer_found = False;

    /*
     * set the flags needed
     */
    Is_local = (hostcmp (From, Host) == 0);
    Is_root = strsame (Person, "root");
    Is_all = False;
    Is_terse = False;
    op_init = False;
    command = 0;
    
    /*
     * check to see that we have a valid command
     */
    if (Debug > 4) 
	Show_ops ();

    if (Parmcount < 1) {
	(void) printf ("No command.\n");
	return False;
    }
    command = find_command (Parms[0].str);
    if (command == 0) {
	(void) printf ("Unknown command \"%s\".\n", Parms[0].str);
	Parmcount = 0;
	return False;
    }

    strncpy (command_name, Parms[0].str, sizeof (command_name) -1);
    command_name[strlen(command_name)+1] = '\0';

    DEBUG3 ("Control_ops: command %s, flags 0x%x, %s",
	    dispatch[command].name, dispatch[command].flags,
	    dispatch[command].summary);

    if (Log_LPCs) {
	if (Printer && *Printer) {
	    log_request ("%s@%s: control operation '%s %s'", Person, From,
			 dispatch[command].name, Printer);
	} else {
	    log_request ("%s@%s: control operation '%s'", Person, From,
			 dispatch[command].name);
	}
    }
    if (!Is_local && !IS_REMOTE (dispatch[command].flags)) {
	log (XLOG_INFO, "Command %s cannot be done remotely", Parms[0].str);
	Parmcount = 0;
	return False;
    }
    /*
     * call the appropriate routine
     */
    Shift_parms (1);
    /*
     * If no option processing just dispatch
     */
    if (!IS_OPT (dispatch[command].flags)) {
	return ((*dispatch[command].routine) (&dispatch[command]));
    }
    /*
     * if no options and IS_ALL is the default, set up the all flag
     */
    if (Parmcount == 0) {
	if (IS_ALL_DEF (dispatch[command].flags)) {
	    Is_all = True;
	} else {
	    (void) printf ("No parameters. Usage:\n  %s.\n",
			   dispatch[command].summary);
	    return False;
	}
    } else if (strsame (Parms[0].str, "all")) {
	if (strsame (dispatch[command].name, "history")) {
	    (void) strcpy (history_user, Parms[0].str);
	    Is_all = False;
	    all_history = 1;
	    Shift_parms (1);	/* AJCD */
	    if (Parmcount == 0) { /* all with no printers given */
		(void) printf ("Missing parameter for printer. Usage:\n  %s.\n",
			       dispatch[command].summary);
		return False;
	    }
	    if (strsame (Parms[0].str, "all")) {
		Is_all = True;
	    } 
	} else {	
	    if (IS_ALL (dispatch[command].flags)) {
		Is_all = True;
		Shift_parms (1);	/* AJCD */
	    } else {
		(void) printf ("'all' not allowed. Usage:\n  %s.\n",
			       dispatch[command].summary);
		return False;
	    }
	}
    } else if (strsame (Parms[0].str, "terse")) {
	if (strsame (dispatch[command].name, "history")) {
	    (void) printf ("terse not allowed with %s command.\n  %s.\n",
			   dispatch[command].name, dispatch[command].summary);
	    return False;
	}
        if (IS_ALL (dispatch[command].flags)) {
	    Is_terse = True;
	    Is_all = True;
	    Shift_parms (1);	/* AJCD */
	} else {
	    (void) printf ("'all' not allowed. Usage:\n  %s.\n",
			   dispatch[command].summary);
	    return False;
	}
    }
    
    /*
     * set up Printer from parameters
     */
    if (Is_all) {
	DEBUG2 ("Control_ops: 'all' parameter");
	if (Is_terse) {
	    DEBUG2 ("Control_ops: 'terse' parameter");
	}

#ifdef HAVE_SYS_TERMIOS_H
        if (ioctl(2, TIOCGWINSZ, &w) == 0 && w.ws_row > 0) {
	    rows = w.ws_row;
        } else {
	    rows = 24;
        }
#else
        rows = 24;
#endif
        line_counter = 3;
	/* Get the number of printers */
   	number_of_printers = 0;
	for (list = All_printers (); *list; ++list) {
	    number_of_printers++;
	}

	/* getting memeory for the 2-dim array lpc_printer_list */
	lpc_printer_list = (char **) malloc ((size_t) number_of_printers * sizeof (char *));
	if (lpc_printer_list == NULL) {
	    logerr_die (XLOG_INFO, "Control_ops: Malloc for **lpc_printer_list failed");
	}
	i = 0;
	for (list = All_printers (); *list; ++list, i++) {
	    lpc_printer_list[i] = (char *) malloc ((size_t) 22 * sizeof (char));
	    if (lpc_printer_list[i] == NULL) {
		logerr_die (XLOG_INFO, "Control_ops: Malloc for *lpc_printer_list failed");
	    }
	    (void) strncpy (lpc_printer_list[i], *list, 21);
	}
	/* Now sort the array of printers */
	qsort ((char **) lpc_printer_list, (size_t) number_of_printers,
	       (size_t) sizeof (lpc_printer_list[0]), printer_comp);
	for (i = 0; i < number_of_printers; i++) {
	    (void) strcpy (p_tmp, lpc_printer_list[i]);
	    Printer = p_tmp;
	    if (service (&dispatch[command]) == 0) {
		for (i = 0; i < number_of_printers; i++) {
		    free (lpc_printer_list[i]);
		}
		free ((char **) lpc_printer_list);
		lpc_printer_list = NULL;
		return False;
	    }
	    /* paging is done now */
	    if (!tk_interface_used) {
                if (line_counter >= rows) {
                    (void) fprintf (stdout, 
				    "Press RETURN to continue or 'b' for back one page or 'q' to quit: ");
		    (void) fflush (stdout);
		    (void) gets (dummys);
                    if (dummys[0] == 'b' || dummys[0] == 'B') {
                        i = i - line_counter;
                        if (i < 0) {
                            i = -1;
                        }
                    } else {
                        if (dummys[0] == 'q' || dummys[0] == 'Q') {
			    goto stop_output;
                        }
                    }
                    line_counter = 3;
                    dummys[0] = ' ';
                    dummys[1] = '\0';
                }
            }
	}
    stop_output:
	for (i = 0; i < number_of_printers; i++) {
	    free (lpc_printer_list[i]);
	}
	free ((char **) lpc_printer_list);
	lpc_printer_list = NULL;
    } else { /* single printer */
	while (Parmcount > 0) {
	    if (strstr (Parms[0].str, "#") != NULL) { /* found message */
	        break;
	    } else {
		if (strsame (dispatch[command].name, "history")) {
		    printer_found = False;
		    for (list = All_printers (); *list; ++list) {
			if (strsame (Parms[0].str, *list)) {
			    /* the second argument is a printer not a username */
			    Printer = Parms[0].str;
			    Shift_parms (1);
			    printer_found = True;
			    break;
			}
		    }
		    if (!printer_found) { /* must be a username */
			if (strlen (history_user) == 0) {
			    (void) strcpy (history_user, Parms[0].str);
			    Shift_parms (1);
			    Printer = Parms[0].str;
			}
			Printer = Parms[0].str;
			Shift_parms (1);
		    }
		} else { 
		    Printer = Parms[0].str;
		    Shift_parms (1);
		}
		if (service (&dispatch[command]) == 0) {
		    return False;
		}
		if (IS_NO_PR_LIST (dispatch[command].flags)) {
		    break;
		}
	    }
	}
    }
    history_user[0] = '\0';
    return True;
}

/***************************************************************************
 * 1. check on the printcap entry
 * 2. chdir to the spool directory
 * 3. dispatch the particular routine
 ***************************************************************************/

static int
service (struct dispatch *cmd) 
{
    int perms = 'C';		/* Permission Checking */

    DEBUG4 ("service: printer '%s', cmd '%s'", Printer, cmd->name);
    if (Get_pc_entry (Printer, Status_pc_vars, Status_pc_len) == 0) {
	printf ("Printer '%s' does not exist.\n", Printer);
	(void) fflush (stdout);
	return False;
    }
    if (RM && NW && IS_ON_REMOTE (cmd->flags)) {
	(void) printf ("%s: Remote printer, use \'remote %s %s\'.\n",
		       Printer, cmd->name, Printer);
	(void) fflush (stdout);
	return False;
    }
    /*
     * we may have a server specified
     */
    if (SS && *SS) {
	if (Get_multi_pc_entry (Printer, SS,
				Status_pc_vars, Status_pc_len) == 0)
	    {
		printf ("Server '%s' queue '%s' does not exist.\n",
			Printer, SS);
		(void) fflush (stdout);
		return False;
	    }
	LO = Printer;
	SV = NULL;
    }
    if (SD == 0 || *SD == 0) {
	log (XLOG_INFO,
	     "service: Printer %s does not have spool directory", Printer);
	return False;
    }
    chdir_SD ();
    /*
     * check on the privileges needed You have to be local and root OR have C privs
     */
    if (IS_PRIV (cmd->flags) && !(Is_local && Is_root)) {
	if (Checkperm (FQDN, Person, First_name, &perms, (int *) 0, 0) <= 0) {
	    (void) printf ("Sorry user %s, but you do not have printer control permission on '%s'!\n", 
			   Person, Printer);
	    (void) fflush (stdout);
	    return False;
	}
    }
    return ((*cmd->routine) (cmd));
}

/***************************************************************************
 * look in the command table for a match.  The "distinct" entry is used
 * to determine the numbers of characters for a match.
 ***************************************************************************/

static int
find_command (char *str) 
{
    int i;

    for (i = 1; i < ncmds; ++i) {
	if ((strncmp (str, dispatch[i].name, dispatch[i].distinct) == 0)
	    && (strncmp (str, dispatch[i].name, strlen (str)) == 0))
	    {
		return (i);
	    }
    }
    return False;
}

#ifdef C_BAD
/***************************************************************************
 * Cannot decide what to do with the command
 ***************************************************************************/
static int
C_bad (void) {
    (void) printf ("Bad command %s.\n", Parms[0].str);
    return True;
}
#endif /* C_BAD */

/***************************************************************************
 * Print a help message for each command listed,
 * or a simple list of commands
 ***************************************************************************/
static int
C_help (void)
{
    int i, cmd;
    char c[2];

    if (Parmcount < 2 || strsame (Parms[1].str, "all")) {
	for (i = 1; i < ncmds; ++i) {
	    /*
	     * Do a kind of VERY crude paging...
	     */
	    if (i == 20) {
		(void) fprintf (stdout, "Press RETURN to continue or q to quit help: ");
		(void) fgets (c, 2, stdin);
		(void) fflush (stdin);
		if ((c[0] == 'q') || (c[0] == 'Q')) {
		    i = 1;
		    return True;
		}
	    }
	    Print_sum (i);
	}
    } else {
	for (i = 1; i < Parmcount; ++i) {
	    cmd = find_command (Parms[i].str);
	    if (cmd > 0) {
		Print_sum (cmd);
	    } else {
		(void) printf ("Not a command: %s.\n", Parms[i].str);
	    }
	}
    }
    (void) fflush (stdout);
    return True;
}

/***************************************************************************
 * prints the command summary line
 ***************************************************************************/
static void
Print_sum (int cmd) {
    (void) printf ("%s\n", dispatch[cmd].summary);
}


/***************************************************************************
 * terminate gracefully
 ***************************************************************************/
static int
C_exit (void)
{
    printf ("LPC terminated.\n");
    exit (0);
    /* NOTREACHED, to keep lint happy */
    return (0);
}

/***************************************************************************
 * C_stop (void)
 * 1. Sets the DISABLE_PRINT perm bit in the lockfile perms
 * 2. Put message in Parms[] if it begins with a "#" into .status file
 * Returns: 1 if successful, 0 if not;
 ***************************************************************************/
static int
C_stop (void)
{
    int i, j, no_message, s;
    char statstr[BUFSIZ];
    
    *statstr = '\0';
    no_message = 1;
    DEBUG4 ("C_stop: printer %s, lock '%s'", Printer, LO);
    (void) Checklockfile (LO, (pid_t *) 0, (char *) 0, 0, &LO_statb);
    if ((check_if_printer_is_attached () ) && (called_from_unattach == False)) {
	if (strcmp (command_name, "down") == 0) {
	    (void) fprintf (stdout, "  Printer '%s' is attached to '%s'; use \"down %s\" to bring it down.\n", 
			    Printer, attached_printer, attached_printer);
	} else {
	    (void) fprintf (stdout, "  Printer '%s' is attached to '%s'; use \"stop %s\" to stop it.\n", 
			    Printer, attached_printer, attached_printer);
	}
	(void) fflush (stdout);
	return (False);
    }

    if ((s = chmod_daemon (LO, (int) (LO_statb.st_mode & 0777) | DISABLE_PRINT)) < 0) {
	logerr (XLOG_INFO, "cannot chmod lockfile %s", LO);
    } else {
        if (q_on_startstop) {
	    for (i = 0; i < Parmcount; i++) {
	        if (strstr (Parms[i].str, "#") != NULL) { /* issued a message */
		    strcat (statstr, Parms[i].str+1);
		    strcat (statstr, " ");
		    no_message = 0;
		    for (j = i + 1; j < Parmcount; j++) {
			strcat (statstr, Parms[j].str);
			strcat (statstr, " ");
		    }
		}
	    }
	    if (no_message) { /* No message issued, use default */
		default_message_stop (statstr);
	    } 
	    if (!called_from_one_job) {
		setstatus (statstr);
		Displayq (False, False);
	    }
	}
    }
    return (s >= 0);
}

/***************************************************************************
 * 1. Clears the DISABLE_PRINT perm bit in the lockfile perms
 * 2. Starts the Printer
 * Returns: 1 if successful, 0 if not;
 ***************************************************************************/
static int
C_start (void)
{
    int s;

    DEBUG4 ("C_start: printer %s, lock '%s'", Printer, LO);

    (void) Checklockfile (LO, (pid_t *) 0, (char *) 0, 0, &LO_statb);
    if (check_if_printer_is_attached ()) {
	if (strcmp (command_name, "up") == 0) {
	    (void) fprintf (stdout, "  Printer '%s' is attached to '%s'; use \"up %s\" to bring it up.\n", 
			    Printer, attached_printer, attached_printer);
	} else {
	    (void) fprintf (stdout, "  Printer '%s' is attached to '%s'; use \"start %s\" to start it.\n", 
			    Printer, attached_printer, attached_printer);
	}
	(void) fflush (stdout);
	return (False);
    }

    if ((s = chmod_daemon (LO, (int) (LO_statb.st_mode & ENABLE_PRINT))) < 0) {
	logerr (XLOG_INFO, "cannot chmod lockfile %s", LO);
    }
    
    /*
     * start the server
     */
    if (s >= 0) {
	s = C_restart ();
    }
    return (s >= 0);
}

/***************************************************************************
 * C_disable (void)
 * 1. Sets the DISABLE_QUEUE perm bit in the lockfile perms
 * 2. Put message in Parms[] if it begins with a "#" into .status file
 * Returns: 1 if successful, 0 if not;
 ***************************************************************************/
static int
C_disable (void)
{
    int i, j, no_message, s;
    char statstr[BUFSIZ];

    *statstr = '\0';
    no_message = 1;
    DEBUG4 ("C_disable: printer %s, lock '%s'", Printer, LO);
    (void) Checklockfile (LO, (pid_t *) 0, (char *) 0, 0, &LO_statb);

    if (check_if_printer_is_attached ()) {
	(void) fprintf (stdout, "  Printer '%s' is attached to '%s'; use \"disable %s\" to disable it.\n", 
			Printer, attached_printer, attached_printer);
	(void) fflush (stdout);
	return (False);
    }

    if ((s = chmod_daemon (LO, (int) (LO_statb.st_mode & 0777) | DISABLE_QUEUE)) < 0) {
	logerr (XLOG_INFO, "cannot chmod lockfile %s", LO);
    } else {
	if (q_on_startstop) {
	    for (i = 0; i < Parmcount; i++) {
	        if (strstr (Parms[i].str, "#") != NULL) { /* issued a message */
		    strcat (statstr, Parms[i].str+1);
		    strcat (statstr, " ");
		    no_message = 0;
		    for (j = i + 1; j < Parmcount; j++) {
			strcat (statstr, Parms[j].str);
			strcat (statstr, " ");
		    }
		}
	    }
	    if (no_message) {  /* No message issued, use default */
		default_message_disable (statstr);
	    }
	    setstatus (statstr);
	    Displayq (False, False);
	}
    }
    return (s >= 0);
}

/***************************************************************************
 * 1. Clears the DISABLE_QUEUE perm bit in the lockfile perms
 * Returns: 1 if successful, 0 if not;
 ***************************************************************************/
static int
C_enable (void)
{
    int s;

    DEBUG4 ("C_enable: printer %s, lock '%s'", Printer, LO);
    (void) Checklockfile (LO, (pid_t *) 0, (char *) 0, 0, &LO_statb);
    if (check_if_printer_is_attached ()) {
	(void) fprintf (stdout, "  Printer '%s' is attached to '%s'; use \"enable %s\" to enable it.\n", 
			Printer, attached_printer, attached_printer);
	(void) fflush (stdout);
	return (False);
    }

    if ((s = chmod_daemon (LO, (int) (LO_statb.st_mode & ENABLE_QUEUE))) < 0) {
	logerr (XLOG_INFO, "cannot chmod lockfile %s", LO);
    } else {
	if (q_on_startstop) {
	    Displayq (False, False);
	}
    }
    return (s >= 0);
}

/***************************************************************************
 * 1. Attempts to fire up the server
 * Returns: 1 if successful, 0 if not;
 ***************************************************************************/
static int
C_restart (void)
{
    /*
     * start the server
     */

    DEBUG4 ("C_restart: printer %s, lock '%s'", Printer, LO);
    if (check_if_printer_is_attached ()) {
	(void) fprintf (stdout, "  Printer '%s' is attached to '%s'; use \"restart %s\" to restart it.\n", 
			Printer, attached_printer, attached_printer);
	(void) fflush (stdout);
	return (False);
    }

    (void) Startserver ();
    (void) sleep (1);
    if (!called_from_one_job) {
	if (q_on_startstop) {
	    Displayq (False, False);
	}
    }
    (void) utime (SD, NULL);
    return True;
}

/**************************************************************************
 * killserver(): kills the current server to stop printing
 **************************************************************************/
static int
killserver (void)
{
    int s;			/* Success of operation */
    pid_t pid;			/* server PID */
    /*
     * Kill the current server to stop printing now.
     */
    s = True;
    if (Checklockfile (LO, &pid, (char *) 0, 0, &LO_statb)) {
	if (kill (pid, SIGINT) < 0) {
	    (void) logerr (XLOG_INFO, "server (pid %d) not killed", pid);
	    s = False;
	} else {
	    (void) printf ("%s '%s': server (pid %d) killed.\n",
			   SS ? "Server" : "Printer", Printer, (int) pid);
	}
    } 

    (void) utime (SD, NULL);
    (void) fflush (stdout);
    return (s);
}

/***************************************************************************
 * 1. Does C_stop(void)
 * 2. kills off server if there is one
 * Returns: 1 if successful, 0 if not;
 ***************************************************************************/
static int 
C_abort (void)
{
    int s = 0;			/* Success of operation */

    DEBUG4 ("C_abort: printer %s, lock '%s'", Printer, LO);
    if ((check_if_printer_is_attached () == True) && (called_from_unattach == False)) {
	(void) fprintf (stdout, "  Printer '%s' is attached to '%s'; use \"abort %s\" to abort it.\n", 
			Printer, attached_printer, attached_printer);
	(void) fflush (stdout);
	return (False);
    }

    q_on_startstop = False;
    /*    if (called_from_unattach == False) { */
    s = C_stop ();
    /* } */
    q_on_startstop = True;
    if (s) {
	s = killserver ();
	if (q_on_abortpromote == 1) {
	    Displayq (False, False);
	}
    }
    return (s);
}

/***************************************************************************
 * C_kill(void)
 * 1. Does C_abort(void)
 * 2. Does C_start(void)
 * Returns: 1 if successful, 0 if not;
 ***************************************************************************/
static int
C_kill (void)
{
    int s;			/* Success of operation */

    DEBUG4 ("C_kill: printer %s, lock '%s'", Printer, LO);
    if (check_if_printer_is_attached ()) {
	(void) fprintf (stdout, "  Printer '%s' is attached to '%s'; use \"kill %s\" to kill it.\n", 
			Printer, attached_printer, attached_printer);
	(void) fflush (stdout);
	return (False);
    }

    q_on_startstop = False;
    s = C_abort ();
    q_on_startstop = True;
    if (s) {
	q_on_startstop = False;
	s = C_start ();
	q_on_startstop = True;
    }
    return (s);
}

/***************************************************************************
 * 1. Removes all entries in the specified spool directory.
 * Returns: 1 if successful, 0 if not;
 ***************************************************************************/
static int
C_clean (void)
{
    int c;			/* ACME Integers, Inc. */
    DIR *dirp;
    plp_dir_t *d;
    pid_t pid;			/* server PID */
    char banner_file[BUFSIZ];

    DEBUG4 ("C_clean: printer %s, lock '%s'", Printer, LO);

    if (check_if_printer_is_attached ()) {
	(void) fprintf (stdout, "  Printer '%s' is attached to '%s'; use \"clean %s\" to clean it.\n", 
			Printer, attached_printer, attached_printer);
	(void) fflush (stdout);
	return (False);
    }

    if ((dirp = opendir_SD ()) == NULL) {
	logerr (XLOG_INFO, "cannot open spool directory %s");
	return False;
    }

    /* First the active server will be killed if there is one */
    if (Checklockfile (LO, &pid, (char *) 0, 0, &LO_statb)) {
	if (kill (pid, SIGINT) < 0) {
	    (void) logerr (XLOG_INFO, "server (pid %d) not killed", pid);
	} else {
	    (void) printf ("Printer '%s': server (pid %d) killed.\n",
			   Printer, (int) pid);
	}
    }
    
    (void) fprintf (stdout, "Cleaning in queue/printer '%s'.\n", Printer);
    line_counter++;
    while ((d = readdir (dirp)) != NULL) {
	c = d->d_name[0];
	if ((c == 'c' || c == 't' || c == 'd' || c == 'h') && d->d_name[1] == 'f') {
            if (unlink_daemon (d->d_name) < 0) {
		(void) printf ("Cannot remove %s from queue %s.\n", 
			       d->d_name, Printer);
            } else {
		(void) printf ("Removed %s from queue %s.\n", d->d_name, Printer);
		line_counter++;
	    }
	    (void) fflush (stdout);	    
	}
    }
    closedir (dirp);
    (void) sprintf (banner_file, "%s/banner_pg", SD);
    (void) unlink_daemon (banner_file);
    (void) sprintf (banner_file, "%s/banner_epg", SD);
    (void) unlink_daemon (banner_file);
    return True;
}

static void
parse_active (char *buf, char *pri, char **host)
{
    char *cp1, *cp2;

    *pri = buf[2];
    if (*pri < 'A' || *pri > 'Z')
	*pri = ' ';

    cp1 = buf;
    cp2 = buf + (*pri == ' ' ? 2 : 3);

    while (isdigit(*cp2))
	*cp1++ = *cp2++;

    *cp1 = '\0';

    *host = cp2;
}

/***************************************************************************
 * C_status (void)
 * Print the status of each queue listed - short format.
 ***************************************************************************/
static int 
C_status (void)
{
    char buf[BUFSIZ];            /* buffer */
    char servers[40];            /* buffer for printers */
    pid_t pid;			/* active server and its pid */
    int active = 0;              /* marker for active servers */ 
    int pstat, qstat;	         /* Printer, queue and reorder status */
    static int jobs;             /* number of jobs in the queue */
    short forms_info_printed_s;  /* buffers the contens of forms_info_printed */

    if (Get_pc_entry (Printer, Status_pc_vars, Status_pc_len) == 0) {
	printf ("Printer '%s' does not exist.\n", Printer);
	(void) fflush (stdout);
	return False;
    }

    if (SD == 0 || *SD == 0) {
	printf ("Printer %s does not have spool directory\n", Printer);
	(void) fflush (stdout);
	return False;
    }
    
    /* start by getting queue information */
    chdir_SD ();
    buf[0] = '\0';
    active = Checklockfile (LO, &pid, buf, sizeof (buf), &LO_statb);
    qstat = (LO_statb.st_mode & DISABLE_QUEUE);
    pstat = (LO_statb.st_mode & DISABLE_PRINT);
    jobs = Getq ();

    if ((Is_terse) && (qstat == 0) && (pstat == 0) && (Jobcount == 0))
	{
	    return True;
	}
    
    /* Now fill the sort arry with the information */
    (void) strcpy (servers, Printer);
    (void) fprintf (stdout, "%s:\n", servers);

    /* prints the queue status */
    (void) fprintf (stdout, 
		    "\tqueuing is %s\n\tprinting is %s\n",
		    qstat ? "disabled" : "enabled",
		    pstat ? "disabled" : "enabled");
    switch (jobs) {
    case 0:
	(void) fprintf (stdout, "\tno entries\n");
	break;
    case 1:
	(void) fprintf (stdout, "\t1 entry in spool area\n");
	break;
    default:
	(void) fprintf (stdout, "\t%d entries in spool area\n", jobs);
	break;
    }

    forms_info_printed_s = forms_info_printed;
    forms_info_printed = 1;
    if (active) {
	(void) fprintf (stdout, "\t");
	printstatus (0);
    } else {
	(void) fprintf (stdout, "\tno daemon present\n");
    }
    forms_info_printed = forms_info_printed_s; 
    (void) fflush (stdout);
    line_counter = line_counter + 6;
    return True;
}

 
/***************************************************************************
 * C_lstatus (void)
 * Print the status of each queue listed or all the queues - long format.
 ***************************************************************************/
static char *header_format      = "%-20.20s %3s %4s %8s %8s %5s %3s %4s %.20s\n";
static char *queue_format       = "%-20.20s %3s %4d %8s %8s";
static char *active_format      =			 " %5d  %c %5s %.20s\n";
static char *server_format      = "  Server %-26.26s    %8s";
static char *attach_format      = "%.8s %2s %-8s";

static int
C_lstatus (void)
{
    extern int Get_Default_Printer (int);

    FILE *afp;                  /* attach file pointer */  
    static int att_mark;        /* marker for attached printer */
    int active;			/* marker for active servers */ 
    pid_t pid;			/* active server and its pid */
    char afname[MAXPARMLEN];    /* attach file name */
    char buf[BUFSIZ];		/* buffer */
    int pstat, qstat, rstat;	/* Printer, queue and reorder status */
    char unattached_printer[MAXPARMLEN]; /* Name of the unattached printer */
    char servers[BUFSIZ];
    char ss[MAXPARMLEN];	/* Save SS */
    char *sp, *ep;		/* ACME Pointers, Inc. */
    char pri, *host;
    char location[4];           /* for the location string */

    att_mark = False;

    if ((afp = fopen_daemon (Attach_file, "r"))) { /* Try to open attach file */
	if (fscanf (afp, "%s", afname) == 1) {
	    if (strsame (afname, Printer)) {
		fatal (XLOG_INFO, "Printer '%s' attached to itself", Printer);
	    }
	    (void)strcpy(unattached_printer, Printer);
	    Printer = afname;
	    att_mark = True;
	}
	fclose(afp);
    }
    if (att_mark) {
	DBGLOCAL3 ("C_lstatus: printer %s attached to %s",	
		   unattached_printer, Printer);
    } else {
	DBGLOCAL3 ("C_lstatus: printer %s", Printer);
    }

    if (Get_pc_entry (Printer, Status_pc_vars, Status_pc_len) == 0) {
	(void) fprintf (stdout, "Printer '%s' does not exist.\n", Printer);
	(void) fflush (stdout);
	return False;
    }

    /* we have a server here */
    if (SS && *SS) {
	(void) strncpy (ss, SS, sizeof (ss));
	if (Get_multi_pc_entry (Printer, SS,
				Status_pc_vars, Status_pc_len) == 0)
	    {
		(void) fprintf (stdout, "Server '%s' queue '%s' does not exist.\n",
				Printer, SS);
		(void) fflush (stdout);
		return False;
	    }
	SV = NULL;
	(void) strncpy (SS, ss, sizeof (ss));
    } else {
	ss[0] = '\0';
    }

    if (SD == 0 || *SD == 0) {
	(void) fprintf (stdout, "Printer %s does not have spool directory\n", 
			Printer);
	(void) fflush (stdout);
	return False;
    }
    chdir_SD ();

    /* start by getting queue information */
    buf[0] = '\0';
    active = Checklockfile (LO, &pid, buf, sizeof (buf), &LO_statb);

    qstat = (LO_statb.st_mode & DISABLE_QUEUE);
    rstat = (LO_statb.st_mode & FORCE_REQUE);
    pstat = (LO_statb.st_mode & DISABLE_PRINT);
    Jobcount = Getq ();

    if ((Is_terse) && (qstat == 0) && (pstat == 0) && (Jobcount == 0))
	{
	    return True;
	}

    /* displays heading if not already displayed */
    if (!op_init) {
	(void) fprintf (stdout, header_format,
			"Name", "Loc", "Jobs", "Queueing", "Printing",
			"PID", "Pri", "Job", "Host");
	op_init = True;
    }

    DEBUG4 ("C_lstatus: active '%d', Jobcount '%d', '%s'",
	    active, Jobcount, buf);

    if (att_mark) { 
	(void) sprintf (servers, attach_format, 
			unattached_printer, "=>", Printer); 
    } else if (*ss) {
	/* Get active server info */
	active = Checklockfile (ss, &pid, buf, sizeof (buf), &LO_statb);
	(void) sprintf (servers, attach_format, Printer, "<=", ss);
    } else {
	(void) strcpy (servers, Printer);
    }

    if (Get_Default_Printer (1)) {
	sprintf (location, "def");
    } else {
	if (att_mark) { 
	    sprintf (location, "att");
	} else {
	    if (SA) {
		sprintf (location, "sav");
	    } else {
		RM ? sprintf (location ,"rem") : sprintf (location, "loc");
	    }
	}
    }

    /* prints the queue status */
    (void) fprintf (stdout, queue_format, servers, 
		    location, Jobcount,
		    qstat ? "disabled" : "enabled",
		    pstat ? "disabled" : "enabled");
    
    if (SV == 0 || *SV == 0) {
	if (Jobcount > 0 && active) {
	    parse_active (buf, &pri, &host);
	    (void) fprintf (stdout, active_format, (int) pid, pri, buf, host);
	} else {
	    if (ST && *ST && stat(ST, &LO_statb) == 0) {
		(void) fprintf (stdout, "  %s %s", pstat ? "Down since" : "Idle since", ctime (&LO_statb.st_mtime) + 4);
	    } else {
		(void) fprintf (stdout, pstat ? "  Down\n" : "  Idle\n"); 
	    }
	}
    } else {
	(void) fprintf (stdout, "\n");
    }
    line_counter++;

    if (rstat) {
	(void) fprintf (stdout, "  Reorder requested\n");
	line_counter++;
    }

    if (SV == NULL || *SV == '\0') {
	/* prints the Printer status */
	if (active || !Is_all || qstat || pstat || Jobcount) {
	    printstatus (0);
	    line_counter++;
	}
    } else {
	(void) strcpy (servers, SV);
	for (sp = servers; sp; sp = ep) {
	    ep = (char *) strchr (sp, ',');
	    if (ep) {
		*ep = '\0'; ep++;
	    }
	    active = Checklockfile (sp, &pid, buf, sizeof (buf), &LO_statb);
	    pstat = (LO_statb.st_mode & DISABLE_PRINT);

	    if (Get_multi_pc_entry (sp, Printer,
				    Status_pc_vars, Status_pc_len) == 0)
		{
		    (void) printf ("%s (sv from %s): no printcap entry.\n", 
				   sp, Printer);
		    continue;
		}
	    (void) printf (server_format, sp, pstat ? "disabled" : "enabled");

	    if (active == 0) {
		if (ST && *ST && stat(ST, &LO_statb) == 0) {
		    (void) fprintf (stdout, "  %s %s", pstat ? "Down since" : "Idle since",  
				    ctime (&LO_statb.st_mtime) + 4);
		} else {
		    (void) fprintf (stdout, pstat ? "  Down\n" : "  Idle\n");
		}
	    } else {
		parse_active (buf, &pri, &host);
		(void) fprintf (stdout, active_format, (int) pid, pri, buf, host);
	    }
	    if (active || !Is_all || pstat) {
		(void) fputs("  ", stdout);
		printstatus (0);
	    }
	    line_counter++;
	}
    }
    line_counter++;
    (void) fflush (stdout);
    return True;
}

/***********************************************************************
 * int C_history (void)
 * prints the history of a queue
 ***********************************************************************/
static int
C_history (void)
{
    char history_file[255];
    FILE *history_fp;
    time_t secs;
    char user[32];
    char machine[32];
    int seq_number;
    char priority[2];
    char jobname[22];
    int perms = 'C';
    int result;
    char buf[128];
    int permission = 0;
    static int head_printed = 0;
    extern void Remote_history (char *);

    /* Not for the save queue! */
    if (SA) 
	return True;

    DEBUG4 ("C_history: printer %s", Printer);

    if (!Use_history) {
	(void) printf ("History feature not active.\nTo enable it please edit the file \"%s\" and add the line \"use-history\".\n", CONFIG_FILE);
	(void) fflush (stdout);
	return False;
    }	
    
    /* check permissions */
    permission = Checkperm (FQDN, Person, First_name, &perms, (int *) 0, 0);


    /* check arguments in parms string */
    if (*history_user) {
	/* we have the "a" in history_user or exactly "all" or...*/
	if (((history_user[0] == 'a') && (history_user[1] == '\0')) || 
	    ((history_user[0] == 'a') && (history_user[1] == 'l') && 
	     (history_user[2] == 'l')&& (history_user[3] == '\0'))) {
	    all_history = 1;
	    DEBUG4 ("C_history: all_history is set");
	} else {
	    /* a username here; */
	    all_history = 0;
	    DEBUG4 ("C_history: all_history is not set");
	}
    }

    if (!head_printed) {
	Print_queuehead_hist ();
	head_printed = 1;
    }
    /* construct the history file name */
    (void) sprintf (history_file, "%s/.history", SD);
    user_to_daemon ();

    if ((history_fp = fopen (history_file, "r")) == NULL) {
	(void) fprintf (stdout, "No history available for this queue/printer.\n");
	(void) fflush (stdout);
    } else {
	while (fscanf (history_fp, "%ld", &secs) != EOF) {
	    (void) fscanf (history_fp, "%s", user);
	    (void) fscanf (history_fp, "%s", machine);
	    (void) fscanf (history_fp, "%d", &seq_number);
	    (void) fscanf (history_fp, "%s", priority);
	    (void) fgets (buf, 128, history_fp);
	    /* strip the "/n" */
	    buf[strlen (buf)-1] = '\0';
	    result = (int) atoi (&buf[strlen (buf)-2]);
	    buf[strlen (buf)-2] = '\0';
	    /* all data are read in; now evaluate it */
	    DEBUG4 ("C_history: checking history for %s and jobs of %s.", 
		    user, Person);
	    if (permission) {
		if (all_history) {
		    ;
		} else {
		    if (strlen (history_user) > 0) {
			if (strcmp (user, history_user) == 0) {
			    ;
			} else {
			    continue;
			}
		    } else {
			if (strcmp (user, Person) == 0) {
			    ;
			} else {
			    continue;
			}
		    }
		} 
	    } else {
		if (strcmp (user, Person) == 0) {
		    ;
		} else {
		    continue;
		}
	    }

	    (void) strncpy (jobname, buf, 21);
	    jobname[21] = '\0';
	    user[9] = '\0';
	    (void) fprintf (stdout,"%-15s %-9s %-16.*s %3d %s %-22s ",
			    Time_str_hist (secs),
			    user, 
			    16, machine, 
			    seq_number, 
			    priority, 
			    jobname); 
	    switch (result) {
	    case -1:
		(void) fprintf (stdout, " SENT\n");
		break;
	    case 0:
		(void) fprintf (stdout, "ERROR\n");
		break;
	    case 1:
		(void) fprintf (stdout, "  OK\n");
		break;
	    case 2:
	    case 3:
		(void) fprintf (stdout, "ERROR\n");
		break;
	    case 4:
		(void) fprintf (stdout, " SENT\n");
		break;
	    case 5:
		(void) fprintf (stdout, "DELETED\n");
		break;
	    case 6:
		(void) fprintf (stdout, " MOVED\n");
		break;
	    default:
		(void) fprintf (stdout, "UNKOWN\n");
		break;
	    }
	    (void) fflush (stdout);
	}
	fclose (history_fp);
    }
    daemon_to_user ();
    if (RM && *RM) {
	DEBUG4 ("C_history: getting Remote_history():");
	if (all_history) {
	    (void) strcpy (history_user, "all");
	}
	Remote_history (history_user);
	if (all_history) {
	    history_user[0] ='\0';
	}
    }
    all_history = 0;
    head_printed = 0;
    return True;
}

/***********************************************************************
 * int C_print_one (void)
 * prints only the specified job of the queue, even if the queue is
 * closed
 ***********************************************************************/
static int
C_print_one (void)
{
    char Parms_str_save[128];
    struct plp_queue *q;
    int i, Jobcount, job_not_found = 0;
    pid_t active_job;

    DEBUG3 ("C_print_one: printer %s", Printer);
    if (Parmcount != 1) {
	(void) printf ("Missing or bad job specification.\n");
	(void) fflush (stdout);
	return False;
    }
    DEBUG4 ("C_print_one: '%s'(%d)", Parms[0].str, Parms[0].num);
    /* Check if the job exists in the queue */
    if (Get_pc_entry (Printer, Status_pc_vars, Status_pc_len) == 0) {
	(void) fprintf (stdout, "Printer or Queue '%s' does not exist!\n", Printer);
	(void) fflush (stdout);
	return (False);
    }
    if (SD == 0 || *SD == 0) {
	DEBUG2 ("not a Printer");
	return (False);
    }
    chdir_SD ();
    Jobcount = Getq ();
    active_job = Checkactive ();
    if (active_job) {
	(void) fprintf (stdout, "Warning: job %s has active server (pid %d).",
			Parms[1].str, active_job);
	return (False);
    }
    for (i = 0; i < Jobcount; ++i) {
	q = &Queue[i];
	if (q->q_num != Parms[0].num) {
	    job_not_found = 1;
	} else {
	    job_not_found = 0;
	    break;
	}
    }
    
    if (job_not_found) {
	if (FQDN != Host) {
	    (void) fputs ("Remote ", stdout);
	} else {
	    (void) fputs ("Local  ", stdout);
	}
	if ((SV == 0 || *SV == 0) && (RM == 0 || *RM == 0)) {
	    (void) fputs ("printer ", stdout);
	} else { 
	    (void) fputs (" queue  ", stdout);
	}
	(void) fprintf (stdout, "'%s'", Printer);
	(void) fprintf (stdout, " on %s", Host);
	if (CM && *CM) {
	    (void) fputs (" (", stdout);
	    Print_comment (CM);
	    (void) fputs (")", stdout);
	}
	(void) fprintf (stdout, ":\n");
	(void) fprintf (stdout, "Warning: job %d does not exist.\n", Parms[0].num);
	(void) fflush (stdout);
	return (False);
    }

    /* First all jobs will have the hold status */
    (void) strncpy (Parms_str_save, Parms[0].str, sizeof (Parms_str_save) - 1);
    Parms_str_save[strlen (Parms_str_save)+1] = '\0';
    (void) sprintf (Parms[0].str, "*");
    called_from_one_job = 1;
    hold ();

    (void) strcpy (Parms[0].str, Parms_str_save);
    C_topq ();
    print_one (Parms[0].num);

    /* Now start the sever for this one job and stop it immediatly after that */
    (void) C_start ();
    (void) C_stop ();

    (void) strncpy (Parms_str_save, Parms[0].str, sizeof (Parms_str_save) - 1);
    Parms_str_save[strlen (Parms_str_save)+1] = '\0';
    (void) sprintf (Parms[0].str, "*");
    /* Now release all the jobs */
    release ();
    called_from_one_job = 0;
    (void) fflush (stdout);
    return True;
}

/***********************************************************************
 * int C_topq (void)
 * Put the specified jobs at the top of Printer queue.
 ***********************************************************************/
static int
C_topq (void) {
    int i, job_is_there;	/* ACME Integer, Inc. */
    int changed;		/* We changed the queue */
    char *cfname;		/* control file name */
    unsigned short *not_found;


    if (check_if_printer_is_attached ()) {
	(void) fprintf (stdout, "  Printer '%s' is attached to '%s'; use \"topq %s\" to perform it.\n", 
			Printer, attached_printer, attached_printer);
	(void) fflush (stdout);
	return (False);
    }

    if (q_on_abortpromote != 2) {
	if (FQDN != Host) {
	    (void) fputs ("Remote ", stdout);
	} else {
	    (void) fputs ("Local  ", stdout);
	}
	if ((SV == 0 || *SV == 0) && (RM == 0 || *RM == 0)) {
	    (void) fputs ("printer ", stdout);
	} else { 
	    (void) fputs (" queue  ", stdout);
	}
	(void) fprintf (stdout, "'%s'", Printer);
	(void) fprintf (stdout, " on %s", Host);
	if (CM && *CM) {
	    (void) fputs (" (", stdout);
	    Print_comment (CM);
	    (void) fputs (")", stdout);
	}
	(void) fprintf (stdout, ":\n");
	(void) fflush (stdout);
    }
    Setuid_debug("top_q");

    /*
     * put the parameters in the list in the correct position
     */
    if (Parmcount == 0) {
	(void) printf ("error: topq missing job specification.\n");
	(void) fflush (stdout);
	return False;
    }
    DEBUG4 ("C_topq: '%s'(%d)", Parms[0].str, Parms[0].num);
    Jobcount = Getq ();		/* get number of jobs in the queue */

    not_found = (unsigned short *) malloc (Parmcount * sizeof (unsigned short));
    if (not_found == NULL) {
	logerr_die (XLOG_INFO, "C_downq: Malloc for not_found failed.");
    }
    for (i = 0; i < Parmcount; i++) {
	not_found[i] = 0;
    }
    /* go through the queue and find the jobs to be promoted */
    changed = False;
    (void) Checkactive (); /* Check if there is an active job; the priority of
			    * job will not be changed */ 

    job_is_there = 0;
    for (i = 0; i < Jobcount; ++i) {
	if (Match_entry (&Queue[i])) {
	    job_is_there = 1;
	    if (Queue[i].q_daemon == 0) {
		/*
		 * Reposition the job by setting its priority to high level 
		 * and then fix the control file
		 */
		if (promote (Queue[i].q_name, Queue[i].q_user)) {
		    changed++;
		    Queue[i].q_name[0] = '\0';
		    /* set the elements of the array of not_found to 0
		     * to avoid double listing of jobnames
		     */
		    {
			int k;
			for (k = 0; k < i; k++) {
			    if (not_found[k] == Queue[i].q_num) {
				not_found[k] = 0;
			    }
			}
		    }
		}
	    } else { /* Active job, issue a warning message */
		(void) fprintf (stdout, "Warning: job %d active, priority cannot be changed.\n", 
				Queue[i].q_num);
		(void) fflush (stdout);
	    }
	}
	if ((!job_is_there) && (i < Parmcount)){
	    not_found[i] = (unsigned short) Parms[i].num;
	}

    } 
    for (i = 0; i < Parmcount; i++) {
	if (not_found[i]) {
	    (void) fprintf (stdout, 
			    "Warning: job %d does not exist on queue/printer '%s'.\n", 
			    not_found[i], Printer);
	}
    }
    (void) fflush (stdout);
    free ((unsigned short *) not_found);

    /*
     * Put the other high priority jobs lower
     */
    if (!changed) {
	return True;
    }
    for (i = 0; i < Jobcount; i++) {
	cfname = Queue[i].q_name;
	if (*cfname && cfname[2] == 'A') {
	    touch (cfname);
	}
    }
    if (!called_from_one_job) {
	if ( q_on_abortpromote != 2) {
	    (void) fprintf (stdout, "Order changed for ");
	    if (FQDN != Host) {
		(void) fputs ("remote ", stdout);
	    } else {
		(void) fputs ("local  ", stdout);
	    }
	    if ((SV == 0 || *SV == 0) && (RM == 0 || *RM == 0)) {
		(void) fputs ("printer ", stdout);
	    } else { 
		(void) fputs (" queue  ", stdout);
	    }
	    (void) fprintf (stdout, "'%s'", Printer);
	    (void) fprintf (stdout, " on %s.\n", Host);
	    (void) fflush (stdout);
	}
    }

    /*
     * Turn on the public execute bit of the lock file to get lpd to rebuild the queue
     * after the current job.
     */
    (void) Checklockfile (LO, (pid_t *) 0, (char *) 0, 0, &LO_statb);
    if (chmod_daemon (LO, (int) (LO_statb.st_mode & 0777) | FORCE_REQUE) < 0) {
	logerr (XLOG_INFO, "cannot force requeue on lockfile %s", LO);
	return False;
    }
    (void) fflush (stdout);
    return True;
}

/***************************************************************************
 * int promote( char *cfname, char *user )
 * promote a file to the A priority (top of queue), and head of the list.
 ***************************************************************************/
static int
promote (char *cfname, char *user) {
    char buf[BUFSIZ];         /* temporary cfname buffer */
    char c_old;               /* old priority */
    char c_new;               /* new priority */
    char c_job[4];            /* job number (as char's) */
    int num;                  /* real job number  */

    if (check_if_printer_is_attached ()) {
	(void) fprintf (stdout, "  Printer '%s' is attached to '%s'; use \"topq %s\" to perform it.\n", 
			Printer, attached_printer, attached_printer);
	(void) fflush (stdout);
	return (False);
    }

    num = 0;
    (void) strcpy (buf, cfname);
    buf[STARTPR] = 'A';
    DEBUG0 ("renaming %s to %s", cfname, buf);
    user_to_daemon ();
    Setuid_debug ("promote");
    if (strcmp (buf, cfname) && rename (cfname, buf) < 0) {
	logerr (XLOG_INFO, "cannot rename %s to %s", cfname, buf);
	daemon_to_user (); /* We are still daemon in this case, so switch back */
	return False;
    }
    Setuid_debug ("promote");
    daemon_to_user ();
    Setuid_debug ("promote");
    (void) strncpy (c_job, &buf[3], 3);
    c_job[3] = '\0';
    c_old = cfname[2];
    c_new = buf[2];
    num = (int) atoi (c_job);
    
    if (q_on_abortpromote != 2) {
	(void) printf ("Changing priority for job %d, owner %s, from %c to %c.\n",
		       num, user, c_old, c_new);
	(void) fflush (stdout);
    } else {
	(void) printf ("Changed priority to A for job %d, owner %s, after requeueing.\n", 
		       num, user);
    }
    (void) fflush (stdout);
    return True;
}


/***************************************************************************
 * int changepriority (char *cfname, char *user)
 * change the priority of a file to the given priority and reorder queue.
 ***************************************************************************/
static int
changepriority (char *cfname, char *user) 
{
    char buf[BUFSIZ];         /* temporary cfname buffer */
    char c_old;               /* old priority */
    char c_new;               /* new priority */
    char c_job[4];            /* job number (as char's) */
    int num;                  /* real job number  */

    if (check_if_printer_is_attached ()) {
	(void) fprintf (stdout, "  Printer '%s' is attached to '%s'; use \"chprio %s\" to perform it.\n", 
			Printer, attached_printer, attached_printer);
	(void) fflush (stdout);
	return (False);
    }

    (void) strcpy (buf, cfname);
    num = 0;

    if (isdigit (*Parms[2].str)) /* 
				  * The given priority must 
				  * be an uppercase letter, not a digit
				  * We'll check that and convert it, if
				  * necessary (Stef)
				  */
	{
	    (void) printf ("Error: bad priority, must be a single letter (A-Z).\n");
	    (void) fflush (stdout);
	    return False;
	}

    if (islower (*Parms[2].str)) {
	/* Lowercase letters will be "uppercased" (Stef) */
	*Parms[2].str = toupper(*Parms[2].str);
	DEBUG2 ("chprio: toupper = %s", Parms[2].str); 
    }
    
    if ( !(isupper (*Parms[2].str)) ) {
	/* Check if the given priority was really an uppercase letter (Stef) */
	Warnmsg ("Error: bad priority; must be a single letter [A to Z], not %s.",
		 Parms[2].str);
	(void) fflush (stdout);
	return False;
    }	    
    
    buf[STARTPR] = *Parms[2].str;
    DEBUG0 ("changepriority: renaming %s to %s", cfname, buf);
    Setuid_debug("changepriority");
    user_to_daemon ();	
    if (strcmp (buf, cfname) && rename (cfname, buf) < 0) {
	logerr (XLOG_INFO, "changepriority: cannot rename %s to %s", cfname, buf);
	daemon_to_user (); /* We are still daemon in this case, so switch back */
	return False;
    }
    Setuid_debug ("changepriority");
    daemon_to_user ();
    Setuid_debug ("changepriority");
    (void)strncpy (c_job, &buf[3], 3);     /* Saving the jobnumber */
    c_job[3] = '\0';
    c_old = cfname[2];                     /* Saving the jobname */
    c_new = buf[2];
    num = (int) atoi (c_job);               /* Converting the job number */

    (void) printf ("Changing priority for job %d, owner %s, from %c to %c.\n",
		   num, user, c_old, c_new);
    (void) fflush (stdout);
    return True;
}


/***************************************************************************
 * demote a file to the Z priority (at the end of the queue), 
 * and head of the list.
 ***************************************************************************/
static int
demote (char *cfname, char *user) 
{
    char buf[BUFSIZ];         /* temporary cfname buffer */
    char c_old;               /* old priority */
    char c_new;               /* new priority */
    char c_job[4];            /* job number (as char's) */
    int num;                  /* real job number  */

    num = 0;
    (void) strcpy (buf, cfname);
    buf[STARTPR] = 'Z';
    DEBUG0 ("renaming %s to %s", cfname, buf);
    Setuid_debug("demote");
    user_to_daemon ();
    Setuid_debug("demote");
    if (strcmp (buf, cfname) && rename (cfname, buf) < 0)
	{
	    logerr (XLOG_INFO, "cannot rename %s to %s", cfname, buf);
	    daemon_to_user (); /* We are still daemon in this case, so switch back */
	    return False;
	}
    Setuid_debug ("demote");
    daemon_to_user ();
    Setuid_debug ("demote");
    (void) strncpy (c_job, &buf[3], 3); /* Saving the jobnumber */
    c_job[3] = '\0';
    c_old = cfname[2];                /* Saving the jobname */
    c_new = buf[2];
    num = (int) atoi (c_job);         /* Converting the job number */

    (void) printf ("Changing priority for job %d, owner %s, from %c to %c.\n", 
		   num, user, c_old, c_new);
    (void) fflush (stdout);
    return True;
}


/***************************************************************************
 * Change the modification time of the file. Returns boolean if successful.
 ***************************************************************************/
static void
touch (char *cfname) 
{
    FILE *fp;
    int i;

    if ((fp = Lockcf (cfname)) == NULL) {
	logerr (XLOG_INFO, "cannot open %s\n", cfname);
    } else if ((i = getc (fp)) == EOF) {
	logerr (XLOG_INFO, "cannot read %s\n", cfname);
    } else if (fseek (fp, 0L, 0) < 0) {;
    /* set pointer back to top of file */
    logerr (XLOG_INFO, "cannot seek %s\n", cfname);
    } else if (putc (i, fp) == EOF) {
	logerr (XLOG_INFO, "cannot write %s\n", cfname);
    } else if (fflush (fp) == EOF) {
	logerr (XLOG_INFO, "cannot flush %s\n", cfname);
    }
    if (fp != NULL) {
	(void) fclose (fp);
    }
}

/***************************************************************************
 * shows the values of the local options
 ***************************************************************************/
static void
Show_ops (void)
{
    int i;
    (void) printf ("From: %s, Host %s, Person %s\n", FQDN, Host, Person);
    (void) printf ("Is_local %d, Is_root %d\n", Is_local, Is_root);
    (void) printf ("Parmcount %d ", Parmcount);
    for (i = 0; i < Parmcount; ++i) {
	(void) printf (" '%s'(%d)", Parms[i].str, Parms[i].num);
    }
    (void) printf ("\n");
    (void) fflush (stdout);
}

/***************************************************************************
 * Change the priority of a specified job in the queue
****************************************************************************/
static int
C_chprio (void)
{
    int i, job_is_there;		/* ACME Integer, Inc. */
    int changed;			/* We changed the queue */
    char *cfname;       		/* control file name */

    if (check_if_printer_is_attached ()) {
	(void) fprintf (stdout, "  Printer '%s' is attached to '%s'; use \"chprio %s\" to perform it.\n", 
			Printer, attached_printer, attached_printer);
	(void) fflush (stdout);
	return (False);
    }

    if (FQDN != Host) {
	(void) fputs ("Remote ", stdout);
    } else {
	(void) fputs ("Local  ", stdout);
    }
    if ((SV == 0 || *SV == 0) && (RM == 0 || *RM == 0)) {
	(void) fputs ("printer ", stdout);
    } else { 
	(void) fputs (" queue  ", stdout);
    }
    (void) fprintf (stdout, "'%s'", Printer);
    (void) fprintf (stdout, " on %s", Host);
    if (CM && *CM) {
	(void) fputs (" (", stdout);
	Print_comment (CM);
	(void) fputs (")", stdout);
    }
    (void) fprintf (stdout, ":\n");
    (void) fflush (stdout);

    /*
     * put the parameters in the list in the correct position and check the
     * number of arguments given
     */
    if (Parmcount == 0) {
	(void) printf ("chprio: missing job specification.\n");
	(void) fflush (stdout);
	return False;
    }
    if ( (Parmcount < 2) ) {
	(void) printf ("chprio: missing priority specification.\n");
	(void) fflush (stdout);
	return False;
    } else if (Parmcount > 2 ) {
	(void) printf ("chprio: too many priorities specified.\n");
	(void) fflush (stdout);
	return False;
    }


    DEBUG4 ("C_chprio: '%s'(%d)", Parms[0].str, Parms[0].num);
    /* 
     * get number of jobs in the queue 
     */
    Jobcount = Getq ();
    /*
     * go through the queue and find the jobs to be changed in priority; but only 
     * inactive jobs may be changed in their priority!
     */
    changed = False;
    (void) Checkactive (); /* Check if there is an active job; the priority of
			    * job will not be changed
			    */ 
    job_is_there = 0;
    for (i = 0; i < Jobcount; ++i)
	{
	    if (Match_entry (&Queue[i])) {
		job_is_there = 1;
		if (Queue[i].q_daemon == 0) {
		    /*
		     * Reposition the job by setting its priority to 
		     * the specified level and then fix the control file
		     */
		    if (changepriority (Queue[i].q_name, Queue[i].q_user) )
			{
			    changed++;
			    Queue[i].q_name[0] = '\0';
			}
		} else { /* Active job, issue a warning message */
		    (void) printf
			("Warning: job %d active, priority cannot be changed.\n", 
			 Queue[i].q_num);
		}
	    }
	}
    if (!job_is_there) {
	(void) fprintf (stdout, 
			"Warning: job %d does not exist on queue/printer '%s'.\n", 
			Parms[0].num, Printer);
    }

    /*
     * Reorder the queue
     */
    if (!changed)
	{
	    return True;
	}
    for (i = 0; i < Jobcount; i++)
	{
	    cfname = Queue[i].q_name;
	    DEBUG4 ("C_chprio: Parms[2].str: %s", Parms[2].str);
	    if (*cfname && cfname[2] == *Parms[2].str) {
		touch (cfname);
	    }
	}
   
    (void) fprintf (stdout, "Order changed for ");
    if (FQDN != Host) {
	(void) fputs ("remote ", stdout);
    } else {
	(void) fputs ("local  ", stdout);
    }
    if ((SV == 0 || *SV == 0) && (RM == 0 || *RM == 0)) {
	(void) fputs ("printer ", stdout);
    } else { 
	(void) fputs (" queue  ", stdout);
    }
    (void) fprintf (stdout, "'%s'", Printer);
    (void) fprintf (stdout, " on %s.\n", Host);
   
    /*
     * Turn on the public execute bit of the lock file to
     * get lpd to rebuild the queue after the current job.
     */
    (void) Checklockfile (LO, (pid_t *) 0, (char *) 0, 0, &LO_statb);
    if (chmod_daemon (LO, (int) (LO_statb.st_mode & 0777) | FORCE_REQUE) < 0)
	{
	    logerr (XLOG_INFO, "cannot force requeue on lockfile %s", LO);
	    return False;
	}
    (void) fflush (stdout);
    return True;
}

/***************************************************************************
 * Put the specified jobs at the end of Printer queue.
 ***************************************************************************/
static int
C_downq (void)
{
    int i, job_is_there;	/* ACME Integer, Inc. */
    int changed;			/* We changed the queue */
    char *cfname;		/* control file name */
    unsigned short *not_found;

    if (check_if_printer_is_attached ()) {
	(void) fprintf (stdout, "  Printer '%s' is attached to '%s'; use \"downq %s\" to perform it.\n", 
			Printer, attached_printer, attached_printer);
	(void) fflush (stdout);
	return (False);
    }

    if (FQDN != Host) {
	(void) fputs ("Remote ", stdout);
    } else {
	(void) fputs ("Local  ", stdout);
    }
    if ((SV == 0 || *SV == 0) && (RM == 0 || *RM == 0)) {
	(void) fputs ("printer ", stdout);
    } else { 
	(void) fputs (" queue  ", stdout);
    }
    (void) fprintf (stdout, "'%s'", Printer);
    (void) fprintf (stdout, " on %s", Host);
    if (CM && *CM) {
	(void) fputs (" (", stdout);
	Print_comment (CM);
	(void) fputs (")", stdout);
    }
    (void) fprintf (stdout, ":\n");
    (void) fflush (stdout);

    /*
     * put the parameters in the list in the correct position
     */
    if (Parmcount == 0)
	{
	    (void) printf ("downq: missing job specification.\n");
	    (void) fflush (stdout);
	    return False;
	}
    DEBUG4 ("C_downq: '%s'(%d)", Parms[0].str, Parms[0].num);
    /* 
     * get number of jobs in the queue 
     */
    Jobcount = Getq ();
    not_found = (unsigned short *) malloc (Parmcount * sizeof (unsigned short));
    if (not_found == NULL) {
	logerr_die (XLOG_INFO, "C_downq: Malloc for not_found failed.");
    }
    for (i = 0; i < Parmcount; i++) {
	not_found[i] = 0;
    }
    /*
     * go through the queue and find the jobs to be demoted
     */
    changed = False;
    (void) Checkactive (); /* Check if there is an active job; the priority of
			    * an active job will not be changed
			    */ 
    job_is_there = 0;
    for (i = 0; i < Jobcount; ++i) {
	if (Match_entry (&Queue[i])) {
	    job_is_there = 1;
	    if (Queue[i].q_daemon == 0) {
		/*
		 * Reposition the job by setting its priority to low level
		 * and then fix the control file
		 */
		if (demote (Queue[i].q_name, Queue[i].q_user) )
		    {
			changed++;
			Queue[i].q_name[0] = '\0';
			/* set the elements of the array of not_found to 0
			 * to avoid double listing of jobnames
			 */
			{
			    int k;
			    for (k = 0; k < i; k++) {
				if (not_found[k] == Queue[i].q_num) {
				    not_found[k] = 0;
				}
			    }
			}
		    }
	    } else { /* Active job, issue a warning message */
		(void) fprintf (stdout, "Warning: job %d active, priority cannot be changed.\n", 
				Queue[i].q_num);
		(void) fflush (stdout);
	    }	   
	}
	if ((!job_is_there) && (i < Parmcount)){
	    not_found[i] = (unsigned short) Parms[i].num;
	}
    }
    for (i = 0; i < Parmcount; i++) {
	if (not_found[i]) {
	    (void) fprintf (stdout, 
			    "Warning: job %d does not exist on queue/printer '%s'.\n", 
			    not_found[i], Printer);
	}
    }
    (void) fflush (stdout);
    free ((unsigned short *) not_found);
    /*
     * Put the other priority jobs higher
     */
    if (!changed)
	{
	    return True;
	}
    for (i = 0; i < Jobcount; i++)
	{
	    cfname = Queue[i].q_name;
	    if (*cfname && cfname[2] == 'Z')
		{
		    touch (cfname);
		}
	}
   
    (void) fprintf (stdout, "Order changed for ");
    if (FQDN != Host) {
	(void) fputs ("remote ", stdout);
    } else {
	(void) fputs ("local  ", stdout);
    }
    if ((SV == 0 || *SV == 0) && (RM == 0 || *RM == 0)) {
	(void) fputs ("printer ", stdout);
    } else { 
	(void) fputs (" queue  ", stdout);
    }
    (void) fprintf (stdout, "'%s'", Printer);
    (void) fprintf (stdout, " on %s.\n", Host);
    /*
     * Turn on the public execute bit of the lock file to
     * get lpd to rebuild the queue after the current job.
     */
    (void) Checklockfile (LO, (pid_t *) 0, (char *) 0, 0, &LO_statb);
    if (chmod_daemon (LO, (int) (LO_statb.st_mode & 0777) | FORCE_REQUE) < 0)
	{
	    logerr (XLOG_INFO, "cannot force requeue on lockfile %s", LO);
	    return False;
	}
    (void) fflush (stdout);
    return True;
}


/***************************************************************************
 * C_lpq(void)
 * invoke Lpq with parameters.
 ***************************************************************************/
static int
C_lpq (void)
{
    char buf[BUFSIZ];
    char *bp, *ep;		/* ACME Pointers, Inc. */
    int i;

    ep = buf + sizeof (buf);
    bp = estrcp (buf, "lpq", ep);
    for (i = 0; i < Parmcount; ++i) {
	bp = estrcp (bp, " ", ep);
	bp = estrcp (bp, Parms[i].str, ep);
    }

    DEBUG4 ("C_lpq: '%s'", buf);
    i = secure_system (buf);
    DEBUG4 ("C_lpq: status %d", i);
    return (i);
}

/***************************************************************************
 * C_lprm(void)
 * invoke Lprm with parameters.
 ***************************************************************************/
static int
C_lprm (void)
{
    char buf[BUFSIZ];
    char *bp, *ep;		/* ACME Pointers, Inc. */
    int i;

    ep = buf + sizeof (buf);
    bp = estrcp (buf, "lprm", ep);
    for (i = 0; i < Parmcount; ++i) {
	bp = estrcp (bp, " ", ep);
	bp = estrcp (bp, Parms[i].str, ep);
    }

    DEBUG4 ("C_lprm: '%s'", buf);
    lprm_called = True;
    i = secure_system (buf);
    DEBUG4 ("C_lprm: status %d", i);
    lprm_called = False;
    return (i);
}

/***************************************************************************
 * C_remote(void)
 * do the indicated command on the remote host
 * command has the form: "remote op printer"
 * 1. get the command and see if it can be done remotely
 *    Parms[0] will be op, Parms[1] will be printer
 * 2. check the printer and make sure that it is a remote printer
 * 3. send the remote command to the far end, and get the information
 *    sent back.
 ***************************************************************************/

static int
C_remote (struct dispatch *cmd) {
    char **list;		/* list of printers */
    int i, j, k;		/* ACME Integers and Bearings, Inc. */
    int command;		/* command we are going to do */
    char param[BUFSIZ];		/* buffer for parameter list */

    if (Parmcount < 2) {
	(void) printf ("No or missing command or parameters.\n");
	(void) fflush (stdout);
	return False;
    }
    command = find_command (Parms[0].str);
    if (command == 0) {
	(void) printf ("Unknown command %s.\n", Parms[0].str);
	(void) fflush (stdout);
	return False;
    }

    DEBUG3 ("C_remote: command %s, %d, %s",
	    dispatch[command].name, dispatch[command].flags, dispatch[command].summary);

    if (!IS_REMOTE (dispatch[command].flags)) {
	(void) printf ("Command %s cannot be done remotely.\n",
		       Parms[0].str);
	(void) fflush (stdout);
	return False;
    }
    /*
     * set up parameter list
     */
    *param = '\0';
    if (IS_NO_PR_LIST (dispatch[command].flags)) {
	for (i = 2; i < Parmcount; ++i) {
	    strcat (param, " ");
	    strcat (param, Parms[i].str);
	}
    }
    /* special actions for "history" command */
    if (strsame (dispatch[command].name, "history")) {
	int printer_found = False;
	for (list = All_printers (); *list; ++list) {
	    if (strsame (Parms[1].str, *list)) {
		/* the second argument is a printer not a username */
		Printer = Parms[1].str;
		printer_found = True;
		*param = '\0';
		if (remote_cando (dispatch[command].flags)) {
		    Remote_control (dispatch[command].name, param);
		}
		break;
	    }
	}
	if (!printer_found) { /* must be a username */
	    (void) strcpy (param, Parms[1].str);
	    if (strsame (Parms[2].str, "all")) {
		for (list = All_printers (); *list; ++list) {
		    Printer = *list;
		    if (remote_cando (dispatch[command].flags)) {
			Remote_control (dispatch[command].name, param);
		    }
		}
	    } else {
		Printer = Parms[2].str;
		if (remote_cando (dispatch[command].flags)) {
		    Remote_control (dispatch[command].name, param);
		}
	    }
	}
	return True;
    } 
    /*
     * check on the printer
     */
    if (strsame ("all", Parms[1].str)) {
	DEBUG2 ("'all' parameter");
	for (list = All_printers (); *list; ++list) {
	    Printer = *list;
	    if (remote_cando (dispatch[command].flags)) {
		Remote_control (dispatch[command].name, param);
	    }
	}
    } else {
	for (i = 1; i < Parmcount; ++i) {
	    if (strstr (Parms[i].str, "#") != NULL) { /* found message; it
							 must came at the end of
							 the string! */
	        break;
	    } else {
	        Printer = Parms[i].str;
		for (j = 0; j < Parmcount; j++) {
		    if (strstr (Parms[j].str, "#")) { /* found message */
			param[0] = '\0';
		        strcat (param, " ");          /* send it to the remote lpd */
		        strcat (param, Parms[j].str);
		        for (k = j + 1; k < Parmcount; k++) {
			    strcat (param, " ");
			    strcat (param, Parms[k].str);
			}
		    }
		}
		if (remote_cando (dispatch[command].flags)) {
		    Remote_control (dispatch[command].name, param);
		}
		if (IS_NO_PR_LIST (dispatch[command].flags)) {
		    break;
		}
	    }
	}
    }
    return True;
}

/***************************************************************************
 * returns 0 if not able to do to remote site, 1 otherwise
 ***************************************************************************/
static int
remote_cando (int flags) 
{
    int perms = 'C';		/* Permission Checking */

    if (Get_pc_entry (Printer, Status_pc_vars, Status_pc_len) == 0) {
	printf ("printer '%s' does not exist.\n", Printer);
	(void) fflush (stdout);
	return False;
    }
    if (RM == 0 || RP == 0) {
	printf ("printer '%s' is not remote.\n", Printer);
	(void) fflush (stdout);
	return False;
    }
    /*
     * check on the privileges needed You have to be local and root OR have C privs
     */
    if (IS_PRIV (flags) && !(Is_local && Is_root) &&
	Checkperm (FQDN, Person, First_name, &perms, (int *) 0, 0) <= 0) {
	printf ("Sorry user %s, but you do not have printer control permissions on '%s'!\n",
		Person, Printer);
	(void) fflush (stdout);
	return False;
    }
    return True;
}

/***************************************************************************
 * C_lpd(void)
 * 1. Check to see if there is an active lpd server by checking the
 *    lock file.
 * 2. Check to see if the /dev/printer is available as well.
 ***************************************************************************/

static int
C_lpd (void)
{
    pid_t f;

    (void) Checklockfile (Masterlock, &f, (char *) 0, 0, &LO_statb);
    if (f) {
	(void) printf ("Active LPD %ld\n", (long) f);
    } else {
	(void) printf ("LPD is not active.\n");
    }
    (void) fflush (stdout);
    return True;
}

/***************************************************************************
 * C_up(void)
 * Calls C_enable and C_start and shows the queue
 ***************************************************************************/
static int
C_up (void)
{
    q_on_startstop = False;
    if (C_start () && C_enable ()) {
	/* Display nothing when printer is attached... */
	if (check_if_printer_is_attached () == False) {
	    q_on_startstop = True;
	    Displayq (False, False);
	}
    }
    return (True);
}

/***************************************************************************
 * C_down(void)
 * 1. Call C_stop and C_disable
 * 2. Put message in Parms[] if it begins with a "#" into .status file
 ***************************************************************************/
static int
C_down (void)
{
    int i, j, no_message;
    char statstr[BUFSIZ];

    *statstr = '\0';
    no_message = 1;
    q_on_startstop = False;
    if (C_stop () && C_disable ()) {
	for (i = 0; i < Parmcount; i++) {
	    if (strstr (Parms[i].str, "#") != NULL) { /* issued a message */
		strcat (statstr, Parms[i].str+1);
		strcat (statstr, " ");
		no_message = 0;
		for (j = i + 1; j < Parmcount; j++) {
		    strcat (statstr, Parms[j].str);
		    strcat (statstr, " ");
		}
	    }
	}
	if (no_message) { /* No message issued, use default */
	    default_message_down (statstr);
	}
	setstatus (statstr);
    }
    /* Display nothing if the printer is attached */
    if (check_if_printer_is_attached () == False) {
	q_on_startstop = True;
	Displayq (False, False);
    }
    return True;
}


/***************************************************************************
 * int C_attach (void)
 * attaches the specified printer to another printer 
 ***************************************************************************/
static int
C_attach (void)
{
    FILE *afp;			/* attach file pointer */
    int s;			/* success of abort/start */

    DEBUG3 ("C_attach: printer %s", Printer);
    if (Parmcount != 1) {
	(void) printf ("Missing destination specification.\n");
	(void) fflush (stdout);
	return False;
    }
    if (!*Attach_file) {
	(void) printf ("Warning: 'attach' feature not active.\n");
	(void) fflush (stdout);
	return False;
    }
    DEBUG4 ("C_attach: '%s'(%d)", Parms[0].str, Parms[0].num);

    /* 
     * Check that a printer is not attached to itself (stupid, eh...?)
     */
    if (strsame(Printer, Parms[0].str))
	{
	    log (XLOG_INFO, "\n  Error: printer '%s' cannot be attached to itself!", Printer);
	    return (False);
	}
    if (check_if_printer_is_attached () == True) {
        (void) fprintf (stdout, "  Printer '%s' is already attached to '%s'.\n",
			Printer, attached_printer);
	(void) fflush (stdout);
	return (False);
    }

    q_on_startstop = False;
    q_on_abortpromote = False;
    s = C_abort ();
    if (s >= 0) {
	if ((afp = fopen_daemon (Attach_file, "w"))) {
	    if (fprintf (afp, "%s\n", Parms[0].str) == EOF) {
		log (XLOG_INFO, "\n  Warning: can't attach printer '%s'.", Printer);
		(void) fclose (afp);
		return (False);
	    }
	    (void) fclose (afp);
	    (void) chmod (Attach_file, (S_IRUSR | S_IWUSR));
	} else {
	    log (XLOG_INFO, "\n  Warning: can't open file to attach printer '%s'.", Printer);
	    return (False);
	}
    }
    (void) fflush (stdout);
    q_on_startstop = True;
    q_on_abortpromote = True;
    Displayq (False, False);
    return (s);
}

/***************************************************************************
 * unattaches the specified printer from another printer 
 ***************************************************************************/
static int
C_unattach (void)
{
    struct stat statb;		/* for statting files */
    FILE *afp;                  /* attach file pointer */
    char afname[MAXPARMLEN];    /* attach file name */
    int s;			/* return status */
    
    DEBUG3 ("C_unattach: printer %s", Printer);

    if (!*Attach_file) {
	(void) printf ("Warning: 'attach' feature not active.\n");
	(void) fflush (stdout);
	return False;
    }
    
    q_on_startstop = False;
    q_on_abortpromote = False;
    called_from_unattach = True;
    if ((s = C_abort ())) {
	if (stat (Attach_file, &statb) == 0) {
	    if ((afp = fopen_daemon (Attach_file, "r"))) {
		if (fscanf (afp, "%s", afname) != 1) {
		    fatal (XLOG_INFO, "attach file for printer %s corrupted!", 
			   Printer);
		    (void) fclose (afp);
		    return (False);
		}
	    }
	    (void) fclose (afp);
	    if (unlink_daemon (Attach_file) != 0) {
	        log (XLOG_INFO, "\n  Warning: can't delete file to unattach printer '%s'.",
		     Printer);
		return (False);
	    }
	} else {
	    (void) fprintf (stdout, "  Printer '%s' is not attached.\n", Printer);
	    (void) fflush (stdout);
	    return (False);
	}

	if (s) {
	    if ((RM != NULL) && (*RM != '\0')) { /* It's a local queue */
		(void) fprintf (stdout, "Local   queue  '%s': unattached from '%s'.\n",
				Printer, afname);
	    } else {
		(void) fprintf (stdout, "Local  printer '%s': unattached from '%s'.\n",
				Printer, afname);
	    }
	}
	(void) fflush (stdout);
	/* Don't start it automatically ! */
	/* s = C_start (); */
    }
    
    (void) fflush (stdout);
    q_on_startstop = True;
    q_on_abortpromote = True;
    Displayq (False, False);
    called_from_unattach = False;
    return (s);
}

/***************************************************************************
 * int C_forms (void)
 * changes the forms from a printer to another form
 ***************************************************************************/
static int
C_forms (void) {
    pid_t pid;			/* active server and its pid */
    int active;
    char buf[BUFSIZ];		/* buffer */
    char *possible_forms, *forms;
    FILE *fp;

    DEBUG3 ("C_forms: printer %s", Printer);
    if (Parmcount != 1) {
	(void) printf ("Missing forms specification.\n");
	(void) fflush (stdout);
	return False;
    }
    if (!Forms_file) {
	(void) printf ("Forms feature not active.\n");
	(void) fflush (stdout);
	return False;
    }
    DEBUG4 ("C_forms: '%s'(%d)", Parms[0].str, Parms[0].num);
    forms = Parms[0].str;

    /* get number of jobs in the queue */
    if (Get_pc_entry (Printer, Status_pc_vars, Status_pc_len) == 0) {
	printf ("Printer '%s' does not exist.\n", Printer);
	(void) fflush (stdout);
	return False;
    }
    /* If remote printer return error */
    if ((RM != NULL) && (*RM != '\0')) {
	printf ("Forms command only applies to local printers.\n");
	(void) fflush (stdout);
	return False;
    }

    /* we have a server here */
    if (SS && *SS) {
	if (Get_multi_pc_entry (Printer, SS,
				Status_pc_vars, Status_pc_len) == 0)
	    {
		printf ("Server '%s' queue '%s' does not exist.\n",
			Printer, SS);
		(void) fflush (stdout);
		return False;
	    }
	LO = Printer;
	SV = NULL;
    }

    if (SD == 0 || *SD == 0) {
	log (XLOG_INFO, "Printer '%s' does not have spool directory.", Printer);
	return False;
    }
    chdir_SD ();
    /*
     * start by getting active server information
     */
    buf[0] = '\0';
    active = Checklockfile (LO, &pid, buf, sizeof (buf), &LO_statb);

    /*
     * get numbers of jobs in queue
     */

    Jobcount = Getq ();

    possible_forms = possible_options ();

    if (possible_forms == NULL || !valid_forms (forms, possible_forms)) {
	printf ("%s: '%s' is not a valid form - choose from:\n%s\n", Printer,
		forms, possible_forms ? possible_forms : "(no forms available)");
	(void) fflush (stdout);
	return False;

    } else {
	char *ip;
	char opt_file_name[MAXPATHLEN];

	strcpy (opt_file_name, Forms_file);
	if ((ip = (char *) strchr (Printer, '.')))
	    strcat (opt_file_name, ip);
	fp = fopen_daemon (opt_file_name, "w");
	if (fp == NULL) {
	    logerr (XLOG_INFO, "Couldn't open options file %s",
		    opt_file_name);
	    return False;
	}
	fputs (forms, fp);
	fclose (fp);
	(void) utime (SD, NULL);

	(void) C_start ();
    }

    (void) fflush (stdout);
    return True;
}

/***************************************************************************
 * int C_listforms (void)
 * lists the loaded and possible forms of a printer
 ***************************************************************************/
static int
C_listforms (void) {
    FILE *afp;                  /* attach file pointer */  
    static int att_mark;        /* marker for attached printer */
    char afname[MAXPARMLEN];    /* attach file name */
    char unattached_printer[MAXPARMLEN]; /* Name of the unattached printer */
    char ss[MAXPARMLEN];        /* Save SS */

    extern void Remote_listforms (void);

    DEBUG3 ("C_listforms");   
    att_mark = False;
    /* check for attached printer */
    if ((afp = fopen_daemon (Attach_file, "r"))) { /* Try to open attach file */
        if (fscanf (afp, "%s", afname) == 1) {
            if (strsame (afname, Printer)) {
                fatal (XLOG_INFO, "Printer '%s' attached to itself", Printer);
            }
            (void) strcpy (unattached_printer, Printer);
            Printer = afname;
            att_mark = True;
        }
        fclose(afp);
    }
    if (att_mark) {
        DBGLOCAL3 ("C_listforms: printer %s attached to %s",    
                   unattached_printer, Printer);
    } else {
        DBGLOCAL3 ("C_listforms: printer %s", Printer);
    }
    
    if (Get_pc_entry (Printer, Status_pc_vars, Status_pc_len) == 0) {
        printf ("Printer '%s' does not exist.\n", Printer);
        (void) fflush (stdout);
        return False;
    }
    /* we have a server here */
    if (SS && *SS) {
        (void) strncpy (ss, SS, sizeof (ss));
        if (Get_multi_pc_entry (Printer, SS, Status_pc_vars, Status_pc_len) == 0)
            {
                printf ("Server '%s' queue '%s' does not exist.\n",
                        Printer, SS);
                (void) fflush (stdout);
                return False;
            }
        SV = NULL;
        (void) strncpy (SS, ss, sizeof (ss));
    } else {
        ss[0] = '\0';
    }
    
    if (SD == 0 || *SD == 0) {
        printf ("Printer %s does not have spool directory\n", Printer);
        (void) fflush (stdout);
        return False;
    }
    chdir_SD ();
    if (SA) {
	(void) fputs ("Save area ", stdout);
    } else {
	if (FQDN != Host) {
	    (void) fputs ("Remote ", stdout);
	} else {
	    (void) fputs ("Local  ", stdout);
	}
	if ((SV == 0 || *SV == 0) && (RM == 0 || *RM == 0)) {
	    (void) fputs ("printer ", stdout);
	} else { 
	    (void) fputs (" queue  ", stdout);
	}
    }
    if (att_mark) {
	(void) fprintf (stdout, "'%s'", unattached_printer);
	(void) printf (" [attached to '%s']", Printer);
    } else {
	(void) fprintf (stdout, "'%s'", Printer);
    }
    
    (void) fprintf (stdout, " on %s", Host);
    if (CM && *CM) {
	(void) fprintf (stdout, " (");
	Print_comment (CM);
	(void) fprintf (stdout, ")");
    }
    (void) fprintf (stdout,":\n");
    if (SA) {
	fprintf (stdout, "  No forms available for save queue.\n");
	(void) fflush (stdout);
	return True;
    }
    if (FL && *FL) {
	(void) printf ("  Current forms = %s (possible forms: %s)\n",
		       strsame (current_options (), "none") ? "none" : 
		       current_options (), possible_options ());
    } else {
	fprintf (stdout, "  No forms available.\n");
    }
    if (RM && *RM) {
	DEBUG4 ("C_listforms: getting Remote_listforms ():");
	Remote_listforms ();
    }
    (void) fflush (stdout);
    return True;
}

/***************************************************************************
 * holds the specified jobs in a printerqueue; even if the printer is
 * enabled these jobs will not be printed until they will explicitly
 * released with the release command
 ***************************************************************************/
static int
C_hold (void)
{

    DEBUG3 ("C_hold: printer %s", Printer);
    if (Parmcount != 1) {
	(void) printf ("Missing job specification.\n");
	(void) fflush (stdout);
	return False;
    }
    
    hold ();
    (void) fflush (stdout);
    return True;
}

/***************************************************************************
 * releases specified jobs in a printer queue so that they get printed
 ***************************************************************************/
static int
C_release (void) 
{
    DEBUG3 ("C_release: printer %s", Printer);
    if (Parmcount != 1) {
	(void) printf ("Missing job specification.\n");
	(void) fflush (stdout);
	return False;
    }
    DEBUG4 ("C_release: '%s'(%d)", Parms[0].str, Parms[0].num);

    release ();

    (void) fflush (stdout);
    return True;
}

/***************************************************************************
 * moves the specified job from one printer queue to another printer
 * queue 
 ***************************************************************************/
static int
C_move (void) 
{

    DEBUG3 ("C_move: printer %s", Printer);
    if (Parmcount != 2) {
	(void) printf ("Missing job or destination specification.\n");
	(void) fflush (stdout);
	return False;
    }
    DEBUG4 ("C_move: '%s'(%d) '%s'(%d)", Parms[0].str,
	    Parms[0].num, Parms[1].str, Parms[1].num);


    (void) Checkactive (); /* check to see if the job to be moved is active;
			    * An active job cannot be moved.
			    */
    Jobcount = Getq ();

    move ();
	
    (void) fflush (stdout);
    return True;
}

/***************************************************************************
 * copies the specified job(s) from one printer queue to another printer
 * queue 
 ***************************************************************************/
static int
C_copy (void) 
{

    DEBUG3 ("C_copy: printer %s", Printer);
    if (Parmcount != 2) {
	(void) printf ("Missing job or destination specification.\n");
	(void) fflush (stdout);
	return False;
    }
    DEBUG4 ("C_copy: '%s'(%d) '%s'(%d)", Parms[0].str,
	    Parms[0].num, Parms[1].str, Parms[1].num);


    (void) Checkactive (); /* check to see if the job to be held is active
			    * An active job cannot be copied.
			    */
    Jobcount = Getq ();

    copy ();
	
    (void) fflush (stdout);
    return True;
}

/***************************************************************************
 * requeues the specified job from one printer queue to another printer
 * queue 
 ***************************************************************************/
static int
C_requeue (void) 
{
    Parms[1].str = Printer;

    DEBUG3 ("C_requeue: printer %s", Printer);
    if (Parmcount != 1) {
	(void) printf ("Missing destination specification.\n");
	(void) fflush (stdout);
	return False;
    }

    DEBUG4 ("C_requeue: '%s'(%d) '%s'(%d)", Parms[0].str,
	    Parms[0].num, Parms[1].str, Parms[1].num);
    q_on_startstop = False;
    if (!C_stop () ) {             /* Sets the queue to "no printing" */
	logerr (XLOG_INFO, "C_requeue: stop of queue/printer %s failed.", Printer);
	return False;
    }
    q_on_startstop = True;
    (void) sleep (1);
    requeue ();
    (void) sleep (1);

    if (!mistake) {
	q_on_abortpromote = 2;
	Parmcount = 2;
	Parms[0].num = (int)atoi (Newjobnum);
	Parms[1].str = "\0";
	(void) C_topq ();             /* Puts the job at the top of the queue */
	q_on_abortpromote = True;

	Parmcount = 0;
	Displayq (False, False);
    }
    (void) fflush (stdout);
    return True;
}

/***************************************************************************
 * tries to preprint a specified range of pages of the specified job 
 * returns 1 if successful, otherwise 0
 ***************************************************************************/
static int
C_preprint (void) 
{
    if (Parmcount != 2) {
	(void) printf ("Missing printer or job specification.\n");
	(void) fflush (stdout);
	return False;
    }
    DEBUG4 ("C_preprint: printer %s, job %d", Parms[0].str, Parms[0].num);
    (void) Checkactive (); /* check to see if the job to be moved is active;
			    * An active job cannot be pre-printed.
			    */
    Jobcount = Getq ();
    preprint ();
    return (1);
}

void
default_message_down (char *statstr)
{	
    time_t t_stat;
    char *t_stat_buf;
    
    t_stat = time ((time_t *) NULL);
    malloc_or_die (t_stat_buf, strlen (ctime (&t_stat) + 1));
    t_stat_buf = ctime (&t_stat);
    t_stat_buf[24] = '\0';
    if (SS && *SS) {
	(void) sprintf (statstr, "Server completely closed since %s.", t_stat_buf);
    } else {
	(void) sprintf (statstr, "Queue completely closed since %s.", t_stat_buf);
    }
    return;
}

void
default_message_stop (char *statstr)
{
    time_t t_stat;
    char *t_stat_buf;
    
    t_stat = time ((time_t *) NULL);
    malloc_or_die (t_stat_buf, strlen (ctime (&t_stat) + 1));
    t_stat_buf = ctime (&t_stat);
    t_stat_buf[24] = '\0';	
    (void) sprintf (statstr, "Printing disabled since %s.", t_stat_buf);
    return;
}

void
default_message_disable (char *statstr)
{
    time_t t_stat;
    char *t_stat_buf;
    
    t_stat = time ((time_t *) NULL);
    malloc_or_die (t_stat_buf, strlen (ctime (&t_stat) + 1));
    t_stat_buf = ctime (&t_stat);
    t_stat_buf[24] = '\0';
    if (SS && *SS) {
	(void) sprintf (statstr, "Server closed since %s.", t_stat_buf);
    } else {
	(void) sprintf (statstr, "Queue closed since %s.", t_stat_buf);
    }
    return;
}

/* Not yet ready...*/
static int
preprint (void)
{
    printf ("This function is not yet ready...:-(\n");
    return (1);
}

int
printer_comp (char **l, char **r)
{
    int k;
    k = strcmp (*l, *r);
    return k;
}

#ifndef DEFAULT_PRINTER
#define DEFAULT_PRINTER	"lp"
#endif

int 
Get_Default_Printer (int all) {
    int i;
    char *printer;
    
    if ((printer = getenv ("PRINTER")) != NULL) {
	DEBUG4 ("Get_Printer: PRINTER='%s'", printer);
	if (strcmp (Printer, printer) == 0) {
	    if (Get_pc_entry (Printer, Status_pc_vars, Status_pc_len) == 0) {
		printf ("Printer '%s' does not exist.\n", Printer);
		(void) fflush (stdout);
		return False;
	    }
	    return (1);
	}
    } else if
	((i = Get_pc_entry (DEFAULT_PRINTER,
			    All_pc_vars, All_pc_len)) != 0)
	{
	    printer = DEFAULT_PRINTER;
	    DEBUG4 ("Get_Printer: default '%s'", printer);
	    if (strcmp (Printer, printer) == 0) {
		if (Get_pc_entry (Printer, Status_pc_vars, Status_pc_len) == 0) {
		    printf ("Printer '%s' does not exist.\n", Printer);
		    (void) fflush (stdout);
		    return False;
		}
		return (1);
	    }
	} else if (Use_first_printer) {
	    /* use the default, first in list */
	    printer = First_printer ();
	    DEBUG4 ("Get_Printer: First_printer '%s'", printer);
	    if (strcmp (Printer, printer) == 0) {
		if (Get_pc_entry (Printer, Status_pc_vars, Status_pc_len) == 0) {
		    printf ("Printer '%s' does not exist.\n", Printer);
		    (void) fflush (stdout);
		    return False;
		}
		return (1);
	    }
	} else {
	    Diemsg ("no printer specified");
	}
    if (Get_pc_entry (Printer, Status_pc_vars, Status_pc_len) == 0) {
        printf ("Printer '%s' does not exist.\n", Printer);
	(void) fflush (stdout);
	return False;
    }
    return (0);
}

int 
check_if_printer_is_attached (void)
{
    FILE *afp;
    char afname[MAXPARMLEN];    /* attach file name */
 
    (void) sprintf (afname, "%s/%s", SD, Attach_file);
    if ((afp = fopen_daemon (Attach_file, "r"))) { /* Try to open attach file */
	/* printer is attached */
	if (fscanf (afp, "%s", afname) == 1) {
	    if (strsame (afname, Printer)) {
		fatal (XLOG_INFO, "Printer '%s' attached to itself", Printer);
	    }
	    (void) strcpy (attached_printer, afname);
	}
	fclose(afp);
	return (1);
    }
    /* printer not attached, return 0 */
    return (0);
}
