/***************************************************************************
 * U. Minnesota LPD Software * Copyright 1987, 1988, Patrick Powell
 ***************************************************************************
 * MODULE: Checkperm.c
 * printer permission checking
 ***************************************************************************
 * Checkperm - determine if a user has permissions to act on a printer queue
 *
 * Input:
 *  permfile - name of the permissions file,
 *    each line of which has the form:
 *		hostname userid printer_name op pr max current
 * 		hostname - the host name of the invoker
 * 		userid - the user name of the invoker
 * 		printer_name - the printer queue to be checked
 *		op -the maximum operation level, A is 'Highest', Z is 'lowest'
 *			Note: R is for LPR, M is for move job,
 *                            C is for LPC, * allows all
 *		pr - the maximum priority level, A is 'Highest', Z is 'lowest'
 *		max - max number of pages allowed
 *		current - current number of pages
 *
 *	# -- comment
 *      #hostname username groupname printername op pr maxpages currentpages
 *      attila    papowell *         imagen      C  A  1000     100
 *      !*        cs9*     *         lpr         *  *
 *      !*        *        cs4       imagen      *  *
 *      *         jsmith   *         imagen      *  *  0        0
 *      *         root     *         *           *  *
 *
 *	Note: * is the wildcard indicator and matches all characters
 *	Note: if maxpages == 0, then page count is unlimited.
 *  Note: if a ! is in the first column  and there is a match
 *    (i.e.- we should grant permission), then permission is refused.
 *    In the example above, this prevents users whose names start with
 *    cs9 will be forbidden access to the lpr queue.
 *
 * Output:
 * 0 - if userid does not have permissions to act on printer queue
 * nonzero - permission granted
 * <0 - permission barred
 *
 * If any of arguments is not specified (i.e.- 0 or NULL value),
 * then checking is not done for the field.
 *
 * The currentpages field is going to be updated every time a printing is
 * done. This is responsibility of the filter which will call the addpages
 * script to do the update. The prmsline is the ordinal number of the 
 * printer perms file to be incremented. It is passed to the update script
 * to be used when a regular expression entry is to be charged for the
 * printout.
 *****************************************************************************/

#include "lp.h"
#include "library/errormsg.h"
#include "library/filestack.h"
#include "library/filecache.h"
#include "library/utils.h"

#ifdef NIS
#include <rpc/rpc.h>
#include <rpcsvc/ypclnt.h>
#include <rpcsvc/yp_prot.h>
extern int innetgr ();
#endif

/*
 * match( str, pattern )	   returns: 1 = matches, 0 = match failed
 * 
 * null pattern is equivalent to '*'.
 * 
 * -- str is null, cannot match -- ignore (ie. pass) (match returns 1)
 * -- pattern is <fixed>*, str is <fixed><anything> (match returns 1)
 * -- pattern is !<fixed>*, str is <fixed><anything> (match returns 0)
 * -- if not above, match returns 0
 */

static int
match (char *str, char *pattern) {
    int c;

    DBGPERMS9 ("match: match str '%s', pattern '%s'", str, pattern);
    if (str == NULL) {
	DBGPERMS8 ("match success");
	return True;
    } else {
	if ((pattern == NULL) || strsame (pattern, "*")) {
	    DBGPERMS8 ("match: match success");
	    return True;
	}
	/* else fall through... */
    }

    while ((c = *pattern) != '\0') {
	DBGPERMS9 ("match: match partial str '%s', pattern '%s'", str, pattern);
	switch (c) {
	    /*
	     * match 0 or more characters in the string
	     */
	case '*':
	    if (pattern[1] == '\0') {
		return True;	/* AJCD */
	    }
	    if (match (str, pattern + 1)) {
		return True;
	    }
	    if (*str && match (str + 1, pattern)) {
		return True;
	    }
	    return False;
	case '?':
	    if (*str == 0) {
		return False;
	    }
	    break;
	default:
	    if (c != *str) {
		return False;
	    }
	    break;
	}
	++pattern;
	++str;
    }
    return (*pattern == *str);
}

static inline void
strcpy_nocase (char *to, char *from) {
    char *x, *y;
    for (x = from, y = to; *x; x++, y++) {
	if (isalpha (*x) && isupper (*x)) {
	    *y = *x + 040;
	} else {
	    *y = *x;
	}
    }
    *y = '\0';
}

/* these are needed for hostnames and ip-addresses. */

static inline char *charstar_ntoa (char *bytes) {
    static char str[128];

#define IpOctet(x)	((unsigned char) bytes[x])

    (void) sprintf (str, "%d.%d.%d.%d", IpOctet(0), IpOctet(1), IpOctet(2), IpOctet(3));
    return (str);
}

static int
match_ipaddresses (char *strp, char *patternp) {
    struct hostent *hent;
    char **addr;			/* ptr to h_addr list */

    DBGPERMS9 ("match_ipaddresses: match str '%s', pattern '%s'", strp, patternp);
    if (strp == NULL) {
	DBGPERMS8 ("match success");
	return True;
    } else {
	if ((patternp == NULL) || strsame (patternp, "*")) {
	    DBGPERMS8 ("match_ipaddresses: match success");
	    return True;
	}
	/* else fall through... */
    }

    /* if we can't resolve the name, just do the string match. */
    if ((hent = gethostbyname (strp)) == NULL)
	return (match (strp, patternp));

    /* cycle through the list of the host's addresses. */
    for (addr = (char **) hent->h_addr_list; *addr != NULL; addr++) {
	if (match (charstar_ntoa (*addr), patternp))
	    return True;
    }
    return False;
}

static int
match_hostnames (char *strp, char *patternp) {
    struct hostent *hent;
    char **alias;			/* ptr to h_aliases list */
    /*@@@ unsigned long addr; *//* inet_addr has an unsigned long return 
                                 * value; but in error cases it returns -1! */
    long addr;

    DBGPERMS9 ("match_hostnames: match str '%s', pattern '%s'", strp, patternp);
    if (strp == NULL) {
	DBGPERMS8 ("match success");
	return True;
    } else {
	if ((patternp == NULL) || strsame (patternp, "*")) {
	    DBGPERMS8 ("match_hostnames: match success");
	    return True;
	}
	/* else fall through... */
    }

    /* if we can't resolve the address, just do the string match. */
    addr = inet_addr (strp);
    if ((addr == -1) ||
	    ((hent = gethostbyaddr ((char *) &addr,
	    IP_ADDRESS_LEN, AF_INET)) == NULL))
    {
	return (match (strp, patternp));
    }

    /* check the host's canonical name. */
    if (match (hent->h_name, patternp))
	return True;

    if (hent->h_aliases) {
	/* cycle through the list of the host's aliases */
	for (alias = hent->h_aliases; *alias != NULL; alias++) {
	    if (match (*alias, patternp))
		return True;
	}
    }
    return False;
}

static int
match_host (char *str, char *pattern) {
    char str_lc[MAXPARMLEN + 1];
    char pattern_lc[MAXPARMLEN + 1];
    char *strp, *patternp;
    int ipaddr_flag_set = 0;

    if (str == NULL) {
	strp = NULL;

    } else if (isdigit (*str)) {
	ipaddr_flag_set |= 1;	/* 1st bit: str is ip-address */
	strp = str;

    } else {
	strcpy_nocase (str_lc, str); strp = str_lc;
    }

    if (pattern == NULL) {
	patternp = NULL;

    } else if (isdigit (*pattern)) {
	ipaddr_flag_set |= 2;	/* 2nd bit: pattern is ip-address */
	patternp = pattern;

    } else {
	strcpy_nocase (pattern_lc, pattern); patternp = pattern_lc;
    }

    if (ipaddr_flag_set == 1) {		/* str is ip-address, pat isn't */
	return (match_hostnames (strp, patternp));
	
    } else if (ipaddr_flag_set == 2) {	/* pat is ip-address, str isn't */
	return (match_ipaddresses (strp, patternp));

    } else {
	/* either both are ip-addrs or both are strings. */
	return (match (strp, patternp));
    }
}

static char *fq_host;
static void make_fq_host (char *host) {
    if (host == NULL) {
	return;
    }
    if (strchr (host, '.') == NULL) {
	fq_host = NULL;
    } else {
	/* this'll work for IP-addresses too! */
	fq_host = strdup (host);
    }
}

/*
 * scan_perms(char *buf) scans a line of permissions and returns status:
 * -2 deny (out of pages)	-1 deny 	1 allow		0 neither
 */
static int
scan_perms (char *buf, char *host, char *user, char *groups, char *printerq, int *op, int *pr_level, int pages) {
    char hostname[MAXPARMLEN + 1];
    char username[MAXPARMLEN + 1];
    char groupname[MAXPARMLEN + 1];
    char printerqname[MAXPARMLEN + 1];
    char operation[MAXPARMLEN + 1];
    char priority[MAXPARMLEN + 1];
    int max, current;
    char *str;
    int i;
    int invert, must_match;
    int grp_found;

    must_match = invert = False;
    priority[0] = '\0';
    operation[0] = '\0';
    current = max = 0;

    str = buf;
    switch (*buf) {
    case '!':
	invert = True;
	++str;
	break;
    case '~':
	must_match = True;
	++str;
	break;
    default:
	break;
    }

    /* sscanf returns number of fields converted */
    i = sscanf (str, "%s%s%s%s%s%s%d%d", hostname, username, groupname,
		printerqname, operation, priority, &max, &current);
    if (i == 0) {
	return False;
    } else if (i < 4) {
	fatal (XLOG_CRIT, "Error in perms file! line='%s'", buf);
	return False;
    }

    DBGPERMS6 ("scan_perms: host=%s,user=%s,group=%s,printer=%s,oper=%s,priority=%s,max%d,current%d",
	hostname, username, groupname, printerqname, operation, priority, max, current);

    /* do the simple string tests before any complicated tests that may
     * involve system calls -- this speeds things up.
     */
    if (!match (printerq, printerqname)) {
	DBGPERMS6 ("scan_perms: printerq bad, %s, %s",
		 printerq ? printerq : "<NONE>", printerqname);
	if (must_match)
	    return -1;
	return False;
    }
    DBGPERMS6 ("scan_perms: printerq ok, %s, %s",
	     printerq ? printerq : "<NONE>", printerqname);

    /* compare the operations */
    if (op && operation[0] != '*' && *op < operation[0]) {
	DBGPERMS6 ("scan_perms: op bad, %c, %c",
		 op ? *op : '?', operation[0]);
	if (must_match)
	    return -1;
	return False;
    }
    DBGPERMS6 ("scan_perms: op ok, %c, %c", op ? *op : '?', operation[0]);

    /*
     * if the hostname field is '*', this means any host can do it otherwise we should
     * check to see if a valid host name. If the first character of the hostname is '@',
     * we check if the host is in the netgroup specified.
     */
#ifdef NIS
    if (hostname[0] == '@') {	/* look up machine in netgroup */
	if (!host || !*host || !innetgr (hostname + 1, host, NULL, NULL)) {
	    DBGPERMS6 ("scan_perms: host %s not in netgroup %s",
			host && *host ? host : "<NULL>", hostname + 1);
	    if (must_match)
		return -1;
	    return False;
	}
    } else
#endif /* NIS */

    if (host && *host) {
	if (fq_host == NULL) {
	    char *domain;

	    /* if there's a domain in the hostname field of the perms
	     * file, and we've got an unqualified hostname in the control
	     * file, try appending the domain to the unqualified hostname
	     * and testing it with gethostbyname(). If ghbn works, we'll
	     * append the domain and pass it on to the matching routine;
	     * otherwise, we'll treat it as a failure.
	     */

	    if ((*hostname < '0' || *hostname > '9') &&
		((domain = strchr (hostname, '.')) != NULL))
	    {
		if ((fq_host = host_in_domain (host, domain)) != NULL) {
		    host = fq_host;

		} else {
		    /* the perms file has a domain, but the unqualified
		     * hostname is not in this domain -> not a match.
		     */
		    free (fq_host);
		    fq_host = NULL;
		    DBGPERMS6 ("scan_perms: host bad, %s, %s",
			     host ? host : "<NONE>", hostname);
		    if (must_match)
			return -1;
		    return False;
		}
	    }
	} else {
	    host = fq_host;
	}
	if (fq_host) {
	    assert (*host == *fq_host);		/* don't work too hard */
	}
    }
    if (hostname[0] != '@' && !match_host (host, hostname)) {
	DBGPERMS6 ("scan_perms: host bad, %s, %s",
		host ? host : "<NONE>", hostname);
	if (must_match)
	    return -1;
	return False;
    }
    DBGPERMS6 ("scan_perms: host ok, %s, %s", host ? host : "<NONE>", hostname);

    /*
     * compare the user names. If first character of username is '@', test if in
     * netgroup.
     */
#ifdef NIS
    if (username[0] == '@') {	/* look up user in netgroup */
	if (!user || !*user || !innetgr (username + 1, NULL, user, NULL)) {
	    DBGPERMS6 ("scan_perms: user %s not in netgroup %s",
		     user && *user ? user : "<NONE>", username + 1);
	    if (must_match)
		return -1;
	    return False;
	}
    } else
#endif /* NIS */
    if (!match (user, username)) {
	DBGPERMS6 ("scan_perms: user bad, %s, %s",
		    user ? user : "<NONE>", username);
	if (must_match)
	    return -1;
	return False;
    }
    if (invert && user == NULL && !strsame (username, "*")) {
	/* no username was specified, a distinct pattern was
	 * used, and we're to invert the results. This would result
	 * in ALL USERS being refused, instead of just the named
	 * user, so invert the test.
	 */
	DBGPERMS6 ("scan_perms: user bad, %s, %s",
		    user ? user : "<NONE>", username);
	if (must_match)
	    return -1;
	return False;
    }
    DBGPERMS6 ("scan_perms: user ok, %s, %s",
	     user ? user : "<NONE>", username);

    /* compare the group names */
    if (groups[0]) {
	grp_found = False;
	while (*groups && !grp_found) {
	    grp_found += match (groups, groupname);
	    if (!grp_found) {
		groups += strlen (groups) + 1;
	    }
	}
    } else {
	/* we cannot say anything about the group. If the pattern
	 * is a straight-out wildcard, pass this check; otherwise,
	 * fail it.
	 */
	if (strsame (groupname, "*")) {
	    grp_found = True;
	} else {
	    grp_found = False;
	}
    }
    if (!grp_found) {
	DBGPERMS6 ("scan_perms: group bad, %s, %s",
		     groups[0] ? groups : "<NONE>", groupname);
	if (must_match)
	    return -1;
	return False;
    }
    DBGPERMS6 ("scan_perms: group ok, %s, %s",
	     groups[0] ? groups : "<NONE>", groupname);

    DBGPERMS6 ("scan_perms: OK, must_match %d, wanted %c(%d), have %c(%d), invert %d",
	     must_match, op ? *op : '?', operation[0], op ? *op : '?',
	     operation[0], invert);

    /*
     * Well, you had to have a match, and you did.  So this means that you passed, but
     * you do not use this entry.
     */
    if (must_match)
	return False;
    DBGPERMS6 ("scan_perms: pages- %d, max %d, current %d", pages, max, current);

    /* So far, so good. Now a couple of final checks before allowing */
    if (pages && max && current > max) {
	return -2;			/* page limit exceeded */
    }

    if (pr_level && priority[0] && priority[0] != '*') {
	/* we want to check priory and there are limits */
	if (priority[0] > *pr_level) {
	    /* sorry, this is too high */
	    *pr_level = priority[0];
	}
    }
    /* the final hurdle; did it allow or deny you? */
    if (invert)
	return -1;
    return True;
}

#ifdef NIS
static int
perms_nis (char *outval, char *host, char *user, char *group, char *printerq, int *op, int *pr_level, int pages) {
    char *np;
    int permission;

    if (outval == 0) {
	logerr (XLOG_INFO, "perms_nis: no NIS match");
	return False;
    }
    DBGPERMS5 ("perms_nis: Checking NIS perms db");

    make_fq_host (host);

    /*
     * scan through the list looking for a match; outval is a string consisting
     * of newline-separated entries
     */
    permission = False;
    while (permission == 0 && outval && *outval) {
	if ((np = strchr (outval, '\n')))
	    *np++ = '\0';	/* terminate current entry */

	if (*outval && (*outval == '\n' || *outval == '#')) {
	    outval = np;
	    continue;
	}
	permission = scan_perms (outval, host, user, group, printerq,
				 op, pr_level, pages);
	outval = np;
    }

    DBGPERMS4 ("perms_nis: perms_nis %d, printerq %s for %s@%s",
	     permission,
	     printerq ? printerq : "NONE", user ? user : "NONE", host ? host : "NONE");

    return (permission);
}

/*
 * Checkperm_NIS checks the NIS database for permissions entries The keys
 * "printer_perms.printerq", "printer_perms.host" and "printer_perms" are tried until a
 * match is found.
 */
static int
Checkperm_NIS (char *host, char *user, char *group, char *printerq,
	int *op, int *pr_level, int pages)
{
    int res = False;		/* result */
    char *np;			/* next character pointer */
    char *outval;
    int outvallen;
    char *stddomain;
    char buf[YPMAXMAP];		/* NIS map */

    if (yp_get_default_domain (&stddomain))
	return False;

    /* Try NIS lookup for permissions */
    (void) strcpy (buf, "printer_perms");
    np = buf + strlen (buf);

    /* tag on queue name */
    *np = PNAME_SEPCH;
    (void) strcpy (np + 1, printerq);

    if (!NIS_printcap_byname || !*NIS_printcap_byname)
	logerr_die (XLOG_CRIT, "No NIS printcap byname map!");

    if (yp_match (stddomain, NIS_printcap_byname, buf, strlen (buf),
		  &outval, &outvallen) == 0) {
	res = perms_nis (outval, host, user, group, printerq, op,
			 pr_level, pages);
	free (outval);
    }
    if (!res) {
	/* tag on host name */
	(void) strcpy (np + 1, ShortHost);

	if (yp_match (stddomain, NIS_printcap_byname, buf, strlen (buf),
		      &outval, &outvallen) == 0) {
	    res = perms_nis (outval, host, user, group, printerq,
			     op, pr_level, pages);
	    free (outval);
	}
    }
    if (!res) {
	/* try just printer_perms */
	*np = '\0';

	if (yp_match (stddomain, NIS_printcap_byname, buf, strlen (buf),
		      &outval, &outvallen) == 0) {
	    res = perms_nis (outval, host, user, group, printerq,
			     op, pr_level, pages);
	    free (outval);
	}
    }
#ifndef YP_GET_DOMAIN_BROKEN
    free (stddomain);
#endif /* NOT defined YP_GET_DOMAIN_BROKEN */
    return (res);
}

#endif /* NIS */

#ifdef HESIOD
static int
perms_hes (char **hesptr, char *host, char *user, char *group, char *printerq, int *op, int *pr_level, int pages) {
    int permission;

    if (hesptr == 0) {
	logerr (XLOG_INFO, "perms_hes: no Hesiod match");
	return False;
    }
    DBGPERMS5 ("perms_hes: Checking Hesiod perms db");

    make_fq_host (host);

    /*
     * scan through the entries looking for a match
     */
    permission = False;
    while (permission == False && *hesptr != NULL) {
	if (*hesptr[0] == '\n' || *hesptr[0] == '#') {
#ifdef HESIOD_FREE
	    (void) free (*hesptr++);
#else
	    hesptr++;
#endif /* HESIOD_FREE  */
	    continue;
	}
	permission = scan_perms (*hesptr, host, user, group, printerq,
				 op, pr_level, pages);
#ifdef HESIOD_FREE
	(void) free (*hesptr++);
#else
	hesptr++;
#endif /* HESIOD_FREE */
    }
#ifdef HESIOD_FREE
    while (*hesptr)		/* free remaining pointers */
	(void) free (*hesptr++);
#endif /* HESIOD_FREE */

    DBGPERMS4 ("perms_hes: perms_hes %d, printerq %s for %s@%s",
	     permission,
	     printerq ? printerq : "NONE", user ? user : "NONE", host ? host : "NONE");

    return (permission);
}

/*
 * Checkperm_hes checks the Hesiod database for permissions entries The keys
 * "printer_perms.printerq", "printer_perms.host" and "printer_perms" are tried until a
 * match is found
 */
static int
Checkperm_hes (char *host, char *user, char *group, char *printerq,
	int *op, int *pr_level, int pages) {
    int res = False;		/* result */
    char **hp;			/* hesiod match pointer */
    char *np;			/* next character pointer */
    char buf[HES_MAXKEYLEN];	/* Hesiod key */

    /* Try Hesiod lookup for permissions */
    (void) strcpy (buf, "printer_perms");
    np = buf + strlen (buf);

    /* tag on queue name */
    *np = PNAME_SEPCH;
    (void) strcpy (np + 1, printerq);

    if (!Hesiod_printcap_key || !*Hesiod_printcap_key) {
	logerr_die (XLOG_CRIT, "hesiod-printcap-key not set!");
    }
    if ((hp = hesiod_match (buf, Hesiod_printcap_key, NULL))) {
	res = perms_hes (hp, host, user, group, printerq, op,
			 pr_level, pages);
    }
    if (!res) {
	/* tag on host name */
	(void) strcpy (np + 1, ShortHost);

	if ((hp = hesiod_match (buf, Hesiod_printcap_key, NULL))) {
	    res = perms_hes (hp, host, user, group, printerq, op,
			     pr_level, pages);
	}
    }
    if (!res) {
	/* try just permissions */
	*np = '\0';

	if ((hp = hesiod_match (buf, Hesiod_printcap_key, NULL))) {
	    res = perms_hes (hp, host, user, group, printerq, op,
			     pr_level, pages);
	}
    }
    return (res);
}

#endif /* HESIOD */

#ifdef LOCAL

static struct file_cache perm_cache;
static int reading_perm_from_cache;

static char *
perm_entry_file (char *buf, int size, FILE *permfile) {
    if (!Use_perm_cache) {
	DBGPERMS9 ("perm_entry_file: perms caching disabled");
	assert (permfile != NULL);
	return (Gets (buf, size, &permfile, Permfile_include_path));

    } else if (!reading_perm_from_cache) {
	/* a per-printer permfile -- we don't cache these */
	assert (permfile != NULL);
	return (Gets (buf, size, &permfile, Permfile_include_path));
    }

    if (perm_cache.size == 0) {
        fatal (XLOG_NOTICE, "perm cache is unset!");
    }

    assert (perm_cache.entries != NULL);
    if (perm_cache.entries[perm_cache.ind] == NULL) {
        /* there's nothing more to read */
        rewind_cache (&perm_cache);
        return NULL;
    }

    (void) strncpy (buf, perm_cache.entries[perm_cache.ind], size - 1);
    
    perm_cache.ind++;

    return buf;
}

static int
perms_file (FILE *permfile, char *host, char *user, char *group, char *printerq, int *op, int *pr_level, int pages) {
    int permission;
    char buf[BUFSIZ];
 
    DBGPERMS5 ("perms_file: Checking perms files");
    make_fq_host (host);

    /*
     * scan through the file looking for a match
     */
    permission = False;
    prmsline = 0;

    while (permission == False &&
	perm_entry_file (buf, sizeof (buf), permfile) != NULL)
    {
	/* prmsline gets incremented every time a new line is read from
	 * Permsfile. If it is nonzero then that line of the printer_perms
	 * file should be incremented by the accounting script
	 */
	prmsline++;
	permission = scan_perms (buf, host, user, group, printerq,
				 op, pr_level, pages);
    }
    DBGPERMS4 ("perms_file: perms_file %d, printerq %s for %s@%s",
	    permission, printerq ? printerq : "NONE",
	    user ? user : "NONE", host ? host : "NONE");

    if (permfile != NULL) {
	(void) fclose (permfile);
    }

    return (permission);
}

/*
 * Checkperm_file checks the XU variable and then the permfile for permissions
 */
static int
Checkperm_file (char *host, char *user, char *group, char *printerq, int *op, int *pr_level, int pages) {
    int res = False;		/* result of first check */
    char *pf = XU;
    FILE *permfile;

    reading_perm_from_cache = False;
    if (pf && *pf) {
	char buf[MAXPATHLEN];
	DBGPERMS3 ("checkperm: reading from XU perms file");
	if (*pf != '/') {	/* make sure we've got a full path */
	    (void) sprintf (buf, "%s/%s", SD, pf);
	    pf = buf;
	}
	if ((permfile = fopen (pf, "r")) != NULL) {
	    res = perms_file (permfile, host, user, group, printerq,
			      op, pr_level, pages);
	} else {
	    logerr (XLOG_WARNING, "cannot open XU file '%s'", pf);
	}
    }
    if (!res) {
	if (Use_perm_cache && perm_cache.entries && perm_cache.entries[0]) {
	    /* we have a cache, use it. */
	    DBGPERMS5 ("checkperm: reading perms from cache");
	    reading_perm_from_cache = True;
	    rewind_cache (&perm_cache);
	    permfile = NULL;

	} else {
	    DBGPERMS5 ("checkperm: reading perms from file");
	    if ((permfile = fopen_path (Permfile_path, "r")) == NULL) {
		fatal (XLOG_INFO, "failed to open permissions file: %s", Errormsg(errno));
	    }
	}
	res = perms_file (permfile, host, user, group, printerq,
			  op, pr_level, pages);
    }
    return (res);
}

void build_perm_cache (void) {
    char buf[BUFSIZ];
    FILE *permfile;
    struct file_cache *perm_cache_ref;

    if (!Use_perm_cache) {
	DBGPERMS6 ("build_perm_cache: perms caching disabled");
	return;
    }

    perm_cache_ref = &perm_cache; 
    chkandrealloc_cache (&perm_cache_ref);
    flush_cache (&perm_cache);

    if ((permfile = fopen_path (Permfile_path, "r")) == NULL) {
	fatal (XLOG_INFO, "failed to open permissions file: %s",
		Errormsg(errno));
    }

    while (Gets (buf, sizeof (buf), &permfile, Permfile_include_path) != NULL)
    {
	DBGPERMS7 ("build_perm_cache: caching perm entry %d [%s]", 
		   perm_cache.ind, buf);
	perm_cache_ref = &perm_cache; 
	chkandrealloc_cache (&perm_cache_ref);
	cache_addline (&perm_cache, buf);
    }
    cache_addeof (&perm_cache);
    rewind_cache (&perm_cache);
    fclose (permfile);
}

void
createPermsCache (void) {
    if (!Use_perm_cache) {
	DBGPERMS6 ("createPermsCache: perms caching disabled");
        return;
    }
    perm_cache.size = 0;
    build_perm_cache ();
}

void
flush_perm_cache (void) {
    if (!Use_perm_cache) {
	return;
    }
    DBGPERMS2 ("flush_perm_cache: flushing perm cache");
    build_perm_cache ();
}

#endif /* LOCAL */

/*
 * Checkperm checks the permissions according to the current configuration. Files,
 * Hesiod, and NIS are consulted until a deny or allow is found. Only local domains are
 * searched for permissions.
 */
int
Checkperm (char *host, char *user, char *printerq, int *op, int *pr_level, int pages) {
    int res = False;
    int i;
    char groups[NGROUPS_MAX * GRP_NAM_MAX];	/* groupnames to check for */

#if (defined(NIS) || defined(HESIOD)) && defined(CROSS_DOMAIN)
    char qname[MAXPARMLEN];	/* printer name plus domain */

    if (out_domain && printerq) {
	(void) strcpy (qname, printerq);	/* add "@domain" to printer name */
	(void) strcat (qname, "@");
	(void) strcat (qname, out_domain);
	printerq = qname;
    }
#endif /* (defined(NIS) || defined(HESIOD)) && defined(CROSS_DOMAIN) */

    groups[0] = '\0';
    if (user) {			/* AJCD: get user's group for use in perms file */
	struct passwd *pwent;
	struct group *grent;
	char *gp = groups;

/*	lower_case (user); */	/* make canonical for accounting */
	if (((pwent = getpwnam (user)) != NULL) &&
		((grent = getgrgid (pwent->pw_gid)) != NULL))
	{
	    (void) strcpy (gp, grent->gr_name);
	    gp += strlen (grent->gr_name) + 1;		/* primary group */
	    while ((grent = getgrent ())) {
		char **mem;
		for (mem = grent->gr_mem; mem && *mem; mem++) {
		    if (strsame (*mem, user) &&
				((gp - groups) < (NGROUPS_MAX - 1) * GRP_NAM_MAX))
		    {
			(void) strcpy (gp, grent->gr_name);
			gp += strlen (grent->gr_name) + 1;
		    }
		}
	    }
	    (void) endgrent ();
	    *gp = '\0';
	}
    }

    if (host) {
	lower_case (host);	/* make canonical for accounting */
    }
 
    DBGPERMS5 ("Checkperm: checking perms for: host=%s,user=%s,group=%s,printerq=%s,op=%c,priority=%c,pages=%d",
	     host ? host : "NONE", user ? user : "NONE",
	     groups[0] ? groups : "NONE", printerq ? printerq : "NONE",
	     op ? *op : '?', pr_level ? *pr_level : '?', pages);

    for (i = 0; i < num_confs && res == False; i++)
	if (conf[i].domain == NULL)	/* only check in local domain */
	    switch (conf[i].type) {
#ifdef LOCAL
	    case LOCAL_TYPE:
		res = Checkperm_file (host, user, groups,
				      printerq, op, pr_level, pages);
		break;
#endif /* LOCAL */
#ifdef NIS
	    case NIS_TYPE:
		res = Checkperm_NIS (host, user, groups,
				     printerq, op, pr_level, pages);
		break;
#endif /* NIS */
#ifdef HESIOD
	    case HESIOD_TYPE:
		res = Checkperm_hes (host, user, groups,
				     printerq, op, pr_level, pages);
		break;
#endif /* HESIOD */
	    default:
		log (XLOG_INFO, "Cannot lookup permissions of type %d.",
			conf[i].type);
	    }

    return (res);
}
