/***************************************************************************
 * U. Minnesota LPD Software * Copyright 1987, 1988, Patrick Powell
 ***************************************************************************
 * MODULE: Errormsg.c
 * error messages and logging
 * log:
 *  --message on stderr and on stdout if "Echo_on_stdout" is set
 *  --if the severity is high enough, will also log using syslog().
 *  --saves the last "non-Debug" log message
 * logerr:
 *  --formats the error message corresponding to "errno" and calls log();
 * fatal:
 *  -- log() and exit() with Errorcode
 * logerr_die:
 *  -- logerr() and exit() with Errorcode
 **************************************************************************/

#include "lp.h"
#include "library/errormsg.h"

#define LOG_BUFSIZ	(BUFSIZ * 2)

static void use_syslog (int, char *);

/****************************************************************************/

#ifdef HAVE_SYS_NERR
#define num_errors()	(sys_nerr)
extern int sys_nerr;
#else
#define num_errors()	(35)		/* conservative estimate */
#endif

/****************************************************************************/

#ifdef HAVE_STRERROR
#ifndef	strerror			/* could be a macro */
extern char *strerror ();
#endif
#else

#ifdef HAVE_SYS_ERRLIST
extern char *sys_errlist[];
#define strerror(err)	(sys_errlist[err])
#else

#undef num_errors
#define num_errors()	(-1)		/* always use "errno=%d" */
#endif /* !HAVE_SYS_ERRLIST */
#endif /* !HAVE_STRERROR */

/****************************************************************************/

/*
 * Errormsg(err) returns a printable form of the errormessage corresponding to errno.
 */

char *
Errormsg (int err) {
    char *cp;
    static char msgbuf[16];	/* holds "errno=%d". */

    if (err >= 0 && err <= num_errors()) {
	cp = strerror (err);
    } else {
	(void) sprintf (msgbuf, "errno=%d", err);
	cp = msgbuf;
    }
    return (cp);
}

struct msgkind {
    int *var;
    char *str;
};

static struct msgkind msg_name[] = {
    {&XLOG_CRIT, " (CRIT)"},
    {&XLOG_ERR, " (ERR)"},
    {&XLOG_WARNING, " (WARN)"},
    {&XLOG_NOTICE, " (NOTICE)"},
    {&XLOG_INFO, " (INFO)"},
    {&XLOG_DEBUG, ""},
    {0}
};

static
char *logmsg (int kind) {
    int i;
    static char b[35];

    for (i = 0; msg_name[i].var; ++i) {
	if (*msg_name[i].var == kind) {
	    return (msg_name[i].str);
	}
    }
    (void) sprintf (b, "bad message type %d", kind);
    return (b);
}

#ifdef HAVE_SYSLOG_H
#include <syslog.h>
#else
static FILE *f;
#endif

/* we cut down the message to avoid attacks based on 8lgm
 * advisory 22 (UNIX syslog internal buffer). 512 bytes will
 * cover everything -- it's a pretty long line anyway!
 */
#define SYSLOG_BUFSIZ	512

void
safe_syslog (int kind, char *realmsg) {
    char msg[SYSLOG_BUFSIZ + 1];
    strncpy (msg, realmsg, SYSLOG_BUFSIZ);
    msg[SYSLOG_BUFSIZ] = '\0';
    (void) syslog (kind, msg);
}

static int logOpen = False;

static void
use_syslog (int kind, char *msg) {
    DEBUG7 ("use_syslog: syslog: %s\n", msg);

    /* testing mode indicates that this is not being used
     * in the "real world", so don't get noisy. */
    if (Testing_mode)
	return;

#ifndef HAVE_SYSLOG_H
    if (f != NULL || (f = fopen ("/dev/console", "w")) != NULL) {
	(void) fprintf (f, "syslog: LPD %s\n", msg);
	(void) fflush (f);
    }

#else				/* HAVE_SYSLOG_H */
#if defined(LOG_PID) && defined(LOG_NOWAIT)
    /* assume we can use openlog(). */
    if (logOpen == False) {
	if (Name && *Name) {
	    openlog (Name, LOG_PID | LOG_NOWAIT, SYSLOG_FACILITY);
	} else {
	    /* guess lpd. */
	    openlog ("lpd", LOG_PID | LOG_NOWAIT, SYSLOG_FACILITY);
	}
	logOpen = True;
    }
    safe_syslog (kind, msg);

#else
    safe_syslog (SYSLOG_FACILITY | kind, msg);
#endif
#endif				/* HAVE_SYSLOG_H */
}

static char log_buf[LOG_BUFSIZ];

static void
log_backend (int kind) {
    static char stamp_buf[256];
    static char stamp_and_log[LOG_BUFSIZ];
    unsigned int stamp_and_log_len;
    int tmp_errno;

    tmp_errno = errno;
    plp_block_signals ();
    /*
     * syslogd does most of the things we do after this point, so there's no point in
     * filling up the logs with redundant timestamps; log the message here.
     */
    if (kind <= XLOG_INFO)
	use_syslog (kind, log_buf);

    /*
     * Now build up the state info: timestamp, hostname, argv[0], pid, etc.
     * Don't worry about efficiency here -- it shouldn't happen very often
     * (unless we're debugging).
     */

    *stamp_buf = '\0';
    (void) sprintf (stamp_buf, "%s", Time_str ());
    if (ShortHost && *ShortHost) {
	(void) sprintf (stamp_buf + strlen (stamp_buf), " %s", ShortHost);
    }
    if (Name && *Name) {
	(void) sprintf (stamp_buf + strlen (stamp_buf), " %s", Name);
    }
    if (Debug || Testing_mode) {
        (void) sprintf (stamp_buf + strlen (stamp_buf), "[%ld]", (long) getpid ());
	(void) sprintf (stamp_buf + strlen (stamp_buf), "%s", logmsg (kind));
    }

    (void) sprintf (stamp_and_log, "%s %s\n", stamp_buf, log_buf);
    stamp_and_log_len = strlen (stamp_and_log);

    /* use writes here: on HP/UX, using fprintf really screws up.
     * TODO: use writev().
     */

    if (Echo_on_stdout) {
	write (1, stamp_and_log, stamp_and_log_len);
    }
    write (2, stamp_and_log, stamp_and_log_len);

    if (kind != XLOG_DEBUG) {
	(void) strcpy (Last_errormsg, stamp_and_log);	/* for error mails */
    }
    plp_unblock_signals ();
    errno = tmp_errno;
}

/* VARARGS2 */
void
#ifdef HAVE_STDARGS
log (int kind, char *msg,...)
#else
log (va_alist) va_dcl
#endif
{
#ifndef HAVE_STDARGS
    int kind;
    char *msg;
#endif
    VA_LOCAL_DECL

    *log_buf = '\0';
    if (Printer && *Printer) { (void) sprintf (log_buf, "%s: ", Printer); }
    VA_START (msg);
    VA_SHIFT (kind, int);
    VA_SHIFT (msg, char *);
    (void) vsprintf (log_buf + strlen (log_buf), msg, ap);
    VA_END;
    log_backend (kind);
}

/* VARARGS2 */
void
#ifdef HAVE_STDARGS
fatal (int kind, char *msg,...)
#else
fatal (va_alist) va_dcl
#endif
{
#ifndef HAVE_STDARGS
    int kind;
    char *msg;
#endif
    VA_LOCAL_DECL

    *log_buf = '\0';
    if (Printer && *Printer) { (void) sprintf (log_buf, "%s: ", Printer); }

    VA_START (msg);
    VA_SHIFT (kind, int);
    VA_SHIFT (msg, char *);
    (void) vsprintf (log_buf + strlen (log_buf), msg, ap);
    VA_END;

    log_backend (kind);
    cleanup (0);
    exit (Errorcode);
}

/* VARARGS2 */
void
#ifdef HAVE_STDARGS
logerr (int kind, char *msg,...)
#else
logerr (va_alist) va_dcl
#endif
{
#ifndef HAVE_STDARGS
    int kind;
    char *msg;
#endif
    VA_LOCAL_DECL
    int err = errno;

    *log_buf = '\0';
    if (Printer && *Printer) { (void) sprintf (log_buf, "%s: ", Printer); }
    VA_START (msg);
    VA_SHIFT (kind, int);
    VA_SHIFT (msg, char *);
    (void) vsprintf (log_buf + strlen (log_buf), msg, ap);
    VA_END;
    (void) sprintf (log_buf + strlen (log_buf), ": %s", Errormsg (err));
    log_backend (kind);
    errno = err;
}

/* VARARGS2 */
void
#ifdef HAVE_STDARGS
logerr_die (int kind, char *msg,...)
#else
logerr_die (va_alist) va_dcl
#endif
{
#ifndef HAVE_STDARGS
    int kind;
    char *msg;
#endif
    VA_LOCAL_DECL
    int err = errno;

    *log_buf = '\0';
    if (Printer && *Printer) { (void) sprintf (log_buf, "%s: ", Printer); }

    VA_START (msg);
    VA_SHIFT (kind, int);
    VA_SHIFT (msg, char *);
    (void) vsprintf (log_buf + strlen (log_buf), msg, ap);
    VA_END;

    (void) sprintf (log_buf + strlen (log_buf), ": %s", Errormsg (err));
    log_backend (kind);
    cleanup (0);
    exit (Errorcode);
}

/***************************************************************************
 * Diemsg( char *m1, *m2, ...)
 * print error message to stderr, and die
 ***************************************************************************/

/* VARARGS1 */
void
#ifdef HAVE_STDARGS
Diemsg (char *msg,...)
#else
Diemsg (va_alist) va_dcl
#endif
{
#ifndef HAVE_STDARGS
    char *msg;
#endif
    VA_LOCAL_DECL

    if (Name && *Name) {
	(void) fprintf (stderr, "%s: ", Name);
    }
    (void) fprintf (stderr, "fatal error: ");

    VA_START (msg);
    VA_SHIFT (msg, char *);
    (void) vfprintf (stderr, msg, ap);
    VA_END;

    (void) fprintf (stderr, "\n");
    cleanup (0);
    exit (1);
}

/***************************************************************************
 * Warnmsg( char *m1, *m2, ...)
 * print warning message to stderr
 ***************************************************************************/

/* VARARGS1 */
void
#ifdef HAVE_STDARGS
Warnmsg (char *msg,...)
#else
Warnmsg (va_alist) va_dcl
#endif
{
#ifndef HAVE_STDARGS
    char *msg;
#endif
    VA_LOCAL_DECL

    if (Name && *Name) {
	(void) fprintf (stderr, "%s: ", Name);
    }
    (void) fprintf (stderr, "warning: ");

    VA_START (msg);
    VA_SHIFT (msg, char *);
    (void) vfprintf (stderr, msg, ap);
    VA_END;

    (void) fprintf (stderr, "\n");
}

/* VARARGS1 */
void
#ifdef HAVE_STDARGS
logDebug (char *msg,...)
#else
logDebug (va_alist) va_dcl
#endif
{
#ifndef HAVE_STDARGS
    char *msg;
#endif
    VA_LOCAL_DECL

    *log_buf = '\0';
    if (Printer && *Printer) { (void) sprintf (log_buf, "%s: ", Printer); }
    VA_START (msg);
    VA_SHIFT (msg, char *);
    (void) vsprintf (log_buf + strlen (log_buf), msg, ap);
    VA_END;
    log_backend (XLOG_DEBUG);
}


/*
 * this shouldn't be called before Setuplog() or Std_environ(), as they'll close the
 * syslog's fd (which this keeps open).
 */

/* VARARGS1 */
void
#ifdef HAVE_STDARGS
log_request (char *fmt,...)
#else
log_request (va_alist) va_dcl
#endif
{
#ifndef HAVE_STDARGS
    char *fmt;
#endif
    VA_LOCAL_DECL
    static int log_is_open = 0;
    char buf[LOG_BUFSIZ];

    /* non-setuid mode indicates that this is not being used
     * in the "real world", so don't get noisy. */

    if (Testing_mode)
	return;

    if (!log_is_open) {
        openlog ("lpd", LOG_PID, SYSLOG_FACILITY);
        log_is_open++;
    }

    VA_START (fmt);
    VA_SHIFT (fmt, char *);
    (void) vsprintf (buf, fmt, ap);
    VA_END;

    safe_syslog (LOG_INFO, buf);
    if (Debug) {
	(void) strcpy (log_buf, buf);
	log_backend (XLOG_DEBUG);
    }

    /* exiting will close the logfile */
}
