/***************************************************************************
 * U. Minnesota LPD Software * Copyright 1987, 1988, Patrick Powell
 * version 3.3.0 Justin Mason July 1994
 ***************************************************************************
 * MODULE: Link_support.c
 ***************************************************************************
 * Support for the inter-machine communications
 * Link_open(int retry)
 *  opens a link to the remote host
 * Link_close()
 *  closes the link to the remote Printer
 * Link_send( char *l)
 *  sends a line to the remote host
 * Link_line( int retry, char *l)
 *  opens the link (with or without retry)
 *  sends a line to the remote host
 * Link_confirm()
 *  gets a single character back, confirming the action
 * Link_ack( c )
 *  sends a single character back, confirming the action
 * Link_copy( FILE *fp, count)
 *  copies a file to the remote host;
 *  if count bytes not transfered, an error is generated
 * Link_get()
 *  gets and prints all information  on stdout
 * Link_port_num()
 *	gets remote port number for connection
 ***************************************************************************/

#include "lp.h"
#include "library/errormsg.h"

static int Link_fd;		/* fd for the socket */
static int reserveport (int, int);

/* these are used for non-RFC-compliant lpd's (like WSLPD).
 * Set up alarms so PLP doesn't hang forever during transfers.
 */

static plp_signal_t (*old_alarm_fun)(int);
static int alarm_timed_out;

/* Some timeout values, in seconds. These can be changed externally.
 */
int Connect_timeout = 30;
int Confirm_timeout = 30;

static plp_signal_t 
connect_alarm (int sig) {
    alarm_timed_out = True;
}

/***************************************************************************
 * Link_open(retry)
 * 1. Set up an inet socket;  a socket has a local host/local port and
 *    remote host/remote port address part.  The LPR software runs SUID
 *    root,  and will attempt to open a privileged port (number less than
 *    PRIV);  at the remote end this is checked to make sure that the
 *    remote machine is running SUID root.  Primitive,  but it is adequate
 *    in a trusting environment.
 *
 * (See Advanced 4.3 BSD IPC Tutorial for more information) The server on
 * the remote machine will be listening on the port given by the entry in
 * the /etc/service for the particular service desired.  If we are using
 * the "real" version of lpd this will be the "Printer" service, the test
 * version will use a "test" service.  The LPD daemon will open a socket
 * and bind to the appropriate "port" number.  The
 * getservbyname("printer","tcp"), is used to get the port information
 * (sp->s_port), which is used in a bind call.
 *
 * When we want to communicate with the lpd daemon on the remote machine,
 * listening on that particular port, we call getseverbyname(...) to get
 * the port number and protocol.  The remote host expects the local port
 * to be in a range that is available only to a root UID process, i.e.-
 * less than IPPORT_RESERVED.  When we open the local port, we get a local
 * port in this range.  At the remote end, the port number is checked to
 * ensure that it is in a valid range.  Since the reserved ports can only
 * be accessed by UID 0 (root) processes, this would appear to prevent
 * ordinary users from directly contacting the remote daemon.
 *
 ***************************************************************************/

/*
 * getport() gets a port,  in the correct range,  to the remote machine This code comes
 * from a description of the RLOGIN and LPD materials
 */
int
getport (char *hostname, int port_num) {
    struct hostent *host;	/* host entry pointer */
    int sock;			/* socket */
    int i;			/* generic integers */
    unsigned long inet_a;

#if (SOCK_FAMILY==AF_UNIX)
    struct sockaddr_un sin;	/* unix socket address */
#else
    struct sockaddr_in sin;	/* inet socket address */
#endif

    /*
     * Zero out the sockaddr_in struct
     */
    bzero ((char *) &sin, sizeof (sin));
    /*
     * Get the host address and port number to connect to.
     */
    DBGREM4 ("getport: host %s", hostname);
#if (SOCK_FAMILY==AF_INET)
    if ((host = (gethostbyname (hostname))) != NULL) {
	/*
	 * set up the address information
	 */
	bcopy (host->h_addr, (caddr_t) & sin.sin_addr, host->h_length);
	sin.sin_family = host->h_addrtype;
    } else {
	DBGREM3 ("getport: unknown host '%s', trying inet address", hostname);
	inet_a = inet_addr (hostname);
	if (inet_a == (unsigned long) (-1)) {
	    fatal (XLOG_INFO, "getport: unknown host '%s'", hostname);
	}
	sin.sin_addr.s_addr = inet_a;
	sin.sin_family = AF_INET;
	DBGREM4 ("getport: inet address %s", hostname);
    }
#endif
    /*
     * get the server name and protocol information from /etc/services
     */
    if (port_num == 0) {
	port_num = Link_port_num ();
    }
    /*
     * set up the address information
     */
#if (SOCK_FAMILY==AF_UNIX)
    sin.sun_family = SOCK_FAMILY;
    sprintf (sin.sun_path, "/tmp/lpdsock%d", port_num);
#else
    sin.sin_port = port_num;
#endif

    /*
     * Try connecting to the server.
     */
    DBGREM3 ("getport: trying connection to %s, port %d", hostname, port_num);

    /*
     * After this call, WE ARE ROOT! Remember to change back.
     */
    sock = reserveport (Maxportno, Minportno);
    if (sock < 0) {
	root_to_user ();
	return (-1);
    }
    old_alarm_fun = plp_signal (SIGALRM, connect_alarm);
    alarm_timed_out = 0; alarm (Connect_timeout);

#if (SOCK_FAMILY==AF_UNIX)
    i = connect (sock, (struct sockaddr *) & sin,
	    sizeof (sin.sun_family) + strlen (sin.sun_path));

    if (i < 0) {
	if (DbgRem > 2) {
	    if (errno == EINTR && alarm_timed_out) {
		log (XLOG_DEBUG, "timed out waiting for connect to unix sock");
	    } else {
		logerr (XLOG_DEBUG, "connect failed to unix sock");
	    }
	}
	if (errno == EINTR && alarm_timed_out)
	    errno = ETIMEDOUT;
	(void) close (sock);
	(void) plp_signal (SIGALRM, old_alarm_fun);
	alarm (0);
	root_to_user ();
	return (-1);
    }
#else
    i = connect (sock, (struct sockaddr *) & sin, sizeof (sin));
    if (i < 0) {
	if (DbgRem > 2) {
	    if (errno == EINTR && alarm_timed_out) {
		log (XLOG_DEBUG, "timed out waiting for connect to %s", RM);
	    } else {
		logerr (XLOG_DEBUG, "connect failed to %s", RM);
	    }
	}
	if (errno == EINTR && alarm_timed_out)
	    errno = ETIMEDOUT;
	(void) close (sock);
	(void) plp_signal (SIGALRM, old_alarm_fun);
	alarm (0);
	root_to_user ();
	return (-1);
    }
#endif
    (void) plp_signal (SIGALRM, old_alarm_fun);
    alarm (0);
    root_to_user ();
    return (sock);
}

/*
 * reserveport(int port_no, min_port_no) Reserve a port, starting at port_no, down to
 * min_port_no. while port_no > min_port_no try to grab the port; if you can, then return
 * socket else return -1 Returns: socket if successful, -1 if not
 */
#if (SOCK_FAMILY==AF_UNIX)
static int
reserveport (int port_no, int min_port_no) {
    int sock;

    user_to_root ();
    sock = socket (SOCK_FAMILY, SOCK_STREAM, 0);
    if (sock < 0) {
	root_to_user ();
	logerr_die (XLOG_INFO, "reserveport socket call failed");
    } else if (sock == 0) {
	root_to_user ();
	logerr_die (XLOG_INFO, "reserveport: socket returns 0");
    }
    return (sock);
}

#else
#ifdef HAVE_RRESVPORT

static int
reserveport (int port_no, int min_port_no) {
    int sock;

    user_to_root ();		/* WE STAY ROOT AFTER THIS! Call root_to_user() yourself */

    /* with rresvport(), we can ignore min_port_no. */
    if ((sock = rresvport (&port_no)) == -1) {
	logerr_die (XLOG_INFO, "rresvport");
    }
    DBGREM4 ("reserveport got socket %d", port_no);
    return (sock);
}

#else				/* no rresvport() */

/*
 * This often doesn't work. YMMV; SunOS and Solaris don't like it, at least. Use
 * rresvport() if you can.
 */
static int
reserveport (int port_no, int min_port_no) {
    struct sockaddr_in sin;
    int sock;

    sock = socket (AF_INET, SOCK_STREAM, 0);
    if (sock < 0) {
	logerr_die (XLOG_INFO, "reserveport socket call failed");
    } else if (sock == 0) {
	logerr_die (XLOG_INFO, "reserveport: socket returns 0");
    }
    user_to_root ();		/* grep: reserveport, no rresvport() */
#ifdef TEST_PRIV
    printf ("reserveport: invoked by: uid=%d, euid=%d\n", getuid (), geteuid ());
#endif
    while (port_no >= min_port_no) {
	bzero ((char *) &sin, sizeof (sin));

	sin.sin_family = AF_INET;
	sin.sin_addr.s_addr = 0;
	sin.sin_port = htons ((u_short) port_no);

	if (bind (sock, (struct sockaddr *) & sin, sizeof (sin)) != -1) {
	    DBGREM4 ("reserveport got socket %d", port_no);
	    /* WE STAY ROOT AFTER THIS! Call root_to_user() yourself */
	    return (sock);
	}
	if (errno != EADDRINUSE && errno != EADDRNOTAVAIL) {
	    /* WE STAY ROOT AFTER THIS! Call root_to_user() yourself */
	    logerr_die (XLOG_INFO, "reserveport: bind failed");
	}
	--port_no;
    }
    /* WE STAY ROOT AFTER THIS! Call root_to_user() yourself */
    (void) close (sock);
    return (-1);
}

#endif				/* no rresvport() */
#endif				/* AF_INET sockets */

int
Link_open (int retry) {
    unsigned int i = 1;
    int l;

    if (Link_fd) {
	return (JSUCC);
    }
    while (Link_fd == 0) {
	DBGREM3 ("Link open: attempt %d (max retries %d)", i, retry);
	l = getport (RM, 0);
	if (l < 0) {
	    if ((retry >= 0) && (i++ > (unsigned) retry)) {
		DBGREM3 ("Link_open: retries expired; failed to open link");
		return (JFAIL);
	    }
	    sleep ((unsigned) 15);

	} else if (l == 0) {
	    fatal (XLOG_INFO, "Link_open: Argh! cannot happen - fd = 0");
	} else {
	    Link_fd = l;
	}
    }
    DBGREM3 ("Link_open: opened connection to %s", RM);
    return (JSUCC);
}

/***************************************************************************
 * 1. close the link
 ***************************************************************************/
void
Link_close (void) {
    if (Link_fd) {
	DBGREM4 ("Link_close: closing connection");
	(void) close (Link_fd);
    }
    Link_fd = 0;
}

/***************************************************************************
 * Link_line( int retry; char *line )
 * send a line to the remote end
 * if retry != 0, blocks until connection made
 ***************************************************************************/

int
Link_line (int retry, char *str) {
    if (Link_open (retry) != JSUCC) {
	return (JFAIL);
    }
    return (Link_send (str));
}

int
Link_send (char *str) {
    int i, l;			/* ACME Integers, Inc. */

    l = strlen (str);
    if (Link_fd == 0) {
	log (XLOG_INFO, "Link_send: link to %s not open", RM);
	return (JFAIL);
    }
    if (l != 0 && (i = write (Link_fd, str, l)) != l) {
	if (i < 0) {
	    logerr (XLOG_INFO, "write error to remote site %s", RM);
	}
	Link_close ();
	return (JFAIL);
    }
    DBGREM4 ("Link_send: sent to %s: '%d'%s", RM, *str, str + 1);
    return (JSUCC);
}

/***************************************************************************
 * Link_confirm()
 *  gets a single character back, confirming the action
 ***************************************************************************/

int
Link_confirm (void) {
    char buf[256];		/* buffer, one line (ish) */
    int n;

    if (Link_fd == 0) {
	log (XLOG_INFO, "link to %s not open", RM);
	return (JFAIL);
    }
    old_alarm_fun = plp_signal (SIGALRM, connect_alarm);
    alarm_timed_out = 0; alarm (Confirm_timeout);

    n = read (Link_fd, buf, 1);
    (void) plp_signal (SIGALRM, old_alarm_fun);
    alarm (0);

    if (n != 1) {
	if (errno == EINTR && alarm_timed_out) {
	    log (XLOG_INFO, "timed out waiting for confirm from %s", RM);
	} else {
	    if (n == 0) {
		logerr (XLOG_INFO, "remote host %s: failed to ack transfer", RM);
	    } else {
		logerr (XLOG_INFO, "remote host %s: illegal ack code '%s'", RM, buf);
	    }
	}
	Link_close ();
	return (JNOCONF);		/* the job may still have printed */
    }

    if (buf[0] == 0) {
	DBGREM4 ("Link_confirm: successful confirm from %s", RM);
	return (JSUCC);
    } else if (buf[0] == '\002') {
	DBGREM4 ("Link_confirm: spool area full from %s", RM);
	return (JFULL);
    } else {
	/* read more, in case it's a line (shouldn't be, though!) */
	n = read (Link_fd, buf + sizeof(char), sizeof (buf) - 2);
	if (n >= 0) { buf[n] = '\0'; }	/* just in case */
	DBGREM4 ("Link_confirm: bad confirm from %s: \"%s\"", RM, buf);
	Link_close ();
	return (JFAIL);
    }
}

/***************************************************************************
 * Link_copy( FILE *fp; long count; char *name)
 *  copies a file to the remote nost;
 *  if count bytes not transfered, an error is generated
 ***************************************************************************/

int
Link_copy (FILE *fp, long count, char *name) {
    char is_data_file;
    char buf[BUFSIZ];		/* buffer */
    int i, l;			/* ACME Integer, Inc. */
    long total;			/* total number of bytes */
    long sum = 0; 
    int percent = 0;            /* percentage of printed bytes */
    FILE *fp_progress;          /* file pointer for the progress file */
    char file_progress[BUFSIZ]; /* progress file name */

    DBGREM4 ("Link_copy: sending %s (%d bytes) to %s", name, count, RM);
    if (Link_fd == 0) {
	log (XLOG_INFO, "link is not open");
	goto error;
    }
    total = 0;

    /* we don't want to have the size of the control-file, so we check the 
     * first 2 chars in *name; 
     * if *name starts with "cf", it's a control-file otherwise 
     * one or more data-files */
    if (name[0] == 'c' && name [1] == 'f') { /* cf file */
	is_data_file = False;
    } else {
	is_data_file = True;
    }

    if (is_data_file) {
     /* Name of the progress file */
	DBGREM4 ("Link_copy: generating progress file");
	(void) sprintf (file_progress, "%s/progress", SD);
	if ((fp_progress = fopen_daemon (file_progress, "w")) == NULL) {
	    log (XLOG_WARNING, "Print_copy: file %s couldn't be opened.", 
		 file_progress);
	}
	/* in case that this file is not owned by daemon...*/
	(void) chown (file_progress, DaemonUID, DaemonGID);
	DBGREM4 ("Link_copy: progress-file %s successfully openend for writing",
		 file_progress);
    }
    while ((i = fread (buf, 1, sizeof (buf), fp)) > 0) {
	total = total + i;
	if (total > count) {
	    log (XLOG_DEBUG,
		 "file '%s': length is %d instead of %d bytes",
		 name, total, count);
	    goto error;
	}
	if ((l = write (Link_fd, buf, i)) != i) {
	    if (l < 0) {
		logerr (XLOG_INFO, "write error while sending '%s' to %s",
			name, RM);
	    } else {
		logerr (XLOG_INFO, "partial write while sending '%s' to %s",
			name, RM);
	    }
	    goto error;
	}
	if (is_data_file) {
	    sum += l;
	    percent = (int)((sum * 100) / count);
	    if (fseek (fp_progress, 0L, SEEK_SET) < 0) {
		log (XLOG_WARNING, "Link_copy: error during positioning in progess-file.");
	    }
	    (void) fprintf (fp_progress, "%d", percent);
	    (void) fflush (fp_progress);
	    DBGREM4 ("Link_copy: printed %d%%", percent);
	}
    }
    if (i < 0) {
	logerr (XLOG_INFO, "file '%s' read error", name);
	goto error;
    } else if (total != count) {
	log (XLOG_DEBUG,
	     "file '%s', copied %d instead of %d bytes to %s",
	     name, total, count, RM);
	goto error;
    }
    if (is_data_file) { /* remove the progress file, it's no longer needed */
	(void) fclose (fp_progress);
	(void) unlink (file_progress);
    }
    return (JSUCC);
error:
    Link_close ();
    if (is_data_file) { /* remove the progress file */
	(void) fclose (fp_progress);
	(void) unlink (file_progress);
    }
    return (JFAIL);
}

int
Link_ack (int c) {
    char buf[1];
    int succ;

    buf[0] = c;

    if (write (Link_fd, buf, 1) != 1) {
	DBGREM4 ("Link_ack: ack '%d' write error to %s", c, RM);
	succ = JFAIL;
	Link_close ();
    } else {
	DBGREM4 ("Link_ack: ack '%d' sent to site %s", c, RM);
	succ = JSUCC;
    }
    return (succ);
}

/***************************************************************************
 * reads all information from the link, and prints on stdout
 ***************************************************************************/
void
Link_get (void) {
    int i;			/* ACME Integers, Inc. */
    char buf[BUFSIZ];		/* buffer */

    if (Link_fd == 0) {
	fatal (XLOG_INFO, "link is not open");
    }
    while ((i = read (Link_fd, buf, sizeof (buf))) > 0) {
	DEBUG3 ("link_get: got %d bytes", i);
	(void) write (1, buf, i);
    }
}

/***************************************************************************
 * - look up the service in the service directory using getservent
 * - if the port number has been set, don't do it a second time.
 ***************************************************************************/

int
Link_port_num (void) {
    struct servent *sp;
    unsigned short int port;

    if (Lpr_port_num == 0) {
	if (!Printer_port || !*Printer_port) {
	    logerr_die (XLOG_INFO, "printer-port is not set!");
	}
	if ((*Printer_port >= '0') && (*Printer_port <= '9')) {
	    /*
	     * starts with a number -> must be a port number. let sscanf do our range
	     * checking: %hu -> unsigned short int
	     */
	    if (sscanf (Printer_port, "%hu", &port) != 1) {
		logerr_die (XLOG_INFO,
			    "%s is not in the valid port range", Printer_port);
	    }
	    Lpr_port_num = port;

	} else {
	    if ((sp = getservbyname (Printer_port, "tcp")) == 0) {
		logerr_die (XLOG_INFO,
			    "getservbyname(\"%s\",tcp) failed", Printer_port);
	    }
	    Lpr_port_num = sp->s_port;
	}
    }
    return (Lpr_port_num);
}

/* "bomb-proof" write -- it'll handle partial writes correctly.
 * quick fix from _UNIX Network Programming_, Stevens (yet another
 * bible!)
 */
int
BPWrite (int fd, char *buf, size_t buflen) {
    int nleft, written;

    nleft = buflen;
    while (nleft > 0) {
	written = write (fd, buf, nleft);
	if (written <= 0) {
	    if (errno == EINTR)
		continue;
	    return written;
	}
	nleft -= written;
	buf += written;
    }
    return buflen;
}
