/***************************************************************************
 * U. Minnesota LPD Software * Copyright 1987, 1988, Patrick Powell
 ***************************************************************************
 * MODULE: checkpc.c
 * Check out a printcap; create spool dirs, fix ownerships, etc.
 ***************************************************************************
 * checkpc: check out the entries in the printcap,
 * checking to see that the default files and directories are present
 * or absent as are indicated by the printcap entry.
 *
 * checkpc [-l] [-s] [-a] [-o] [-f] [-Pprinter] [printcapfile]
 * The default is to create all the various files;
 * the following flags are used to disable automatic creation:
 * -l - no log file
 * -s - no status file
 * -o - no output device
 * -a - no accounting file
 * -f - fix, attempt to create with the desired perms
 * -P printer this printer only
 * printcapfile: name of the printcap file to be used; default is /etc/printcap
 */

#include "lp.h"
#include "library/errormsg.h"
#include "library/readconf.h"
#include "patchlevel.h"

#define FILEMODE	0644
#define DEVMODE		0600
#define DIRMODE		002755		/* drwxr-sr-x (g+s == inherit gid) */

static int nolog;			/* no log file */
static int noout;			/* no output file */
static int nostat;			/* no status file */
static int noacct;			/* no accounting file */
static int fix;				/* fix if bad or missing */
static struct passwd *pwent;		/* password entry */
static struct group *grent;		/* group entry */
static int DU, DG;
static char *Daemon_group;
static int canChown;			/* can do chown */

static void fixup (int, char *, int, char *, char *);
static int fix_system (char *);

extern void Get_Daemon (void);

/*
 * check and fix if necessary
 */
static int all_check(char *check_label, int remove, int type, char *path,
	int perms, char *owner, char *group)
{
    struct stat statbuf;
    char cmd[BUFSIZ];
    char label[BUFSIZ];
    int ok;

    assert (path != (char*) NULL);
    (void) sprintf (label, "%s, %s=%s", Printer, check_label, path);

    ok = True;
    if (stat (path, &statbuf) < 0) {
	printf ("%s: cannot be stated\n", label);
	if (remove) {
	    return True;
	}
	if (fix) {
	    fixup (type, path, perms, owner, group);
	}
	return False;
    }

    if (remove) {
	(void) printf ("%s: removing\n", label);
	(void) sprintf (cmd, "rm -rf %s", path);
	if (fix_system (cmd))
	    return False;
    }

    if ((statbuf.st_mode & type) == 0) {
	(void) printf ("%s: not a %s\n", label,
			(type == S_IFDIR) ? "directory" : "file/device");
	return False;
    }
    if (strcmp (path, "/dev/null")) {
	/* don't try and fix /dev/null!! */
	if (statbuf.st_uid != DU) {
	    ok = False;
	    (void) printf ("%s: incorrect owner (%s)\n", label, owner);
	    if (fix && canChown) {
		fixup (type, path, perms, owner, NULL);
		return False;
	    }
	}
	if (group && statbuf.st_gid != DG) {
	    ok = False;
	    (void) printf ("%s: incorrect group (%s)\n", label, group);
	    if (fix && canChown) {
		fixup (type, path, perms, NULL, group);
		return False;
	    }
	}
	/* if mode & perms == perms, it's grand for our purposes */
	if ((statbuf.st_mode & perms) != perms) {
	    ok = False;
	    (void) printf ("%s: perms are %o\n", label,
			(unsigned int) (statbuf.st_mode & 0777));
	    if (fix) {
		fixup (type, path, perms, owner, group);
		return False;
	    }
	}
    }
    return (ok);
}

/*
 * doprinter: 1. get the printcap entry 2. check the spool direcotry for ownership, etc.
 * 3. check the log, accounting, and lock file for existence.
 */

static void doprinter(void) {
    DEBUG2 ("checking", Printer);

    if (Get_pc_entry (Printer, All_pc_vars, All_pc_len) == False) {
	(void) printf ("%s: not a printer\n", Printer);
	return;
    }

    /* check to see if the spool directory exists */
    (void) all_check ("sd", 0, S_IFDIR, SD, DIRMODE, Daemon_user, Daemon_group);

    if (chdir (SD) < 0) {
	logerr (XLOG_DEBUG, "cannot chdir to '%s'!\n", SD);
	return;
    }

    if (noout == False && LP && *LP) {
	/* ignore the LP check if it's an LP-pipe */
	if (*LP != '|') {
	    (void) all_check ("lp", noout, S_IFREG | S_IFBLK,
				LP, DEVMODE, Daemon_user, Daemon_group);
	}
    }
    if (nolog == False && LF && *LF) {
	(void) all_check ("lf", nolog, S_IFREG, LF, FILEMODE, Daemon_user, Daemon_group);
    }
    if (LO && *LO) {
	(void) all_check ("lo", 0, S_IFREG, LO, FILEMODE, Daemon_user, Daemon_group);
    }
    if (!noacct && AF && *AF) {
	(void) all_check ("af", noacct, S_IFREG, AF, FILEMODE, Daemon_user, Daemon_group);
    }
    if (nostat == False && ST && *ST) {
	(void) all_check ("st", nostat, S_IFREG, ST, FILEMODE, Daemon_user, Daemon_group);
    }
    if (PS && *PS) {
	(void) all_check ("ps", nostat, S_IFREG, PS, FILEMODE, Daemon_user, Daemon_group);
    }
}

plp_signal_t
cleanup(int sig) {;}

static void
usage (void) {
    (void) printf ("Usage: %s [-aflso][-Pprinter] [printcapfile]\n", Name);
    exit (1);
}

static int fix_system (char *cmd) {
    int status;
    static char path_decl[BUFSIZ];
    static char *cmd_with_path = NULL;

    assert(cmd!=(char*)NULL);
    (void) printf ("\tFIX: %s\n", cmd);

    if (cmd_with_path == NULL) {
	/* try and ensure chown is in the path. */
	(void) strcpy (path_decl, "PATH=/etc:/usr/etc:/usr/sbin:$PATH; ");
	cmd_with_path = &(path_decl [strlen (path_decl)]);
    }

    (void) strcpy (cmd_with_path, cmd);

    if ((status = system (path_decl)) != 0) {
	(void) printf ("\tFIX: failed, try by hand (pwd=%s)!\n", SD);
	return status;

    } else {
	return 0;
    }
}

/*
 * try to create the file/directory with the appropriate permissions
 */
static void fixup (int type, char *path,
	int perms, char *owner, char *group)
{
    char cmd[BUFSIZ];
    struct stat statbuf;

    if (stat (path, &statbuf) < 0) {
	if (type == S_IFDIR) {
	    (void) sprintf (cmd, "mkdir %s", path);
	} else {
	    (void) sprintf (cmd, "touch %s", path);
	}
        fix_system (cmd);
    }

    if (stat (path, &statbuf) < 0) {
	(void) printf ("%s: cannot stat -- try by hand!\n", path);
    }

    if (canChown) {
	/* fix up ownership and perms, you are root */
	if (type == S_IFDIR) {
	    if (owner) {
		(void) sprintf (cmd, "cd %s; chown %s .", path, owner);
		fix_system (cmd);
	    }
	    if (group) {
		(void) sprintf (cmd, "cd %s; chgrp %s .", path, group);
		fix_system (cmd);
	    }
	} else {
	    if (owner) {
		(void) sprintf (cmd, "chown %s %s", owner, path);
		fix_system (cmd);
	    }
	    if (group) {
		(void) sprintf (cmd, "chgrp %s %s", group, path);
		fix_system (cmd);
	    }
	}
    }

    (void) sprintf (cmd, "/bin/chmod %o %s", perms, path);
    fix_system (cmd);
}

int
main(int argc, char *argv[], char *envp[]) {
    char **prlist;
    int option;
    char *tokpath;

    while ((option = Getopt (argc, argv, "D:afolsP:")) != EOF) {
	switch (option) {
	case 'D':		/* turn on Debugging */
	    Parse_debug (Optarg);
	    break;
	default:
	    usage ();
	    break;
	case 'a':		/* no acct file */
	    ++noacct;
	    break;
	case 'f':		/* no acct file */
	    ++fix;
	    break;
	case 'o':		/* no output file */
	    ++noout;
	    break;
	case 'l':		/* no log file */
	    ++nolog;
	    break;
	case 's':		/* no status file */
	    ++nostat;
	    break;
	case 'P':		/* printer */
	    if (Optarg) {
		Printer = Optarg;
	    } else {
		usage ();
	    }
	    break;
	}
    }

    Std_environ (argc, argv, envp);
    Readconf ();

    if (Printcap_path == NULL || !*Printcap_path) {
	logerr_die (XLOG_INFO, "printcap-path is unset!");
    }
    if (Optind < argc) {
	tokpath = (char *) malloc (strlen (Printcap_path) + strlen (argv[Optind]) + 2);
	(void) sprintf (tokpath, "%s:%s", argv[Optind], Printcap_path);
	++Optind;
    } else {
	tokpath = (char *) strdup (Printcap_path);
    }
    if (Optind < argc) {
	usage ();
    }
    canChown = (getuid() == 0);

    if (canChown == False) {
	(void) printf ("%s: warning: not running as root, can't chown\n", Name);
    }
    if (Daemon_user == NULL) {
	logerr_die (XLOG_DEBUG, "haven't read config file yet");
    }
    Get_Daemon ();
    pwent = getpwnam (Daemon_user); assert(pwent);
    DU = DaemonUID;
    DG = DaemonGID;

    if ((grent = getgrgid (DG)) == NULL) {
	logerr_die (XLOG_DEBUG, "getgrgid failed for gid %d", DG);
    }
    Daemon_group = grent->gr_name;

    {
	char *pc;

	pc = (char *) strtok (tokpath, ":");
	do {
	    if (pc[0] != '/') {
		(void) Diemsg ("Printcap argument must be absolute path name");
	    }
	} while (strtok (NULL, ":"));
    }
    free (tokpath);

    if (Printer) {
	doprinter ();
    } else {
	(void) printf ("Checking all printers\n");
	for (prlist = All_printers (); (Printer = *prlist); ++prlist) {
	    doprinter ();
	}
    }
    return False;
}
