/***************************************************************************
 * U. Minnesota LPD Software * Copyright 1987, 1988, Patrick Powell
 ***************************************************************************
 * MODULE: pac.c
 * Accounting Summary
 ***************************************************************************/

#include "lp.h"
#include "patchlevel.h"
#include "library/errormsg.h"
#include "library/readconf.h"
#include "library/utils.h"

/***************************************************************************
 *      pac - printer/ploter accounting information
 * SYNOPSIS
 *      pac [ -Pprinter ] [ -pprice ] [ -a ] [ -e ] [ -s ] [-m] [-h]
 * DESCRIPTION
 *      Pac reads the printer/plotter accounting files, accumulating
 *      the number of pages (the usual case) or feet (for raster
 *      devices) of paper consumed by each user, and printing out
 *      how much each user consumed in pages or feet and dollars.
 *      -Pprinter
 *           Selects the printer.  If not specified, the value of
 *           the environment variable PRINTER is used, and if it is
 *           not defined, the default printer for the site.
 *      -pprice
 *           The price per page to be used, instead of the printcap
 *           co (cost) entry.  If none is specified, the default
 *           price of $0.02 per page is used.
 *      -a   Accounting is done for all printers.  A cumulative sum-
 *           mary of usage of all the entries will be printed.
 *      -e   Accounting is done for each printer entry in turn.
 *      -s   Summarize the accounting information in the summary
 *           accounting file.  Used with the -a or -e flag all the
 *           accounting information for each printer can be done at
 *           once.  This option has the side effect of truncating
 *           the original file.  It copies the original file to a
 *           backup before doing any action.
 *      -b   writes a bill on a file in /tmp under the name 
 *           plp_bill.<username>; this may be submitted to the user
 *
 *      The accounting file is specified in the printcap af entry.
 *      The summary accounting file has the .sum suffix added.
 * FILES
 *      printcap af entry
 * NOTES
 *      The accounting information is generated by the various
 *      filters used by the LPD daemon.  For each user file they
 *      should write a line to the accounting file in the following
 *      format.
 *           user     host   printer pages format jobname  seq-number date
 *           papowell umn-cs ps      10     C     Makefile 367        April 13 1988
 *      The summary file will produce entries of the form:
 *           user     host   printer pages format runs  cost
 *           papowell umn-cs ps      1320  f      17   32.00
 ***************************************************************************/

static int All = 0;			/* do for all printers */
static int Each = 0;			/* do for each printer */
static int Summary = 0;			/* do a summary for each printer */
static int NoMachines = 0;		/* disregard machine names */
static int Header = 0;			/* print header */
static long total_pages = 0;
static float pagecost = 0.02;		/* cost per page (or what ever) */
static float total = 0.0;               /* total summary */
static short price_set = 0;
static int print_bill = 0;              /* Flag if a bill should be generated */
/*
 * Hash Table Each user has his own entry
 */

#define	HASHSIZE	97	/* Number of hash buckets */

/* for the currency sign */
char curr_sign[6];

char date[32];

struct entry {
    struct entry *next;		/* Forward hash link */
    char user[32];		/* user name */
    char host[32];		/* host name */
    char printer[32];		/* host name */
    long pages;			/* Feet or pages of paper */
    char format[32];		/* format */
    long runs;			/* Number of runs */
    float price;		/* cost for this */
    struct entry *chain;	/* list link */
};

static struct entry *Hash_table[HASHSIZE];	/* Hash table proper */
static struct entry *All_user, *Sum_user, *Raw_user;

static struct entry *Add_entry (struct entry **, char *, char *, char *, char *);
static struct entry *Make_entry (struct entry **, char *, char *, char *, char *);
static struct entry *Find_entry (char *, char *, char *, char *);
static short new_summary_file = 0;

void Dump_info (), Sort_file (), Update_sum (), Merge_raw (), Merge_raw_all (),
	Merge_raw_sum (), Free_entry (), Get_summary (), Process (),
	Check_dup (), Dienoarg (), Check_str_dup (), Check_int_dup (),
	Get_parms (), Clean_hash (), Get_raw_file (), Generate_Bill (struct entry *, char *);
int comp_array ();

int
main (int argc, char **argv, char **envp) {
    char **list;
    char *lang = NULL;

    /*
     * set umask to avoid problems with user umask
     */
    (void) umask (0);

    /*
     * process the parameters
     */
    Get_parms (argc, argv);
    Std_environ (argc, argv, envp);
    Readconf ();

    if (!All && !Each)
	Get_Printer (0);
    /*
     * set default format
     */
    if (!FROMHOST[0]) {
	(void) strcpy (FROMHOST, Host);
    }

    /* get the value of the environment variable LANG... for internationalization */
    lang = (char *) getenv ("LANG");

    if (!lang) { /* No language, setting to default (DM) */
	(void) strcpy (curr_sign, " DM:");
    } else {
	if (!strcmp (lang, "de")) {
		(void) strcpy (curr_sign," DM: ");
	} else {
	    if (! strcmp (lang, "en_GB")) {
		(void) strcpy (curr_sign, "  : ");
	    } else {
		if (!strcmp (lang, "it")) {
		    (void) strcpy (curr_sign, "Lit.:");
		} else {
		    if (!strcmp (lang, "fr")) {
			(void) strcpy (curr_sign," FF: ");
		    } else {
			(void) strcpy (curr_sign," US$:");
		    }
		}
	    }
	}
    }

    DEBUG4 ("Printer %s, All %d, Each %d, Summary %d",
			Printer, All, Each, Summary);

    if (All | Each) {
	for (list = All_printers (); *list; ++list) {
	    Printer = *list;
	    Process ();
	}
    } else {  
	Process ();
    } 
    if (All) {
	Sort_file (&All_user);
	Dump_info (All_user, NULL);	/* null == all printers */
    }
    return False;
}

/***************************************************************************
 * Get_parms()
 * get the parameters
 ***************************************************************************/
static char *optstr = "D:P:abemp:hs";
void
Get_parms (int argc, char **argv) {
    int option;

    while ((option = Getopt (argc, argv, optstr)) != EOF) {
	switch (option) {
	case 'p':
	    pagecost = (float) atof (Optarg);
	    price_set = 1;
	    break;
	case 'D':
	    Parse_debug (Optarg);
	    break;
	case 'P':
	    if (Optarg) {
		if (Printer) {
		    Check_str_dup (option, Printer, Optarg);
		}
		Printer = Optarg;
	    }
	    break;
	case 'e':
	    Check_dup (option, &Each);
	    break;
	case 'a':
	    Check_dup (option, &All);
	    break;
	case 'b':
	    print_bill = 1;
	    break;
	case 's':
	    Check_dup (option, &Summary);
	    break;
	case 'm':
	    Check_dup (option, &NoMachines);
	    break;
	case 'h':
	    Check_dup (option, &Header);
	    break;
	case '?':
	    break;
	default:
	    fatal (XLOG_INFO, "Get_parms: badparm %c", option);
	}
    }

    /*
     * set up the Parms[] array
     */
    allocParmsIfNeeded ();
    for (; Optind < argc; ++Optind) {
	Parms[Parmcount].str = argv[Optind];
	++Parmcount;
        growParmsIfNeeded ();
    }
}

/***************************************************************************
 * Check_int_dup( int option, int *value, char *arg )
 * 1.  check to see if value has been set
 * 2.  if not, then get integer value from arg
 ***************************************************************************/

void
Check_int_dup (int option, int *value, char *arg) {
    if (!Optarg || !*Optarg) {
	Dienoarg (option);
    }
    if (*value) {
	Diemsg ("duplicate option %c", option);
    }
    if (sscanf (arg, "%d", value) != 1 || *value <= 0) {
	Diemsg ("option %c parameter (%s) is not positive integer",
		option, arg);
    }
}

/***************************************************************************
 * Check_str_dup( int option, char *value, char *arg )
 * 1.  check to see if value has been set
 * 2.  if not, then set it
 ***************************************************************************/

void
Check_str_dup (int option, char *value, char *arg) {
    if (!Optarg || !*Optarg) {
	Dienoarg (option);
    }
    if (*value) {
	Diemsg ("duplicate option %c", option);
    }
    if ((int) strlen (arg) > MAXPARMLEN) {
	Diemsg ("option %c arguement too long (%s)", option, arg);
    }
    (void) strcpy (value, arg);
}

void
Dienoarg (int option) {
    Diemsg ("option '%c' missing argument", option);
}

/***************************************************************************
 * Check_dup( int option, int *value )
 * 1.  check to see if value has been set
 * 2.  if not, then set it
 ***************************************************************************/
void
Check_dup (int option, int *value) {
    if (*value) {
	Diemsg ("duplicate option %c", option);
    }
    *value = 1;
}

/***************************************************************************
 * Process
 * Get the accounting file.
 * 1. Read printcap entry.
 * 2. Get the accounting file information
 * 3. If we are to summarize,  read the summyary
 * 4. Combine the summary and other file information.
 * 5. if we are to summarize, copy the information to the output summary file
 *    save the original file, and truncate it.
 ***************************************************************************/

void
Process (void) {
    char buf[BUFSIZ];
    FILE *fp, *bp, *sp;		/* Accounting and summary file */

    fp = bp = sp = (FILE *) NULL;
    if (Get_pc_entry (Printer, Status_pc_vars, Status_pc_len) == 0) {
	Warnmsg ("Printer %s does not exist\n", Printer);
	return;
    }
    if (SD == 0 || *SD == 0) {
	DEBUG2 ("not a Printer");
	return;
    }
    if (AF == 0 || *AF == 0) {
	Warnmsg ("Printer %s does not have an accounting file!", Printer);
	return;
    }
    chdir_SD ();
    if ((fp = fopen (AF, "r+")) == 0) {
	Warnmsg ("cannot open acct file for %s (%s): %s",
		Printer, AF, Errormsg (errno));
	return;
    }
    (void) sprintf (buf, "%s.bak", AF);
    if ((bp = fopen (buf, "w")) == NULL) {
	logerr (XLOG_NOTICE, "warning: cannot open backup file %s", buf);
    }
    if (Summary) {
	(void) sprintf (buf, "%s.sum", AF);
	sp = fopen (buf, "r+");
    }

    Get_raw_file (fp, bp);

    if (Summary) {
	if (sp != NULL) {
	    Get_summary (sp);
	    new_summary_file = 0;
	} else {
	    /* summary file doesn't exist yet, so we write a new one */
	    sp = fopen (buf, "w");
	    new_summary_file = 1;
	}
	Merge_raw_sum (&Raw_user);
	Sort_file (&Sum_user);
	Update_sum (sp);
	if (Each) {
	    Dump_info (Sum_user, Printer);
	}
    } else if (All) {
	Merge_raw_all (&Raw_user);
    } else {
	Dump_info (Raw_user, Printer);
    }

    if (Sum_user) {
	Free_entry (Sum_user);
	Sum_user = 0;
    }

    Free_entry (Raw_user);
    Raw_user = 0;
    if (sp) {
	(void) fclose (sp);
	(void) chmod (buf, 00664);
    }
    if (bp) {
	(void) fclose (bp);
	(void) chmod (buf, 00664);
    }
    if (fp) {
	(void) fclose (fp);
	(void) chmod (AF, 00664);
    }
}

plp_signal_t cleanup (int signal) { exit (1); }

/*
 * Get_raw_file(): read and copy the raw accounting file into the backup,
 * set up the raw information.
 */

void
Get_raw_file (FILE *f, FILE *b) {
    char buf[BUFSIZ];
    char user[BUFSIZ];
    char host[BUFSIZ];
    char printer[BUFSIZ];
    char format[BUFSIZ];
    char jobname[BUFSIZ];
    char date_1[4], date_2[4], date_3[4], date_4[9], date_5[5];
    long pages;
    int i, seq; 
    static int first_record;
    struct entry *p; 

    Raw_user = 0;
    first_record = 1;

    Clean_hash ();

    while (fgets (buf, sizeof (buf), f)) {
	(void) fputs (buf, b);

	/* break out the fields */
	if (first_record) {
	    first_record = 0;
	    i = sscanf (buf, "%s %s %s %ld %s %s %d %s %s %s %s %s\n", user, host, printer, &pages, format, jobname, &seq, date_1, date_2, date_3, date_4, date_5);
	    (void) sprintf (date, "%s %s %s %s %s", date_1, date_2, date_3, date_4, date_5);
	} else {
	    i = sscanf (buf, "%s %s %s %ld %s\n", user, host, printer, &pages, format);
	}
	if (i < 5 || (int) strlen (user) > 32 || (int) strlen (host) > 32
	    || pages < 0 || (int) strlen (format) > 32) {
	    Warnmsg ("bad format of accounting information in this line:\n \"%s\"\n",
		     buf);
	}

	if (NoMachines) {
	    p = Add_entry (&Raw_user, user, NULL, printer, format);
	} else {
	    p = Add_entry (&Raw_user, user, host, printer, format);
	}
	p->pages += pages;
	p->runs += 1;
    }
}

/*
 * Get_summary(): read and copy the raw accounting file into the backup,
 * set up the raw information.
 */

void
Get_summary (FILE *f) {
    char buf[BUFSIZ];
    char user[BUFSIZ];
    char host[BUFSIZ];
    char printer[BUFSIZ];
    char format[BUFSIZ];
    long pages;
    long jobs;
    int i;
    struct entry *p;

    Sum_user = 0;

    Clean_hash ();

    while (fgets (buf, sizeof (buf), f)) {

	/* break out the fields */
	i = sscanf (buf, "%s %s %s %ld %s %ld\n",
		    user, host, printer, &pages, format, &jobs);

	if ( strncmp (user ,"=", 1) ) { /* ignore comment line that starts with "=" */ 
	    if (i < 4 || (int) strlen (user) > 32 || (int) strlen (host) > 32
		|| pages < 0 || (int) strlen (format) > 32) {
		Warnmsg ("bad format of accounting information: %s\n", buf);
	    }
	    if (NoMachines) {
		p = Add_entry (&Sum_user, user, NULL, printer, format);
	    } else {
		p = Add_entry (&Sum_user, user, host, printer, format);
	    }
	    p->pages += pages;
	    p->runs += jobs;
	}
    }
}

void
Clean_hash (void) {
    int i;
    for (i = 0; i < HASHSIZE; ++i) {
	Hash_table[i] = 0;
    }
}

int
Hash (char *s) {
    int i;

    i = s[1] * 3 + s[0] + s[strlen (s) - 1] * 7;
    return (i % HASHSIZE);
}

static struct entry *
Find_entry (char *user, char *host, char *printer, char *format) {
    struct entry *p;
    int i;

    i = Hash (user);

    p = Hash_table[i];

    while (p) {
	if (strsame (user, p->user)
	    && (NoMachines || (hostcmp (host, p->host) == 0))
	    && strsame (printer, p->printer)
	    && strsame (format, p->format)) {
	    return (p);
	}
	p = p->next;
    }
    return (p);
}

static struct entry *entry_list;
#define NENTRY 25

static struct entry *
Make_entry (struct entry **list, char *user, char *host,
	char *printer, char *format)
{
    struct entry *p;
    static struct entry blank;	/* all zeros */
    int i;

    if (entry_list == 0) {
	p = entry_list =
		(struct entry *) malloc (sizeof (struct entry) * NENTRY);
	for (i = 0; i < (NENTRY - 1); ++i) {
	    p[i].next = &p[i + 1];
	}
	p[i].next = 0;
    }
    p = entry_list;
    entry_list = p->next;

    *p = blank;
    p->next = 0;
    p->chain = 0;

    i = Hash (user);
    p->next = Hash_table[i];
    Hash_table[i] = p;
    (void) strcpy (p->user, user);
    if (!NoMachines) {
        (void) strcpy (p->host, host);
    }
    (void) strcpy (p->printer, printer);
    (void) strcpy (p->format, format);
    p->chain = *list;

    if ((CO) && (!price_set)) {  /* Costs per page */
	p->price = ((float)CO)/1000.0;
    } else {
	p->price = pagecost;
    }
    
    *list = p;
    return (p);
}

void
Free_entry (struct entry *p) {
    struct entry *q;
    while (p) {
	q = p->chain;
	p->next = entry_list;
	entry_list = p;
	p = q;
    }
}

static struct entry *
Add_entry (struct entry **list, char *user,
	char *host, char *printer, char *format)
{
    struct entry *p;
    p = Find_entry (user, host, printer, format);
    if (p == 0) {
	p = Make_entry (list, user, host, printer, format);
    }
    return (p);
}

void
Merge_raw_sum (struct entry **from) {
    Merge_raw (&Sum_user, from);
}

void
Merge_raw_all (struct entry **from) {
    Merge_raw (&All_user, from);
}

void
Merge_raw (struct entry **s, struct entry **from) {
    struct entry *p, *q;

    /*
     * assume that you have just read in the summary file
     */
    p = *from;
    while (p) {
	q = Make_entry (s, p->user, p->host, p->printer, p->format);
	q->runs += p->runs;
	q->pages += p->pages;
	p = p->chain;
    }
}

void
Update_sum (FILE *fp) {
    struct entry *p;
    int i, n;
    time_t t;
    char   time_buf[26];
    float  price;

    t = time ((time_t *) NULL);
    (void) strcpy (time_buf, (ctime (&t)));
    time_buf[24] = '\0';

    /*
     * next, truncate the file
     */
    if (ftruncate (fileno (fp), 0) < 0) {
	logerr_die (XLOG_INFO, "cannot truncate summary file");
    }
    rewind (fp); /* Setting the file pointer to the beginning of the file */

    /* Getting the number of records */
    for (i = 0, p = Sum_user; p; p = p->chain, ++i); 
    if (!new_summary_file) { 
	n = i / 2;
    } else {
	n = i;
    }
    i = 0;
    (void) fprintf (fp, "==================================================================\n");
    (void) fprintf (fp, "=============      Contents of accounting file      ==============\n");
    (void) fprintf (fp, "==================================================================\n");
    if ((CO) && (!price_set)) {  /* Costs per page */
        price = ((float)CO)/1000.0;
    } else {
	price = pagecost;
    }
    if (NoMachines) {
	for (p = Sum_user; p; p = p->chain) {
	    if (i == n) {
		(void) fprintf (fp, "==================================================================\n");
		(void) fprintf (fp, "=======    Sums calculated on %s    =======\n", time_buf);
		(void) fprintf (fp, "==================================================================\n");
	    
		(void) fprintf (fp, "= User   Host(local)  Printer           PPS Format Runs    Costs =\n");
		(void) fprintf (fp, "=----------------------------------------------------------------=\n");
	    }	    
/*	    (void) fprintf (fp, "%-8s localhost      %-13s %5ld %s     %5ld  %7.2f\n", p->user,  */
	    (void) fprintf (fp, "%-8s localhost      %-13s %5ld %-6s%-5ld  %7.2f\n", p->user, 
			    p->printer, p->pages, p->format, p->runs, (p->pages * price));
	    i++;
	    if ( i > n) {
		total_pages = total_pages + p->pages;
	    }
	}
    } else {
	for (p = Sum_user; p; p = p->chain) {
	    if (i == n) {
		(void) fprintf (fp, "==================================================================\n");
		(void) fprintf (fp, "=======    Sums calculated on %s    ========\n", time_buf);
		(void) fprintf (fp, "==================================================================\n");
		
		(void) fprintf (fp, "= User   Host         Printer           PPS Format Runs    Costs =\n");
		(void) fprintf (fp, "=----------------------------------------------------------------=\n");
	    }	    
	    (void) fprintf (fp, "%-8s %-14s %-13s %5ld %-6s %5ld  %7.2f\n", p->user, p->host,
			    p->printer, p->pages, p->format, p->runs, (p->pages * price));
	    i++;
	    if ( i > n) {
		total_pages = total_pages + p->pages;
	    }
	}
    }
	(void) fprintf (fp, "=----------------------------------------------------------------=\n");
	(void) fprintf (fp, "=                                       Total: %s %13.2f=\n", curr_sign, (total_pages * price));
}

static int
cmpx (const void *l, const void *r) {
    int c;
    struct entry *sl, *sr;

    sl = (struct entry *) l;
    sr = (struct entry *) r;

    if (sl->user == 0) return (-1);
    if (sr->user == 0) return True;
    if ((c = strcmp (sl->user, sr->user)) != 0)
	return (c);

    if (!NoMachines) {
	if (sl->host == 0) return (-1);
	if (sr->host == 0) return True;
	if ((c = hostcmp (sl->host, sr->host)) != 0)
	    return (c);
    }
    
    if (sl->printer == 0) return (-1);
    if (sr->printer == 0) return True;
    return (strcmp (sl->printer, sr->printer));
}

void
Sort_file (struct entry **st) {
    int i, n;
    struct entry *p, *q;

    /* count the number of entries */

    for (i = 0, p = *st; p; p = p->chain, ++i);
    n = i;
    if (n > 0) {
	q = (struct entry *) malloc ((unsigned) n * sizeof (struct entry));
	if (q == NULL) {
	    logerr_die (XLOG_INFO, Malloc_failed_msg);
	}

	for (i = 0, p = *st; i < n; p = p->chain, ++i) {
	    q[i] = *p;
	}
	qsort ((char *)q, i, sizeof (struct entry), cmpx);

	for (i = 0; i < n - 1; ++i) {
	    q[i].chain = &q[i + 1];
	}
	q[i].chain = 0;
    }
}

void
Dump_info (struct entry *p, char *header) {
    time_t t;
    char   time_buf[26];
    char user_merk[16];
    long total_pages = 0;
    int i = 0, j, k = 0;
    struct entry *p_ptr, *p_array;

    printf ("\nHost: %s\n", FROMHOST);

    if (Header) {
	if (header == NULL) {
	    (void) printf ("All printers:\n");
	} else {
	    (void) printf ("Printer \"%s\" (first entry is dated %s):\n", header, date);
	}
	(void) printf ("%-10s %-14s %-13s %5s %-6s %5s %9s   %s\n",
		       "USER", "HOST", "PRINTER", "PPS", "FORMAT", "RUNS",
		       "COSTS", "  PAGEPR."); 
    }

    i = 0;
    p_ptr = p;
    /* Counting the number of records... */
    for (; p_ptr; p_ptr = p_ptr->chain)
      i++;
    j = i;
    p_array = (struct entry *) malloc (i * sizeof (struct entry));
    if (p_array == NULL) {
      logerr_die (XLOG_INFO, "pac: No memory available for arry p_array");
    }

    t = time ((time_t *) NULL);
    (void) strcpy (time_buf, (ctime (&t)));
    time_buf[24] = '\0';
    i = 0;   
    if (NoMachines) {
	for (; p; p = p->chain) {
	    (void) strcpy (p_array[i].user, p->user);
	    (void) strcpy (p_array[i].printer, p->printer);
	    p_array[i].pages = p->pages;
	    (void) strcpy (p_array[i].format, p->format);
	    p_array[i].runs = p->runs;
	    p_array[i].price = p->price;
	    if (print_bill) {
	        Generate_Bill (&p_array[i], time_buf);
	    }
	    i++;
	}

	qsort ((struct entry *) p_array, (size_t) i, 
	       (size_t) sizeof (struct entry), comp_array);
	for (i = 0; i < j; i++) {
	    (void) printf ("%-10s %-14s %-13s %5ld %-6s %5ld %s %7.2f  %3.2f\n", 
			   p_array[i].user, "-", p_array[i].printer, p_array[i].pages,
			   p_array[i].format,
			   p_array[i].runs, curr_sign, 
			   ((float)p_array[i].pages * p_array[i].price),
			   p_array[i].price);  
	    total = total + (float)p_array[i].pages * p_array[i].price;
	    total_pages = total_pages + p_array[i].pages;
	}
    } else {
	for (; p; p = p->chain) {
	    p->host[14] = '\0'; /* brute force to cut the hostname */
	    (void) strcpy (p_array[i].user, p->user);
	    (void) strcpy (p_array[i].host, p->host);
	    (void) strcpy (p_array[i].printer, p->printer);
	    p_array[i].pages = p->pages;
	    (void) strcpy (p_array[i].format, p->format);
	    p_array[i].runs = p->runs;
	    p_array[i].price = p->price;
	    if (print_bill) {
	        Generate_Bill (&p_array[i], time_buf);
		if (strcmp (user_merk, p_array[i].user) != 0) {
		    k++;
		}
		sprintf (user_merk, "%s", p_array[i].user);
	    }
	    i++;
	}
	qsort ((struct entry *) p_array, (size_t) i, 
	       (size_t) sizeof (struct entry), comp_array);

	for (i = 0; i < j; i++) {
	  (void) printf ("%-10s %-14s %-13s %5ld %-6s %5ld %s %7.2f  %3.2f\n",
			 p_array[i].user, p_array[i].host, p_array[i].printer, 
			 p_array[i].pages, 
			 p_array[i].format,
			 p_array[i].runs, curr_sign, 
			 ((float)p_array[i].pages * p_array[i].price),
			 p_array[i].price);  
	  total = total + (float)p_array[i].pages * p_array[i].price;
	  total_pages = total_pages + p_array[i].pages;
	}
    }
    (void) printf ("--------------------------------------------------------------------------------\n");
    (void) printf("Total (calc. on %s): Pages: %-7ld  %s%8.2f\n", 
		  time_buf, total_pages, curr_sign, total);
    if (print_bill) {
        printf ("\npac: %d bills generated in /tmp.\n", k);
    }
}

int 
comp_array (struct entry *l, struct entry * r)
{
  int i;

  i = strcmp (l->user, r->user);
  if (i)
    return (i);
  i = strcmp (l->host, r->host);
  if (i)
    return (i);
  i = strcmp (l->printer, r->printer);
  return (i);
}

void 
Generate_Bill (struct entry *p, char *time_buf)
{

  int i, j, y;
  FILE *bill_fp;
  char filename[128];   /* Holds the filename */
  char matrix[24][255];
  char dummy_s[128];

  for (i = 0; i < 24; i++) {
    for (j = 0; j < 255; j++) {
	   matrix[i][j] = ' ';
    }
  }

  (void) sprintf (filename, "/tmp/PLP_bill.%s", p->user);
  if ((bill_fp = fopen (filename, "a")) == NULL) {
      log (XLOG_INFO, "Error: can't create file %s in /tmp.", filename);
      exit (1);
  }
  
  /* Draw a box around the bill */
  for (i = 0; i < 80; i++) {
    matrix[0][i] = '-';
    matrix[23][i] = '-';
  }
  for (j = 0; j < 24; j++) {
    matrix[j][0] = '|';
    matrix[j][79] = '|';
    matrix[j][80] = '\n';
    matrix[j][81] = '\0';
  }
  matrix[0][0] = ' ';
  matrix[0][79] = ' ';
  matrix[23][0] = ' ';
  matrix[23][79] = ' ';

  /*  sprintf (dummy_s2, "%s", p->user);
  for (i = 0; i < strlen (p->user); i++)
    strcat (dummy_s2, "\b");
  strcat (dummy_s2, p->user);
  for (i = 0; i < strlen (p->user); i++)
    strcat (dummy_s2, "\b");
  strcat (dummy_s2, p->user);
  sprintf (dummy_s3, "%s", p->printer);
  for (i = 0; i < strlen (p->printer); i++)
    strcat (dummy_s3, "\b");
  strcat (dummy_s3, p->printer);
  for (i = 0; i < strlen (p->printer); i++)
    strcat (dummy_s3, "\b");
  strcat (dummy_s3, p->printer); */
  /* prints the user and printername */
  /*  sprintf (dummy_s, "Printer bill for user %s, printer %s:", dummy_s2, dummy_s3);*/
  sprintf (dummy_s, "Printer bill for user %s, printer %s:", p->user, p->printer);
  memcpy (&matrix[4][15], dummy_s, strlen (dummy_s));

  /* prints the date of the bill */
  sprintf (dummy_s, "Date: %s", time_buf);
  memcpy (&matrix[1][2],  dummy_s, strlen (dummy_s));

  /* prints the amount of money to pay */
  sprintf (dummy_s, "You have printed: %5ld %s.", p->pages, p->pages > 1 ? "pages" : "page");
  memcpy (&matrix[8][15],  dummy_s, strlen (dummy_s));
  sprintf (dummy_s, "* price per page:  %2.2f %s", p->price, curr_sign);
  memcpy (&matrix[9][15],  dummy_s, strlen (dummy_s));
  sprintf (dummy_s, "------------------------------");
  memcpy (&matrix[10][15],  dummy_s, strlen (dummy_s));
  sprintf (dummy_s, "Total:       %10.2f %s", (float) p->pages * p->price, curr_sign);
  memcpy (&matrix[11][15],  dummy_s, strlen (dummy_s));
  
  sprintf (dummy_s, "This bill was generated by PLP software (and friends).");
  memcpy (&matrix[22][2],  dummy_s, strlen (dummy_s));
  y = strlen (p->user) * 4 + strlen (p->host) * 4;
  matrix[4][75+y] = '|';
  matrix[4][76+y] = '\n';
  matrix[4][77+y] = '\0';

  /* write the file */
  for (j = 0; j < 24; j++) {
    fprintf (bill_fp, "%s", matrix[j]);
  }
  fclose (bill_fp);
  return;
}
